/* Image Store Cache Memory Driver */

#include <asm/current.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/byteorder.h>
#include <asm/semaphore.h>

#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/delay.h>
#include <linux/stringify.h>

#include <xilinx/xparameters.h>

/* Xilinx Specific Machine Status Register (MSR) Bits */
#define XREG_MSR_APU_AVAILABLE                  0x02000000
#define XREG_MSR_APU_ENABLE                     0x00080000

#define BLKSIZ (112*sizeof(u32))
#define BASEADDR XPAR_XPS_REGISTERINT_BRAM_BASEADDR
#define READ_DONE_INTR XPAR_XPS_INTC_0_RR_PPC_LINUX_IRPT_VIDEO_READ_DONE_INTR
#define WRITE_DONE_INTR XPAR_XPS_INTC_0_RR_PPC_LINUX_IRPT_VIDEO_WRITE_DONE_INTR

#define MODULE_NAME "mossbmf"
MODULE_AUTHOR("GVG");
MODULE_DESCRIPTION("ImageStore Burst Memory FIFO Driver");
MODULE_VERSION("0:0.1-experimental");

/* #define USE_QUAD_OPS 1 */

/*  Custom Instructions for FCM test */
/* lwfcmx=lvewx, stwfcmx=stvewx for GNU compier vs. Xilinx EDK compiler */
static inline void loadblk(u32 *blk, const size_t size)
{
	int i;

	local_irq_disable();
	asm ("nop; nop");
#ifndef USE_QUAD_OPS
	for(i=0;i<size;i+=sizeof(u32)) {
		asm volatile("lvewx 0,%0,%1" : : "b" (blk), "r" (i) );
	}
#else
	for(i=0;i<size;i+=(4*sizeof(u32))) {
		asm volatile("lvx 0,%0,%1" : : "b" (blk), "r" (i) );
	}
#endif
	asm ("nop; nop");
	local_irq_enable();
}

static inline void storeblk(u32 *blk, const size_t size)
{
	int i;

	local_irq_disable();
	asm ("nop; nop");
#ifndef USE_QUAD_OPS
	for(i=0;i<size;i+=sizeof(u32)) {
		asm volatile("stvewx 0,%0,%1" : : "b" (blk), "r" (i) );
	}
#else
	for(i=0;i<size;i+=(4*sizeof(u32))) {
		asm volatile("stvx 0,%0,%1" : : "b" (blk), "r" (i) );
	}
#endif
	asm ("nop; nop");
	local_irq_enable();
}

static int major = 0; /* the major number assigned to our driver */
static int mvcm_init(void);
static void mvcm_cleanup(void);
static unsigned long rdDone = 0; 
static unsigned long rdTally = 0; 
static unsigned long wrDone = 0; 
static unsigned long wrTally= 0; 
static void *cntl = NULL;

DECLARE_MUTEX(mvcm_rmutex);
DECLARE_MUTEX(mvcm_wmutex);
DECLARE_WAIT_QUEUE_HEAD(waitfld);

module_init(mvcm_init);
module_exit(mvcm_cleanup);

int mvcm_open(struct inode *inode,struct file *filep);
int mvcm_release(struct inode *inode,struct file *filep);
ssize_t mvcm_read(struct file *filep,char *buff,size_t count,loff_t *offp );
ssize_t mvcm_write(struct file *filep,const char *buff,size_t count,loff_t *offp );
	
struct file_operations mvcm_fops={
	open: mvcm_open,
	read: mvcm_read,
	write: mvcm_write,
	release:mvcm_release,
};

static inline long mvcm_bytes2blks(const loff_t *offp)
{
	const long remainder = 0xFFFFFFFF%BLKSIZ;
	long blkaddr = 0;
	if(offp) {
		long high32 = (((*offp)>>32) & 0xFFFFFFFF);
		long low32 = ((*offp) & 0xFFFFFFFF);

		blkaddr = low32 / BLKSIZ;
		blkaddr += high32 * (0xFFFFFFFF/BLKSIZ);
		blkaddr += ((high32+1)*(remainder+1))/BLKSIZ;
	}
	return blkaddr;
}

int mvcm_open(struct inode *inode,struct file *filep)
{
	long ctlReg = 0;
	unsigned int minor = iminor(inode);
	printk(KERN_DEBUG MODULE_NAME ": open minor=%d\n", minor);

	if(!cntl) cntl = ioremap(BASEADDR + 0x0230, 4*sizeof(long));

	/* tap the FIFO resets */
	if(cntl) {
		ctlReg  = __arch__swab32(ioread32(cntl));
		set_bit(0, &ctlReg); // CPUFifoWrRst
		set_bit(1, &ctlReg); // CPUFifoRdRst
		iowrite32(__arch__swab32(ctlReg), cntl);
		barrier();
		clear_bit(0, &ctlReg); // CPUFifoWrRst
		clear_bit(1, &ctlReg); // CPUFifoRdRst
		iowrite32(__arch__swab32(ctlReg), cntl);
	}

	rdDone = 0; 
	wrDone = 0; 

	msleep(3);
	return 0;
}

int mvcm_release(struct inode *inode,struct file *filep)
{
	long ctlReg = 0;
	unsigned int minor = iminor(inode);
	void *cntl = ioremap(BASEADDR + 0x0230, 4*sizeof(long));
	printk(KERN_DEBUG MODULE_NAME ": release minor=%d\n", minor);

	/* tap the FIFO resets */
	if(cntl) {
		ctlReg  = __arch__swab32(ioread32(cntl));
		set_bit(0, &ctlReg); /* CPUFifoWrRst */
		set_bit(1, &ctlReg); /* CPUFifoRdRst */
		iowrite32(__arch__swab32(ctlReg), cntl);
		barrier();
		clear_bit(0, &ctlReg); /* CPUFifoWrRst */
		clear_bit(1, &ctlReg); /* CPUFifoRdRst */
		iowrite32(__arch__swab32(ctlReg), cntl);
	}

	if(cntl) iounmap(cntl);
	cntl = NULL;

	rdDone = 0; 
	wrDone = 0; 
	return 0;
}


ssize_t mvcm_read(struct file *filep,char *buff,size_t count,loff_t *offp )
{
	u32 *data = NULL;
        int ret = 0;
	int err = 0;
	off_t offset = 0;
	long ctlReg = 0;
	long blkAddr = 0;
	long numBlks = 0;

	volatile __attribute__((aligned (32))) u32 wd[8]; 

	unsigned int minor = iminor(filep->f_dentry->d_inode);
	printk(KERN_DEBUG MODULE_NAME ": read minor=%d count=%ld\n", minor, (long)count);

	if(count==0) return 0;

        /* keep count within range */
	if(count > 2*PAGE_SIZE) count = 2*PAGE_SIZE;
	numBlks = count / BLKSIZ; 
	if(numBlks%1) numBlks++;

	down_interruptible(&mvcm_rmutex);

	/* set machine state register to initialize APU */
	mtmsr(mfmsr() | XREG_MSR_APU_AVAILABLE | XREG_MSR_APU_ENABLE);
	barrier();

	/* disable video mode read of the FIFO */
	ctlReg  = __arch__swab32(ioread32(cntl));
	clear_bit(5, &ctlReg); clear_bit(6, &ctlReg);
	iowrite32(__arch__swab32(ctlReg), cntl);
	barrier();

	/* set CPUReadbackAddress */
	blkAddr = mvcm_bytes2blks(offp);
	iowrite32(__arch__swab32(blkAddr), cntl+(2*sizeof(long)));

	/* set CPUReadbackLength */
	iowrite32(__arch__swab32(numBlks), cntl+(3*sizeof(long)));

	printk(KERN_DEBUG MODULE_NAME ": read minor=%d blkaddr=%ld size=%ldblks\n", minor, blkAddr, numBlks);

	/* enable video mode read of the FIFO */
        rdDone = 0;
	set_bit(5, &ctlReg); clear_bit(6, &ctlReg);
	iowrite32(__arch__swab32(ctlReg), cntl);
	barrier();

	ret = wait_event_interruptible_timeout(waitfld, (rdDone!=0), HZ);
        rdDone = 0;
        if(ret<0) {
		printk(KERN_ALERT MODULE_NAME " -> read done failed!\n" );
		/* disable video mode read of the FIFO */
		ctlReg  = __arch__swab32(ioread32(cntl));
		clear_bit(5, &ctlReg); clear_bit(6, &ctlReg);
		iowrite32(__arch__swab32(ctlReg), cntl);
                return -ERESTARTSYS;
        }

	data = vmalloc(2*PAGE_SIZE);
	if(data) {
		memset(data, 0x200, 2*PAGE_SIZE); 

		/* read FIFO data */
		storeblk(data, numBlks*BLKSIZ); 

		/* disable video mode read of the FIFO */
		ctlReg  = __arch__swab32(ioread32(cntl));
		clear_bit(5, &ctlReg); clear_bit(6, &ctlReg);
		iowrite32(__arch__swab32(ctlReg), cntl);

		/* copy kernel space buffer to user space*/
		err = copy_to_user(buff,data,count);

		if(data) vfree(data);
		data = NULL;
	}

	up(&mvcm_rmutex);

	if(err) {
		printk(KERN_ALERT MODULE_NAME " -> userspace copy failed!\n" );
		return -EFAULT;
	} 

	*offp += count;
	return count;
}

ssize_t mvcm_write(struct file *filep,const char *buff,size_t count,loff_t *offp )
{
	u32 *data = NULL;
        int ret = 0;
	int err = 0;
	off_t offset = 0;
	long ctlReg = 0;
	long blkAddr = 0;
	long numBlks = 0;

        unsigned int minor = iminor(filep->f_dentry->d_inode);
	printk(KERN_DEBUG MODULE_NAME ": write minor=%d count=%ld\n", minor, (long)count);

	if(count==0) return 0;

        /* keep count within range */
	if(count > 2*PAGE_SIZE) count = 2*PAGE_SIZE;
	numBlks = count / BLKSIZ; 
	if(numBlks%1) numBlks++;

	printk(KERN_DEBUG MODULE_NAME ": write minor=%d blkaddr=%ld count=%ld\n", minor, blkAddr, count);
	ctlReg  = __arch__swab32(ioread32(cntl));

        /* allocate data buffer */
	data = vmalloc(2*PAGE_SIZE);
	if(!data) {
		return -ENOMEM;
	}

	if(data) {
		memset(data, 0x200, 2*PAGE_SIZE); 

	        /* copy user space buffer to kernel space*/
	        err = copy_from_user(data,buff,count);
		if(err) {
	                printk(KERN_ALERT MODULE_NAME " -> Kernel copy failed!\n" );
			if(data) vfree(data);
			data = NULL;
	                return -EFAULT;
	        }
	}

	down_interruptible(&mvcm_wmutex);

	/* set machine state register to initialize APU */
	mtmsr(mfmsr() | XREG_MSR_APU_AVAILABLE | XREG_MSR_APU_ENABLE);
	barrier();

	if(data) {
		/* inhibit FIFO from writing to cache memory */
		clear_bit(4, &ctlReg); clear_bit(7, &ctlReg);
		iowrite32(__arch__swab32(ctlReg), cntl);
		barrier();
 
		/* load FIFO here */
		loadblk(data, numBlks*BLKSIZ); 

		vfree(data);
		data = NULL;
	}

	/* set CPURecordAddress */
	blkAddr = mvcm_bytes2blks(offp);
	iowrite32(__arch__swab32(blkAddr), cntl+(1*sizeof(long)));

	/* start writing contents of FIFO to cache memory */
        wrDone = 0;
	set_bit(4, &ctlReg); 
	iowrite32(__arch__swab32(ctlReg), cntl);
	barrier();

	/* wait for completion */
	ret = wait_event_interruptible_timeout(waitfld, (wrDone!=0), HZ);
        wrDone = 0;

	/* inhibit FIFO from writing to cache memory */
	clear_bit(4, &ctlReg); clear_bit(7, &ctlReg);
	iowrite32(__arch__swab32(ctlReg), cntl);

	up(&mvcm_wmutex);

        if(ret<0) {
		printk(KERN_ALERT MODULE_NAME " -> write done failed!\n" );
        	return -ERESTARTSYS;
        }

	*offp += count;
	return count;
}

static irqreturn_t mvcm_read_done_handler(int irq, void *addr, struct pt_regs *regs)
{
        rdDone = ~0;
        ++rdTally;
	barrier();
	wake_up_interruptible_all(&waitfld);
        return IRQ_HANDLED;
}

static irqreturn_t mvcm_write_done_handler(int irq, void *addr, struct pt_regs *regs)
{
        wrDone = ~0;
        ++wrTally;
	barrier();
	wake_up_interruptible_all(&waitfld);
        return IRQ_HANDLED;
}

static int mvcm_init(void)
{
	int i = 0;
	struct resource *cntl = NULL;
	struct resource *fifo = NULL;
	printk(KERN_DEBUG MODULE_NAME ": initializing\n");
	major = register_chrdev(0,MODULE_NAME,&mvcm_fops);
	if(!major){
		printk(KERN_ALERT MODULE_NAME ": failed to register\n");
		return -EIO;
	}
	printk(KERN_INFO MODULE_NAME ": major=%d, pvr=%x\n", major, mfspr(0x11f));

        cntl = request_mem_region(
				BASEADDR + 0x0230, 
				4*sizeof(long),
				MODULE_NAME
				);
	if(!cntl) {
        	printk(KERN_ALERT MODULE_NAME ": cannot access control space\n");
		return -EFAULT;
	}

        /* initialize interrupt handler */
	rdDone = 0; 
        i = request_irq(READ_DONE_INTR,
                        &mvcm_read_done_handler, 0, "mvcm", &rdDone);
        if(i<0) {
                printk(KERN_ERR MODULE_NAME "irq %d unavailable\n", 
			READ_DONE_INTR);
                return i;
        }

	wrDone = 0; 
	i = request_irq(WRITE_DONE_INTR,
                        &mvcm_write_done_handler, 0, "mvcm", &wrDone);
        if(i<0) {
                printk(KERN_ERR MODULE_NAME "irq %d unavailable\n", 
			WRITE_DONE_INTR);
                return i;
        }
	return 0;
}

static void mvcm_cleanup(void)
{
        disable_irq(READ_DONE_INTR);
        disable_irq(WRITE_DONE_INTR);
        free_irq(READ_DONE_INTR, &rdDone);
        free_irq(WRITE_DONE_INTR, &wrDone);
	rdDone = 0; 
	wrDone = 0; 

	if(cntl) iounmap(cntl);
	cntl = NULL;
        release_mem_region(BASEADDR + 0x0230, 4*sizeof(long));

	printk(KERN_DEBUG MODULE_NAME ": terminating\n");
	printk(KERN_ALERT MODULE_NAME ": rdTally=%ld\n", rdTally);
	printk(KERN_ALERT MODULE_NAME ": wrTally=%ld\n", wrTally);
	unregister_chrdev(major,MODULE_NAME);
	return ;
}

