/**
 * @file
 * @brief driver for the GV SDI Capture hardware
 * @copyright 2008 Thomson GV <adam.clarke@thomson.net>
 *
 * Character driver.  Using the standard open() and read() system calls,
 * one can communicate with the FPGA SDI port control registers. 
 */

#include <asm/byteorder.h>
#include <asm/current.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/uaccess.h>

#include <linux/device.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>

#include <xilinx/xparameters.h>
#if 0
#define BASEADDR XPAR_RAMSTORE2PLB_BASEADDR
#else
#define BASEADDR XPAR_XPS_REGISTERINT_BRAM_BASEADDR
#endif

typedef struct {
	loff_t baseaddr;
	size_t size;
} regmap_t;

typedef struct {
	unsigned long top;
	unsigned long left;
	unsigned long bottom;
	unsigned long right;
} frame_t;

typedef struct {
	unsigned long Y;
	unsigned long Cr;
	unsigned long Cb;
} color_t;

#define MODULE_NAME "mosscdr"
MODULE_AUTHOR("GVG");
MODULE_DESCRIPTION("ImageStore SDI Channel Driver");
MODULE_VERSION("0:0.1-experimental");

static const unsigned long PIXELS_IN_BLK = 112L;

static const regmap_t regmap[] = {
	{ 0x0020, 0x0210 }, /* zero means all ports */
	{ 0x0020, 0x0058 },
	{ 0x0078, 0x0058 },
	{ 0x00D0, 0x0058 },
	{ 0x0128, 0x0058 },
	{ 0x0180, 0x0058 },
	{ 0x01D8, 0x0058 },
	{ ~0, ~0 } /* placeholder, end of array */
};

static unsigned int mcdr_major = 0; /* the major number assigned to our driver */
static int mcdr_init(void);
static void mcdr_cleanup(void);

module_init(mcdr_init);
module_exit(mcdr_cleanup);

static int mcdr_open(struct inode *inode,struct file *filep);
static int mcdr_release(struct inode *inode,struct file *filep);
static ssize_t mcdr_read(struct file *filep,char *buff,size_t count,loff_t *offp);
static ssize_t mcdr_write(struct file *filep,const char *buff,size_t count,loff_t *offp);
	
/** @brief supported file operations */
struct file_operations mcdr_fops={
	open: mcdr_open,
	read: mcdr_read,
	write: mcdr_write,
	release:mcdr_release,
};

/** @brief called when the device file is opened in userspace */
int mcdr_open(struct inode *inode,struct file *filep)
{
        unsigned int minor = iminor(inode);
        printk(KERN_DEBUG MODULE_NAME ": open minor=%d\n", minor);

	return 0;
}

/** @brief called when the device file is closed in userspace */
int mcdr_release(struct inode *inode,struct file *filep)
{
        unsigned int minor = iminor(inode);
        printk(KERN_DEBUG MODULE_NAME ": release minor=%d\n", minor);

	return 0;
}

/** @brief read from register associated with the provided minor number */
ssize_t mcdr_read(struct file *filep,char *buff,size_t count,loff_t *offp)
{
	int err;
	char data[PAGE_SIZE];
	unsigned int minor = iminor(filep->f_dentry->d_inode);

	printk(KERN_DEBUG MODULE_NAME "-read[%d]: count=%lu, offset=%lu\n",
				 minor, (unsigned long)count, (unsigned long)*offp );

	if(count > sizeof(data)) count = sizeof(data);

       	memset(data, 0, sizeof(data));

        switch(minor) {
	case  0:
	case  1:
	case  2:
	case  3:
	case  4:
	case  5:
	case  6:
	{
		void *addr = NULL;
		if(count > regmap[minor].size) count = regmap[minor].size;
	        addr = ioremap(BASEADDR + regmap[minor].baseaddr, count);
	        if (addr) {
			size_t i;
			char *p = data;
			for(i=0; i<count; i+=sizeof(long)) {
				long l = ioread32(addr+i);
				p += sprintf(p, " %ld", __arch__swab32(l));
			}
			p += sprintf(p, "\n");
			iounmap(addr);
			addr = NULL;
	        } else {
	                printk(KERN_ERR MODULE_NAME "-read: memory unmappable\n");
	                return -EIO;
		}
		break;
	}
	default:
		{
		printk(KERN_ALERT MODULE_NAME "-read[%d]: minor number out of range\n", minor);
		return -EFAULT;
		}
        }

	/* copy user space buffer to kernel space */
	err = copy_to_user(buff,data,strlen(data));

	if(err) {
		printk(KERN_ALERT MODULE_NAME "-read: userspace copy failed!\n" );
		return -EFAULT;
	}

        *offp += strlen(data);
	return strlen(data);
}

/** @brief write to register associated with the provided minor number */
ssize_t mcdr_write(struct file *filep,const char *buff,size_t count,loff_t *offp)
{
	int err;
	char data[PAGE_SIZE];
	unsigned int minor = iminor(filep->f_dentry->d_inode);

	printk(KERN_DEBUG MODULE_NAME "-write[%d]: count=%lu, offset=%lu\n",
				 minor, (unsigned long)count, (unsigned long)*offp );

	if(count > sizeof(data)) count = sizeof(data);

        memset(data, 0, sizeof(data));

	/* copy user space buffer to kernel space*/
	err = copy_from_user(data,buff,count);
	if(err) {
		printk(KERN_ALERT MODULE_NAME "-write: kernel copy failed!\n" );
		return -EFAULT;
	}

        switch(minor) {
	case  0:
	case  1:
	case  2:
	case  3:
	case  4:
	case  5:
	case  6:
	{
		char *eol = NULL;
		void *addr = NULL;
		long l[22];
		size_t i=0;
		i = sscanf(data, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld"
		                  "%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",
				&l[ 0], &l[ 1], &l[ 2], &l[ 3], &l[ 4], &l[ 5], &l[ 6], &l[ 7], &l[ 8], &l[ 9], &l[10], 
				&l[11], &l[12], &l[13], &l[14], &l[15], &l[16], &l[17], &l[18], &l[19], &l[20], &l[21]
				);
		if(i != 22) {
	                printk(KERN_ERR MODULE_NAME "-write: expected 22 tokens, got %d\n", i);
	                return -EIO;
		}

		eol = strchr(data, '\n');
		if(eol) count = 1+(eol-data);
	        addr = ioremap(BASEADDR + regmap[minor].baseaddr, regmap[minor].size);
		if (addr) {
			for(i=0; i<22; i++) {
				iowrite32(__arch__swab32(l[i]), addr+i*sizeof(long));
			}
			iounmap(addr);
			addr = NULL;
	        } else {
	                printk(KERN_ERR MODULE_NAME "-write: memory unmappable\n");
	                return -EIO;
		}
		break;
	}
	default:
		{
		printk(KERN_ALERT MODULE_NAME "-write[%d]: minor number out of range\n", minor);
		return -EFAULT;
		}
        }

	*offp += count;
	return count;
}

/** @brief initialize module */
static int mcdr_init(void)
{
	struct resource *r = NULL;

	printk(KERN_DEBUG MODULE_NAME ": initializing\n");
	mcdr_major = register_chrdev(0,MODULE_NAME,&mcdr_fops);
	if(!mcdr_major) {
		printk(KERN_ALERT MODULE_NAME ": failed to register\n");
		return -EIO;
	}
	printk(KERN_INFO MODULE_NAME ": major=%d\n", mcdr_major);
        r = request_mem_region( 
			BASEADDR + 0x0020, 0x0230-0x0020, "mcdr_init");
	if(!r) {
		printk(KERN_ALERT MODULE_NAME ": cannot access FPGA registers\n");
		return -EFAULT;
	}

	return 0;
}

/** @brief cleanup module */
static void mcdr_cleanup(void)
{
        release_mem_region(BASEADDR + 0x0020, 0x0230-0x0020);

	printk(KERN_DEBUG MODULE_NAME ": terminating\n");
	unregister_chrdev(mcdr_major,MODULE_NAME);
}

