/* Linux Image Engine Configuration Module */

#include <asm/current.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/byteorder.h>

#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/device.h>

#include <xilinx/xparameters.h>
#if 0
#define BASEADDR XPAR_RAMSTORE2PLB_BASEADDR
#else
#define BASEADDR XPAR_XPS_REGISTERINT_BRAM_BASEADDR
#endif

#define MODULE_NAME "mosscfg"
MODULE_AUTHOR("GVG");
MODULE_DESCRIPTION("ImageStore FPGA Configuration Driver");
MODULE_VERSION("0:0.1-experimental");

static struct resource *r = NULL;
static int major = 0; /* the major number assigned to our driver */
static int mcfg_init(void);
static void mcfg_cleanup(void);

module_init(mcfg_init);
module_exit(mcfg_cleanup);

int mcfg_open(struct inode *inode,struct file *filep);
int mcfg_release(struct inode *inode,struct file *filep);
ssize_t mcfg_read(struct file *filep,char *buff,size_t count,loff_t *offp );
ssize_t mcfg_write(struct file *filep,const char *buff,size_t count,loff_t *offp );
	
struct file_operations mcfg_fops={
	open: mcfg_open,
	read: mcfg_read,
	write: mcfg_write,
	release:mcfg_release,
};

int mcfg_open(struct inode *inode,struct file *filep)
{
	unsigned int minor = iminor(inode);
	printk(KERN_DEBUG MODULE_NAME ": open minor=%d\n", minor);
	return 0;
}

int mcfg_release(struct inode *inode,struct file *filep)
{
	unsigned int minor = iminor(inode);
	printk(KERN_DEBUG MODULE_NAME ": release minor=%d\n", minor);
	return 0;
}

ssize_t mcfg_read(struct file *filep,char *buff,size_t count,loff_t *offp )
{
	int err;
	char data[PAGE_SIZE];
	unsigned int minor = iminor(filep->f_dentry->d_inode);
	void *addr = NULL;

        printk(KERN_DEBUG MODULE_NAME ": read, minor=%d count=%d\n", minor, count);

	addr = ioremap(BASEADDR, sizeof(unsigned long));
	if(!addr) {
		printk(KERN_ALERT MODULE_NAME " -> io remap failed!\n" );
		return -EFAULT;
	}

	switch(minor) {
	case 1: /* MGT reset */
	case 2: /* DCM reset */
	case 3: /* DQ1 training */
	case 4: /* DQ2 training */
	case 5: /* RX1 training */
	case 6: /* RX2 training */
		strncpy(data,"0\n",sizeof(data));
		break;
	default: /* raw */
		{
		long l = __arch__swab32(ioread32(addr));
	       	snprintf(data, sizeof(data), 
				"CTL "
				"%lx %lx %lx %lx %lx %lx %lx %lx %lx %lx " 
				"%lx %lx %lx %lx %lx %lx %lx %lx %lx %lx " 
				"%lx  raw=0x%08lx \n", 
				((l >>20) & 1), ((l >>19) & 1),
				((l >>18) & 1), ((l >>17) & 1),
				((l >>16) & 1), ((l >>15) & 1),
				((l >>14) & 1), ((l >>13) & 1),
				((l >>12) & 1), ((l >>11) & 1),
				((l >>10) & 1), ((l >> 9) & 1),
				((l >> 8) & 1), ((l >> 7) & 1),
				((l >> 6) & 1), ((l >> 5) & 1),
				((l >> 4) & 1), ((l >> 3) & 1),
				((l >> 2) & 1), ((l >> 1) & 1),
				(l & 1), l
			);
		break;
		}
	}

	iounmap(addr);
	addr = NULL;

	/* copy kernel space buffer to user space*/
	err = copy_to_user(buff,data,strlen(data));
	if(err) {
		printk(KERN_ALERT MODULE_NAME " -> userspace copy failed!\n" );
		return -EFAULT;
	}

	*offp += strlen(data);
	return strlen(data);

}

ssize_t mcfg_write(struct file *filep,const char *buff,size_t count,loff_t *offp )
{
	int err;
	char data[PAGE_SIZE];
        unsigned int minor = iminor(filep->f_dentry->d_inode);
	void *addr = NULL;

        printk(KERN_DEBUG MODULE_NAME ": write, minor=%d count=%d\n", minor, count);

        memset(data, 0, sizeof(data));

        /* copy user space buffer to kernel space */
        err = copy_from_user(data,buff,sizeof(data));
	if(err) {
                printk(KERN_ALERT MODULE_NAME " -> Kernel copy failed!\n" );
                return -EFAULT;
        }

        /* one line at a time */
	{
	char *p = strnchr(data, sizeof(data), '\n');
	if(!p) {
		printk(KERN_ALERT MODULE_NAME " -> line detect failed!\n" );
		return -EFAULT;
	}
	if((p-data) > PAGE_SIZE) {
		printk(KERN_ALERT MODULE_NAME " -> line too long!\n" );
		return -EFAULT;
	}
	*(p+1) = '\0';
	}

	addr = ioremap(BASEADDR, sizeof(unsigned long));
	if(!addr) {
		printk(KERN_ALERT MODULE_NAME " -> io remap failed!\n" );
		return -EFAULT;
	}

	switch(minor) {
	case 1: /* tap the MGT reset */
		{
		long ctl  = __arch__swab32(ioread32(addr));
		set_bit(14, &ctl);
		iowrite32(__arch__swab32(ctl), addr);
		barrier();
		clear_bit(14, &ctl);
		iowrite32(__arch__swab32(ctl), addr);
		break;
		}
	case 2: /* tap the DCM reset */
		{
		long ctl  = __arch__swab32(ioread32(addr));
		set_bit(13, &ctl);
		iowrite32(__arch__swab32(ctl), addr);
		barrier();
		clear_bit(13, &ctl);
		iowrite32(__arch__swab32(ctl), addr);
		break;
		}
	case 3: /* tap the Video Controller reset */
		{
		long ctl  = __arch__swab32(ioread32(addr));
		set_bit(11, &ctl);
		iowrite32(__arch__swab32(ctl), addr);
		barrier();
		clear_bit(11, &ctl);
		iowrite32(__arch__swab32(ctl), addr);
		break;
		}
	case 4: /* tap the Video Controller Active */
		{
		long ctl  = __arch__swab32(ioread32(addr));
		set_bit(7, &ctl);
		iowrite32(__arch__swab32(ctl), addr);
		barrier();
		clear_bit(7, &ctl);
		iowrite32(__arch__swab32(ctl), addr);
		break;
		}
	case 5: /* RX1 training */
	case 6: /* RX2 training */
		break;
	default: /* raw */
		{
		long l = 0;
		if(sscanf(data, "CTL raw=%lx", &l) == 1 ) 
			iowrite32(__arch__swab32(l), addr);
		else
			printk(KERN_ALERT MODULE_NAME ": input parsing  failed!\n" );
		break;
		}
	}

	iounmap(addr);
	addr = NULL;

	*offp += strlen(data);
	return strlen(data);
}

static int mcfg_init(void)
{
	printk(KERN_DEBUG MODULE_NAME ": initializing\n");
	major = register_chrdev(0,MODULE_NAME,&mcfg_fops);
	if(!major){
		printk(KERN_ALERT MODULE_NAME ": failed to register\n");
		return -EIO;
	}
	printk(KERN_INFO MODULE_NAME ": major=%d\n", major);
        r = request_mem_region(
				BASEADDR, 
				sizeof(long),
				MODULE_NAME
				);
	if(!r) {
        	printk(KERN_ALERT MODULE_NAME ": cannot access FPGA registers\n");
		return -EFAULT;
	}
	return 0;
}
static void mcfg_cleanup(void)
{
        if(r) release_mem_region(BASEADDR, sizeof(long));

	printk(KERN_DEBUG MODULE_NAME ": terminating\n");
	unregister_chrdev(major,MODULE_NAME);
	return ;
}

