// compile with : g++ -g -o thumbnail.cgi thumbnail.cpp -lrudecgi
#include <fstream>
#include <iomanip>
#include <sstream>
#include <stdexcept>

#include <boost/lexical_cast.hpp>
#include <boost/interprocess/sync/scoped_lock.hpp>
#include <boost/interprocess/sync/named_mutex.hpp>

#include <rude/cgi.h>

#include <log4cpp/Category.hh>
#include <log4cpp/PropertyConfigurator.hh>

#include "blkaddr.h"
#include "400.h"
#include "404.h"
#include "500.h"

namespace l4 = log4cpp;

static const size_t BLKSIZ = 112 * sizeof(long);
static boost::interprocess::named_mutex mutex(boost::interprocess::open_or_create, "thumbnail");

/// \brief BMP file header
/// \note integer values are stored in little-endian format (least-significant byte first)
static const char bmphdr[14] = {
	'B', 'M',		///< the magic number used to identify the BMP file: 0x42 0x4D (ASCII code points for B and M) 
	0x76, 0x2F, 0, 0, 	///< the size of the BMP file in bytes: 54+84*48*3=12150=0x2F76
 	'G', 'V', 'M', 'O', 	///< application identifier: Grass Valley Missouri
	 54, 0, 0, 0, 		///< the offset, i.e. starting address, of the byte where the bitmap data can be found. 
};

/// \brief Device Independent Bitmap (DIB) Header
/// \note uncompressed Windows V3 version for maximum compatibility
static const char dibhdr[40] = {
	 40, 0, 0, 0,  	///< size of this header (40 bytes) 
	 84, 0, 0, 0,  	///< bitmap width in pixels (signed integer). 
	 48, 0, 0, 0,  	///< bitmap height in pixels (signed integer). 
	  1, 0,  	///< number of color planes being used. Must be set to 1. 
	 24, 0,  	///< number of bits per pixel, which is the color depth of the image. Typical values are 1, 4, 8, 16, 24 and 32. 
	  0, 0, 0, 0,  	///< compression method being used. 0 for none.
	0x40, 0x2F, 0, 0,  	///< image size. 84*48*3=12096=0x2F40
	0x61, 0x0F, 0, 0,  	///< horizontal resolution of the image. (pixel per meter, signed integer) 
	0x61, 0x0F, 0, 0,  	///< vertical resolution of the image. (pixel per meter, signed integer) 
	  0, 0, 0, 0,  	///< number of colors in the color palette, or 0 to default to 2n. 
	  0, 0, 0, 0,  	///< number of important colors used, or 0 when every color is important; generally ignored. 
};

/// \brief converts 4 ASCII chars into a 24bit (8x3) RGB tuple
void RGB_PIXEL(const char *data, char *pixel)
{
	pixel[0] = ((((data[2] - 33) & 0x3) << 6) | ((data[3] - 33))); 
	pixel[1] = ((((data[1] - 33) & 0xF) << 4) | ((data[2] - 33) >> 2)); 
	pixel[2] = (((data[0] - 33) << 2) | ((data[1] - 33) >> 4)); 

	if(pixel[0]<16) pixel[2] = 16;
	if(pixel[1]<16) pixel[2] = 16;
	if(pixel[2]<16) pixel[2] = 16;
}

/// \brief prints out a CGI 400:BadRequest responce with an associated image.
/// 84x48x24bit RGB DIB bitmap 
void error400(void)
{
	char pixels[84*48*3];
	char *ptr = pixels+sizeof(pixels);

	while(ptr>pixels){
		ptr -= 3;
		RGB_PIXEL(data400,ptr); 
		data400 += 4; 
	}

	// Print out CGI HTTP Response header
	std::cout << "HTTP/1.0 400 Bad Request\n";
	std::cout << "Cache-Control: no-cache\n";
	std::cout << "Content-Type: image/bmp; name=\"NotFound.bmp\"\n";
	std::cout << "Content-Length: 12150\n";
	std::cout << "Server: Missouri-ImageCache/0.1\n";
	std::cout << "\n";

	// Print out image
	std::cout.write(bmphdr, sizeof(bmphdr));
	std::cout.write(dibhdr, sizeof(dibhdr));
	std::cout.write(pixels, sizeof(pixels));
}

/// \brief prints out a CGI 404:NotFound responce with an associated image.
/// 84x48x24bit RGB DIB bitmap 
void error404(void)
{
	char pixels[84*48*3];
	char *ptr = pixels+sizeof(pixels);

	while(ptr>pixels){
		ptr -= 3;
		RGB_PIXEL(data404,ptr); 
		data404 += 4; 
	}

	// Print out CGI HTTP Response header
	std::cout << "HTTP/1.0 404 Not Found\n";
	std::cout << "Cache-Control: no-cache\n";
	std::cout << "Content-Type: image/bmp; name=\"NotFound.bmp\"\n";
	std::cout << "Content-Length: 12150\n";
	std::cout << "Server: Missouri-ImageCache/0.1\n";
	std::cout << "\n";

	// Print out image
	std::cout.write(bmphdr, sizeof(bmphdr));
	std::cout.write(dibhdr, sizeof(dibhdr));
	std::cout.write(pixels, sizeof(pixels));
}

/// \brief prints out a CGI 500:InternalError responce with an associated image.
/// 84x48x24bit RGB DIB bitmap 
void error500(void)
{
	char pixels[84*48*3];
	char *ptr = pixels+sizeof(pixels);

	while(ptr>pixels){
		ptr -= 3;
		RGB_PIXEL(data500,ptr); 
		data500 += 4; 
	}

	// Print out CGI HTTP Response header
	std::cout << "HTTP/1.0 500 Internal Error\n";
	std::cout << "Cache-Control: no-cache\n";
	std::cout << "Content-Type: image/bmp; name=\"InternalError.bmp\"\n";
	std::cout << "Content-Length: 12150\n";
	std::cout << "Server: Missouri-ImageCache/0.1\n";
	std::cout << "\n";

	// Print out image
	std::cout.write(bmphdr, sizeof(bmphdr));
	std::cout.write(dibhdr, sizeof(dibhdr));
	std::cout.write(pixels, sizeof(pixels));
}

/// \brief extracts an 8bit red value from a 10bit ycbcr data stream
/// \param pos points to a 10bit luminance value, which is preceeded by a Cr value
/// \returns the derived an 8bit (0-255) red value
/// \note this is an approximate (CPU conserving) integer math based implementation,
/// suitable for use in generating thumbnails, but not suitable for use in essence conversion
unsigned char ycbcr2r(char *pos)
{
	long y = *(unsigned short *)pos;
	y /= 4;
	if(y < 0) y=0;
	if(y > 255) y=255;

	long cr = *(unsigned short *)(pos-sizeof(short));
	cr /= 4;
	if(cr < 0) cr=0;
	if(cr > 255) cr=255;

	long r = ( 298 * y + 408 * cr ) / 256 - 222;
	if(r < 16) return 16;
	if(r > 255) return 255;
	return r;
}

/// \brief extracts an 8bit green value from a 10bit ycbcr data stream
/// \param pos points to a 10bit luminance value, which is preceeded by a Cr value, and followed by a Cb value
/// \returns the derived an 8bit (0-255) green value
/// \note this is an approximate (CPU conserving) integer math based implementation,
/// suitable for use in generating thumbnails, but not suitable for use in essence conversion
unsigned char ycbcr2g(char *pos)
{
	long y = *(unsigned short *)pos;
	y /= 4;
	if(y < 0) y = 0;
	if(y > 255) y = 255;

	long cb = *(unsigned short *)(pos+sizeof(short));
	cb /= 4;
	if(cb < 0) cb=0;
	if(cb > 255) cb=255;

	long cr = *(unsigned short *)(pos-sizeof(short));
	cr /= 4;
	if(cr < 0) cr=0;
	if(cr > 255) cr=255;

	long g = ( 298 * y - 100 * cb - 208 * cr ) / 256 + 135;
	if(g < 16) return 16;
	if(g > 255) return 255;
	return g;
}

/// \brief extracts an 8bit blue value from a 10bit ycbcr data stream
/// \param pos points to a 10bit luminance value, which is followed by a Cb value
/// \returns the derived an 8bit (0-255) blue value
/// \note this is an approximate (CPU conserving) integer math based implementation,
/// suitable for use in generating thumbnails, but not suitable for use in essence conversion
unsigned char ycbcr2b(char *pos)
{
	short y = *(unsigned short *)pos;
	y /= 4;
	if(y < 0) y=0;
	if(y > 255) y=255;

	short cb = *(unsigned short *)(pos+sizeof(short));
	cb /= 4;
	if(cb < 0) cb=0;
	if(cb > 255) cb=255;

	long b = ( 298 * y + 516 * cb ) / 256 - 276;
	if(b < 16) return 16;
	if(b > 255) return 255;
	return b;
}

/// \brief prints out a CGI 200:OK responce with an associated image.
/// 84x48x24bit RGB DIB bitmap 
void thumbnail(const size_t hBias, const size_t hGain, const size_t vBias, const size_t vGain, const char *device, loff_t offset, const size_t len)
{
	char pixels[48*84*3];
	char buf[len];

        std::ifstream cm(device); // cache memory stream

	if(cm.is_open()) {
		memset(pixels, 0, sizeof(pixels));
		for(size_t y=0; y<48; y++) {
			cm.seekg(offset+vGain*(y+vBias)*len);
			cm.read(buf, len);
			for(size_t x=0; x<84; x++) {
				size_t index = 3*(x+(47-y)*84);
				char *pos = buf+(hBias+x*hGain);
				pixels[index+0] = ycbcr2b(pos);
				pixels[index+1] = ycbcr2g(pos);
				pixels[index+2] = ycbcr2r(pos);
			}
		}
		cm.close();
	}

	// Print out CGI HTTP Response header
	std::cout << "HTTP/1.0 200 OK\n";
	std::cout << "Content-Type: image/bmp; name=\"thumbnail.bmp\"\n";
	std::cout << "Content-Length: 12150\n";
	std::cout << "Server: Missouri-ImageCache/0.1\n";
	std::cout << "\n";

	// Print out image
	std::cout.write(bmphdr, sizeof(bmphdr));
	std::cout.write(dibhdr, sizeof(dibhdr));
	std::cout.write(pixels, sizeof(pixels));
}

/// \brief lookup item address in the cache
/// \param addr block address of item
/// \returns address of item in cache
loff_t lookup(const std::string blkaddr)
{
	
	loff_t index;
	try  {
		index = boost::lexical_cast<loff_t>(blkaddr);
		index *= BLKSIZ;
	} catch(...) {
		index = ~0LL;
	}
	return index;
}

/// \brief entry point or the application
int main(int argc,const char *argv[])
{
	// Obtain the CGI instance
	rude::CGI cgi;

        l4::Category &log = l4::Category::getRoot();
	try {
		l4::PropertyConfigurator::configure(".l4rc");
		log.noticeStream() << "logging configured" << l4::eol;
	} catch(log4cpp::ConfigureFailure e) {
		log.noticeStream() << "configuration failed" << l4::eol;
	}

	// check for required search parameters: suite, id, frame
	if(!cgi.exists("suite") || !cgi.exists("id") || !cgi.exists("field")) {
		error400();
		l4::Category::shutdown();
		return 0;
	}

	try {
		blkaddr ba;

		boost::interprocess::scoped_lock<boost::interprocess::named_mutex> lock(mutex);

		// lookup item address in the cache
		unsigned suite = boost::lexical_cast<unsigned >(cgi.value("suite"));
		unsigned id = boost::lexical_cast<unsigned >(cgi.value("id"));
		unsigned field = boost::lexical_cast<unsigned >(cgi.value("field"));
		if(!suite) throw std::out_of_range("suite=0");
		if(suite>2) throw std::out_of_range("suite>2");
		if(!id) throw std::out_of_range("id=0");
		if(id>9999) throw std::out_of_range("id>9999");
		if(!field) throw std::out_of_range("field=0");
		if(field>9999) throw std::out_of_range("field>9999");
		std::ostringstream buf;
		buf << suite << ":";
		buf << std::setfill('0') << std::setw(4) << id << ":";
		buf << std::setfill('0') << std::setw(4) << field << ":";
		buf << "v";
		loff_t addr = ba.addr(buf.str());
		addr *= BLKSIZ;

		// print out thumbnail
		switch(ba.used(buf.str())) {
		case 1701:
		case 1708:
			thumbnail(8, 32, 4, 4, "/dev/moCM2", addr, 7*BLKSIZ);
			break;
		case 2016:
			thumbnail(8, 32, 5, 5, "/dev/moCM2", addr, 7*BLKSIZ);
			break;
		case 8640:
			thumbnail(16, 56, 6, 12, "/dev/moCM2", addr, 12*BLKSIZ);
			break;
		case 9720:
			thumbnail(40, 88, 5, 10, "/dev/moCM2", addr, 18*BLKSIZ);
			break;
		default:
			throw std::runtime_error("unknown image format");
			break;
		}
	} catch(std::runtime_error e) {
		error404();
	} catch(...) {
		error500();
	}

	l4::Category::shutdown();
	return 0;
}
