/* Linux Image Engine Configuration Module */

#include <asm/current.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/byteorder.h>

#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/wait.h>
#include <linux/kernel.h>
#include <linux/device.h>

#include <xilinx/xparameters.h>
#if 0
#define RS_DQS_IDLY_CTL (XPAR_RAMSTORE2PLB_BASEADDR + 0x04)
#define RS_DQ_IDLY_CTL (XPAR_RAMSTORE2PLB_BASEADDR + 0x08)
#define RS_DQS_RB (XPAR_RAMSTORE2PLB_BASEADDR + 0x408)
#else
#define RS_DQS_IDLY_CTL (XPAR_XPS_REGISTERINT_BRAM_BASEADDR + 0x04)
#define RS_DQ_IDLY_CTL (XPAR_XPS_REGISTERINT_BRAM_BASEADDR + 0x08)
#define RS_DQS_RB (XPAR_XPS_REGISTERINT_BRAM_BASEADDR + 0x408)
#endif
#define DQS_IDLY_RST (1<<31)
#define DQS_IDLY_INC (1<<30)
#define DQ_IDLY_RST (1<<31)
#define DQ_IDLY_INC (1<<30)
#define LINELEN (7 + 18*2 + 1)

#define MODULE_NAME "mossdq"
MODULE_AUTHOR("GVG");
MODULE_DESCRIPTION("ImageStore DQ Calibration Driver");
MODULE_VERSION("0:0.1-experimental");

static struct resource *r = NULL;
static int major = 0; /* the major number assigned to our driver */
static int mdq_init(void);
static void mdq_cleanup(void);

module_init(mdq_init);
module_exit(mdq_cleanup);

int mdq_open(struct inode *inode,struct file *filep);
int mdq_release(struct inode *inode,struct file *filep);
ssize_t mdq_read(struct file *filep,char *buff,size_t count,loff_t *offp );
ssize_t mdq_write(struct file *filep,const char *buff,size_t count,loff_t *offp );
	
struct file_operations mdq_fops={
	open: mdq_open,
	read: mdq_read,
	write: mdq_write,
	release:mdq_release,
};

int mdq_open(struct inode *inode,struct file *filep)
{
	unsigned int minor = iminor(inode);
	printk(KERN_DEBUG MODULE_NAME ": open minor=%d\n", minor);
	return 0;
}

int mdq_release(struct inode *inode,struct file *filep)
{
	unsigned int minor = iminor(inode);
	printk(KERN_DEBUG MODULE_NAME ": release minor=%d\n", minor);
	return 0;
}

ssize_t mdq_read(struct file *filep,char *buff,size_t count,loff_t *offp )
{
	int err;
	char data[PAGE_SIZE];
	unsigned int minor = iminor(filep->f_dentry->d_inode);
	unsigned rb = 0;
	void *sctl = ioremap(RS_DQS_IDLY_CTL, sizeof(unsigned long));
	void *addr = ioremap(RS_DQS_RB, sizeof(unsigned long));
	size_t tap = ((*offp) & 0xFFFFFFFF);

        printk(KERN_DEBUG MODULE_NAME ": read, minor=%d offp=%d\n", minor, tap);
	tap /= LINELEN;
	tap %= 64;

	if(sctl) {
		if(!tap) {
			iowrite32(__arch__swab32(DQS_IDLY_RST), sctl);
			barrier();

			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(HZ/1000);
			iowrite32(__arch__swab32(0), sctl);
			barrier();

			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(HZ/1000);
		}

		iowrite32(__arch__swab32(DQS_IDLY_INC | 0x3FFFF), sctl);
		barrier();

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ/1000);
		iowrite32(__arch__swab32(DQS_IDLY_INC), sctl);
		barrier();

		iounmap(sctl);
	}

	if(addr) {
		rb = __arch__swab32(ioread32(addr));
		iounmap(addr);
	}
       	snprintf(data, sizeof(data), 
			"DQS:%02d "
			"%c %c %c %c %c %c %c %c %c " 
			"%c %c %c %c %c %c %c %c %c " 
			"\n", 
			tap,
			(rb & (1<< 0) ? '1' : '0'),
			(rb & (1<< 1) ? '1' : '0'),
			(rb & (1<< 2) ? '1' : '0'),
			(rb & (1<< 3) ? '1' : '0'),
			(rb & (1<< 4) ? '1' : '0'),
			(rb & (1<< 5) ? '1' : '0'),
			(rb & (1<< 6) ? '1' : '0'),
			(rb & (1<< 7) ? '1' : '0'),
			(rb & (1<< 8) ? '1' : '0'),
			(rb & (1<< 9) ? '1' : '0'),
			(rb & (1<<10) ? '1' : '0'),
			(rb & (1<<11) ? '1' : '0'),
			(rb & (1<<12) ? '1' : '0'),
			(rb & (1<<13) ? '1' : '0'),
			(rb & (1<<14) ? '1' : '0'),
			(rb & (1<<15) ? '1' : '0'),
			(rb & (1<<16) ? '1' : '0'),
			(rb & (1<<17) ? '1' : '0')
		);

	/* copy kernel space buffer to user space*/
	err = copy_to_user(buff,data,strlen(data));
	if(err) {
		printk(KERN_ALERT MODULE_NAME " -> userspace copy failed!\n" );
		return -EFAULT;
	}

	*offp += strlen(data);
	return strlen(data);

}

ssize_t mdq_write(struct file *filep,const char *buff,size_t count,loff_t *offp )
{
	size_t i;
	int err;
	char data[PAGE_SIZE];
	char buf[18];
	int mask;
        unsigned int minor = iminor(filep->f_dentry->d_inode);
	void *sctl = NULL;
	void *ctl = NULL;
	size_t low = ((*offp) & 0xFFFFFFFF);

        printk(KERN_DEBUG MODULE_NAME ": write, minor=%d offp=%d\n", minor, low);

        memset(data, 0, sizeof(data));

        /* copy user space buffer to kernel space*/
        err = copy_from_user(data,buff,sizeof(data));
	if(err) {
                printk(KERN_ALERT MODULE_NAME " -> Kernel copy failed!\n" );
                return -EFAULT;
        }

	/* TODO: extract the mask from the user space data */
	mask = 0;
	i = sscanf(data, "%c %c %c %c %c %c %c %c %c %c %c %c %c %c %c %c %c %c",
			&buf[ 0], &buf[ 1], &buf[ 2], &buf[ 3], &buf[ 4], &buf[ 5], 
			&buf[ 6], &buf[ 7], &buf[ 8], &buf[ 9], &buf[10], &buf[11], 
			&buf[12], &buf[13], &buf[14], &buf[15], &buf[16], &buf[17] 
		);
	if(i!= sizeof(buf)) {
		printk(KERN_ERR MODULE_NAME "-write: expected %d tokens, got %d\n", sizeof(buf), i);
		return -EIO;
	}
	for( i=0; i<sizeof(buf); i++ ) {
		if(buf[i]=='1') mask += (1<<i);
	}
 	mask &= 0x3FFFF;

	sctl = ioremap(RS_DQS_IDLY_CTL, sizeof(unsigned long));
	if(!sctl) {
                printk(KERN_ALERT MODULE_NAME " -> IR_DQS_IDLY_CTL remap failed!\n" );
                return -EFAULT;
        }
	ctl = ioremap(RS_DQ_IDLY_CTL, sizeof(unsigned long));
	if(!sctl) {
                printk(KERN_ALERT MODULE_NAME " -> IR_DQ_IDLY_CTL remap failed!\n" );
                return -EFAULT;
        }

	if(!low) {
		iowrite32(__arch__swab32(DQS_IDLY_RST), sctl);
		iowrite32(__arch__swab32(DQS_IDLY_RST), ctl);
		barrier();

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ/1000);
		iowrite32(__arch__swab32(0), sctl);
		iowrite32(__arch__swab32(0), ctl);
		barrier();

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ/1000);
	}

	iowrite32(__arch__swab32(DQS_IDLY_INC | mask), sctl);
	iowrite32(__arch__swab32(DQS_IDLY_INC | mask), ctl);
	barrier();

	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(HZ/1000);
	iowrite32(__arch__swab32(DQS_IDLY_INC), sctl);
	iowrite32(__arch__swab32(DQS_IDLY_INC), ctl);
	barrier();

	iounmap(sctl);
	iounmap(ctl);

	*offp += strlen(data);
	return strlen(data);
}

static int mdq_init(void)
{
	printk(KERN_DEBUG MODULE_NAME ": initializing\n");
	major = register_chrdev(0,MODULE_NAME,&mdq_fops);
	if(!major){
		printk(KERN_ALERT MODULE_NAME ": failed to register\n");
		return -EIO;
	}
	printk(KERN_INFO MODULE_NAME ": major=%d\n", major);
        r = request_mem_region( RS_DQS_RB, sizeof(long), MODULE_NAME);
	if(!r) {
        	printk(KERN_ALERT MODULE_NAME ": cannot access FPGA registers\n");
		return -EFAULT;
	}
        r = request_mem_region( RS_DQS_IDLY_CTL, sizeof(long), MODULE_NAME);
	if(!r) {
        	printk(KERN_ALERT MODULE_NAME ": cannot access FPGA registers\n");
		return -EFAULT;
	}
	return 0;
}
static void mdq_cleanup(void)
{
        if(r) {
		release_mem_region(RS_DQS_RB, sizeof(long));
		release_mem_region(RS_DQS_IDLY_CTL, sizeof(long));
	}

	printk(KERN_DEBUG MODULE_NAME ": terminating\n");
	unregister_chrdev(major,MODULE_NAME);
	return ;
}

