/// \file
/// \copyright 2008 Thomson Grass Valley
/// \author adam clarke
/// \date 2008-06-06
/// \brief respond to host controller messages 
///
#include <iomanip>
#include <iostream>
#include <sstream>
#include <vector>
#include <map>
#include <algorithm>
#include <stdexcept>
#include <ctime>

#include <boost/lexical_cast.hpp>
#include <boost/program_options.hpp>

#include <log4cpp/Category.hh>
#include <log4cpp/BasicConfigurator.hh>
#include <log4cpp/PropertyConfigurator.hh>

#include "blkaddr.h"

namespace po = boost::program_options;
namespace l4 = log4cpp;

class range
{
	unsigned upper;
	unsigned lower;
	bool valid;
	l4::Category &log;
public:
	range(const std::string &spec);
	bool overlap(const range &r);
	bool operator!(void) const { return !valid; };
	bool operator==(const range &r) const
		{ return (valid==r.valid) && (upper==r.upper) && (lower==r.lower); };

	unsigned getUpper() const { return upper; };
	unsigned getLower() const { return lower; };
};

range::range(const std::string &spec) 
	: upper(0), lower(0), valid(false), log(l4::Category::getInstance("range"))
{
	// validate input 
	if(spec.find_first_not_of("1234567890-") != std::string::npos)
	{
		return;
	}

	// parse first parameter 
	std::istringstream buf(spec);
	short x = 0;
	short y = 0;
	buf >> x;
	if(buf.good() || buf.eof())
	{
		upper = lower = x;
		valid = true;
	}

	if(buf.good()) {
		// ignore dash delimiter 
		buf.ignore(1);
		
		// parse second parameter 
		buf >> y;
		if(buf.good() || buf.eof())
		{
			lower = std::min(x,y);
			upper = std::max(x,y);
			valid = true;
		}
	}
}

bool range::overlap(const range &r) 
{
	if((r.lower > lower) && (r.lower < upper)) return true;
	return false;
}

class move
{
	l4::Category &log;
public:
	move() : log(l4::Category::getInstance("move")) {};

	void operator()(const std::string cmd)
	{
		unsigned suite = 0;
		size_t suite1 = std::string::npos;
		size_t suite2 = std::string::npos;
		size_t s_begin = std::string::npos;
		size_t s_end = std::string::npos;
		size_t t_begin = std::string::npos;
		size_t t_end = std::string::npos;

		log.infoStream() << "got: " << cmd << l4::eol;

		suite1 = cmd.find("SUITE=1");
		suite2 = cmd.find("SUITE=2");
		if((suite1 == std::string::npos) && (suite2 == std::string::npos)) {
			log.warnStream() << "no SUITE tag" << l4::eol;
			return;
		}
		if(suite1 != std::string::npos) suite = 1;
		if(suite2 != std::string::npos) suite = 2;

		s_begin = cmd.find("SOURCE=");
		if(s_begin == std::string::npos) {
			log.warnStream() << "no SOURCE tag" << l4::eol;
			return;
		}
		s_begin += strlen("SOURCE=");

		s_end = cmd.find_first_not_of("0123456789-", s_begin);
		if(s_end == std::string::npos) s_end = cmd.size();

		range srcs(cmd.substr(s_begin, s_end-s_begin));
		if(!srcs) {
			log.warnStream() << "no SOURCE values" << l4::eol;
			return;
		}

		t_begin = cmd.find("TARGET=");
		if(t_begin == std::string::npos) {
			log.warnStream() << "no TARGET tag" << l4::eol;
			return;
		}
		t_begin += strlen("TARGET=");

		t_end = cmd.find_first_not_of("0123456789-", t_begin);
		if(t_end == std::string::npos) t_end = cmd.size();

		range targets(cmd.substr(t_begin, t_end-t_begin));
		if(!targets) {
			log.warnStream() << "no TARGET values" << l4::eol;
			return;
		}

		if(srcs.overlap(targets)) {
			unsigned src = srcs.getUpper();
			unsigned target = targets.getUpper();
			blkaddr ba;
			while((src>=srcs.getLower()) && (target>=targets.getLower())) {
				log.infoStream() << "got: " << src << " -> " << target << l4::eol;
				ba.move(suite, src, target);
				--src; --target;
			}
		} else {
			unsigned src = srcs.getLower();
			unsigned target = targets.getLower();
			blkaddr ba;
			while((src<=srcs.getUpper()) && (target<=targets.getUpper())) {
				log.infoStream() << "got: " << src << " -> " << target << l4::eol;
				ba.move(suite, src, target);
				++src; ++target;
			}
		}
	};
	void operator() (const unsigned suite, const unsigned src, const unsigned target)
	{
		log.infoStream() << "got: " << src << " -> " << target << l4::eol;
		blkaddr ba;
		if(src && target) {
			if(!ba.move(suite, src, target))
				log.infoStream() << "failed" << l4::eol;
		}
	};
};

class swap
{
	l4::Category &log;
public:
	swap() : log(l4::Category::getInstance("swap")) {};

	void operator()(const std::string cmd)
	{
		unsigned suite = 0;
		size_t suite1 = std::string::npos;
		size_t suite2 = std::string::npos;
		size_t s_begin = std::string::npos;
		size_t s_end = std::string::npos;
		size_t t_begin = std::string::npos;
		size_t t_end = std::string::npos;

		log.infoStream() << "got: " << cmd << l4::eol;

		suite1 = cmd.find("SUITE=1");
		suite2 = cmd.find("SUITE=2");
		if((suite1 == std::string::npos) && (suite2 == std::string::npos)) {
			log.warnStream() << "no SUITE tag" << l4::eol;
			return;
		}
		if(suite1 != std::string::npos) suite = 1;
		if(suite2 != std::string::npos) suite = 2;

		s_begin = cmd.find("SOURCE=");
		if(s_begin == std::string::npos) {
			log.warnStream() << "no SOURCE tag" << l4::eol;
			return;
		}
		s_begin += strlen("SOURCE=");

		s_end = cmd.find_first_not_of("0123456789-", s_begin);
		if(s_end == std::string::npos) s_end = cmd.size();

		range srcs(cmd.substr(s_begin, s_end-s_begin));
		if(!srcs) {
			log.warnStream() << "no SOURCE values" << l4::eol;
			return;
		}

		t_begin = cmd.find("TARGET=");
		if(t_begin == std::string::npos) {
			log.warnStream() << "no TARGET tag" << l4::eol;
			return;
		}
		t_begin += strlen("TARGET=");

		t_end = cmd.find_first_not_of("0123456789-", t_begin);
		if(t_end == std::string::npos) t_end = cmd.size();

		range targets(cmd.substr(t_begin, t_end-t_begin));
		if(!targets) {
			log.warnStream() << "no TARGET values" << l4::eol;
			return;
		}

		unsigned src = srcs.getLower();
		unsigned target = targets.getLower();
		blkaddr ba;
		while((src<=srcs.getUpper()) && (target<=targets.getUpper())) {
			log.infoStream() << "got: " << src << " -> " << target << l4::eol;
			if(!ba.swap(suite, src, target)) {
				if(!ba.move(suite,src,target)) {
					ba.move(suite,target,src);
				}
			}
			src++; target++;
		}
	};
	void operator() (const unsigned suite, const unsigned src, const unsigned target)
	{
		blkaddr ba;
		log.infoStream() << "got: " << src << " -> " << target << l4::eol;
		if(src && target) {
			if(!ba.swap(suite, src, target)) {
				if(!ba.move(suite,src,target)) {
					ba.move(suite,target,src);
				}
			}
		}
	};
};

class append
{
	l4::Category &log;
public:
	append() : log(l4::Category::getInstance("append")) {};

	void operator()(const std::string cmd)
	{
		unsigned suite = 0;
		size_t suite1 = std::string::npos;
		size_t suite2 = std::string::npos;
		size_t s_begin = std::string::npos;
		size_t s_end = std::string::npos;
		size_t t_begin = std::string::npos;
		size_t t_end = std::string::npos;

		log.infoStream() << "got: " << cmd << l4::eol;

		suite1 = cmd.find("SUITE=1");
		suite2 = cmd.find("SUITE=2");
		if((suite1 == std::string::npos) && (suite2 == std::string::npos)) {
			log.warnStream() << "no SUITE tag" << l4::eol;
			return;
		}
		if(suite1 != std::string::npos) suite = 1;
		if(suite2 != std::string::npos) suite = 2;

		s_begin = cmd.find("SOURCE=");
		if(s_begin == std::string::npos) return;
		s_begin += strlen("SOURCE=");

		s_end = cmd.find_first_not_of("0123456789", s_begin);
		if(s_end == std::string::npos) s_end = cmd.size();

		unsigned src = 0;
		if(s_end > s_begin) {
			src = boost::lexical_cast<unsigned>(cmd.substr(s_begin, s_end-s_begin));
		}

		t_begin = cmd.find("TARGET=");
		if(t_begin == std::string::npos) return;
		t_begin += strlen("TARGET=");

		t_end = cmd.find_first_not_of("0123456789", t_begin);
		if(t_end == std::string::npos) t_end = cmd.size();

		unsigned target = 0;
		if(t_end > t_begin) {
			target = boost::lexical_cast<unsigned>(cmd.substr(t_begin, t_end-t_begin));
		}

		blkaddr ba;
		log.infoStream() << "got: " << src << " -> " << target << l4::eol;
		if(src && target) {
			ba.append(suite, src, target);
		}
	};
	void operator() (const unsigned suite, const unsigned src, const unsigned target)
	{
		blkaddr ba;
		log.infoStream() << "got: " << src << " -> " << target << l4::eol;
		if(src && target) {
			ba.append(suite, src, target);
		}
	};
};

class head
{
	l4::Category &log;
public:
	head() : log(l4::Category::getInstance("head")) {};

	void operator()(const std::string cmd)
	{
		unsigned suite = 0;
		size_t suite1 = std::string::npos;
		size_t suite2 = std::string::npos;
		size_t s_begin = std::string::npos;
		size_t s_end = std::string::npos;
		size_t m_begin = std::string::npos;
		size_t m_end = std::string::npos;
		size_t t_begin = std::string::npos;
		size_t t_end = std::string::npos;

		log.infoStream() << "got: " << cmd << l4::eol;

		suite1 = cmd.find("SUITE=1");
		suite2 = cmd.find("SUITE=2");
		if((suite1 == std::string::npos) && (suite2 == std::string::npos)) {
			log.infoStream() << "no SUITE tag" << l4::eol;
			return;
		}
		if(suite1 != std::string::npos) suite = 1;
		if(suite2 != std::string::npos) suite = 2;

		s_begin = cmd.find("SRC=");
		if(s_begin == std::string::npos) return;
		s_begin += strlen("SRC=");

		s_end = cmd.find_first_not_of("0123456789", s_begin);
		if(s_end == std::string::npos) s_end = cmd.size();

		unsigned src = 0;
		if(s_end > s_begin) {
			src = boost::lexical_cast<unsigned>(cmd.substr(s_begin, s_end-s_begin));
		}

		m_begin = cmd.find("MARK=");
		if(m_begin == std::string::npos) return;
		m_begin += strlen("MARK=");

		m_end = cmd.find_first_not_of("0123456789", m_begin);
		if(m_end == std::string::npos) m_end = cmd.size();

		unsigned mark = 0;
		if(m_end > m_begin) {
			mark = boost::lexical_cast<unsigned>(cmd.substr(m_begin, m_end-m_begin));
		}

		unsigned target = 0;
		t_begin = cmd.find("TARGET=");
		if(t_begin != std::string::npos) {
			t_begin += strlen("TARGET=");

			t_end = cmd.find_first_not_of("0123456789", t_begin);
			if(t_end == std::string::npos) t_end = cmd.size();

			if(t_end > t_begin) {
				target = boost::lexical_cast<unsigned>(cmd.substr(t_begin, t_end-t_begin));
			}
		}

		blkaddr ba;
		if(src && mark) {
			ba.head(suite, src, mark, target);
		}
	};
	void operator() (const unsigned suite, const unsigned src, const unsigned mark, const unsigned target)
	{
		log.infoStream() << "got: " << src << "(" << mark << ") -> " << target << l4::eol;
		blkaddr ba;
		if(src && mark) {
			ba.head(suite, src, mark, target);
		}
	};
};

class tail
{
	l4::Category &log;
public:
	tail() : log(l4::Category::getInstance("tail")) {};

	void operator()(const std::string cmd)
	{
		unsigned suite = 0;
		size_t suite1 = std::string::npos;
		size_t suite2 = std::string::npos;
		size_t s_begin = std::string::npos;
		size_t s_end = std::string::npos;
		size_t m_begin = std::string::npos;
		size_t m_end = std::string::npos;
		size_t t_begin = std::string::npos;
		size_t t_end = std::string::npos;

		log.infoStream() << "got: " << cmd << l4::eol;

		suite1 = cmd.find("SUITE=1");
		suite2 = cmd.find("SUITE=2");
		if((suite1 == std::string::npos) && (suite2 == std::string::npos)) {
			log.infoStream() << "no SUITE tag" << l4::eol;
			return;
		}
		if(suite1 != std::string::npos) suite = 1;
		if(suite2 != std::string::npos) suite = 2;

		s_begin = cmd.find("SRC=");
		if(s_begin == std::string::npos) return;
		s_begin += strlen("SRC=");

		s_end = cmd.find_first_not_of("0123456789", s_begin);
		if(s_end == std::string::npos) s_end = cmd.size();

		unsigned src = 0;
		if(s_end > s_begin) {
			src = boost::lexical_cast<unsigned>(cmd.substr(s_begin, s_end-s_begin));
		}
		log.infoStream() << "src: " << src << l4::eol;

		m_begin = cmd.find("MARK=");
		if(m_begin == std::string::npos) return;
		m_begin += strlen("MARK=");

		m_end = cmd.find_first_not_of("0123456789", m_begin);
		if(m_end == std::string::npos) m_end = cmd.size();

		unsigned mark = 0;
		if(m_end > m_begin) {
			mark = boost::lexical_cast<unsigned>(cmd.substr(m_begin, m_end-m_begin));
		}
		log.infoStream() << "mark: " << mark << l4::eol;

		unsigned target = 0;
		t_begin = cmd.find("TARGET=");
		if(t_begin != std::string::npos) {
			t_begin += strlen("TARGET=");

			t_end = cmd.find_first_not_of("0123456789", t_begin);
			if(t_end == std::string::npos) t_end = cmd.size();

			if(t_end > t_begin) {
				target = boost::lexical_cast<unsigned>(cmd.substr(t_begin, t_end-t_begin));
			}
			log.infoStream() << "target: " << target << l4::eol;
		}

		blkaddr ba;
		log.infoStream() << "got: " << src << "(" << mark << ") -> " << target << l4::eol;
		if(src && mark) {
			ba.tail(suite, src, mark, target);
		}
	};
	void operator() (const unsigned suite, const unsigned src, const unsigned mark, const unsigned target)
	{
		blkaddr ba;
		log.infoStream() << "got: " << src << "(" << mark << ") -> " << target << l4::eol;
		if(src && mark) {
			ba.tail(suite, src, mark, target);
		}
	};
};

class marry
{
	l4::Category &log;
public:
	marry() : log(l4::Category::getInstance("marry")) {};

	void operator()(const std::string cmd)
	{
		unsigned suite = 0;
		size_t suite1 = std::string::npos;
		size_t suite2 = std::string::npos;
		size_t s_begin = std::string::npos;
		size_t s_end = std::string::npos;
		size_t t_begin = std::string::npos;
		size_t t_end = std::string::npos;

		log.infoStream() << "got: " << cmd << l4::eol;

		suite1 = cmd.find("SUITE=1");
		suite2 = cmd.find("SUITE=2");
		if((suite1 == std::string::npos) && (suite2 == std::string::npos)) {
			log.infoStream() << "no SUITE tag" << l4::eol;
			return;
		}
		if(suite1 != std::string::npos) suite = 1;
		if(suite2 != std::string::npos) suite = 2;

		s_begin = cmd.find("KEY=");
		if(s_begin == std::string::npos) return;
		s_begin += strlen("KEY=");

		s_end = cmd.find_first_not_of("0123456789", s_begin);
		if(s_end == std::string::npos) s_end = cmd.size();

		unsigned src = 0;
		if(s_end > s_begin) {
			src = boost::lexical_cast<unsigned>(cmd.substr(s_begin, s_end-s_begin));
		}

		t_begin = cmd.find("VIDEO=");
		if(t_begin == std::string::npos) return;
		t_begin += strlen("VIDEO=");

		t_end = cmd.find_first_not_of("0123456789", t_begin);
		if(t_end == std::string::npos) t_end = cmd.size();

		unsigned target = 0;
		if(t_end > t_begin) {
			target = boost::lexical_cast<unsigned>(cmd.substr(t_begin, t_end-t_begin));
		}

		blkaddr ba;
		log.infoStream() << "got: " << src << " -> " << target << l4::eol;
		if(src && target) {
			ba.marry(suite, src, target);
		}
	};
	void operator() (const unsigned suite, const unsigned src, const unsigned target)
	{
		blkaddr ba;
		log.infoStream() << "got: " << src << " -> " << target << l4::eol;
		if(src && target) {
			ba.marry(suite, src, target);
		}
	};
};

/// \brief application entry point, launches a capture or display loop
/// - parse command line args with boost::program_options,
/// \param argc the number of arguments
/// \param argv the command line std::strings
int main(int argc, char *argv[])
{
  int retval = 0;
  l4::Category &log = l4::Category::getRoot();
  std::fstream hfs; ///< host device stream
  move move;
  swap swap;
  append append;
  head head;
  tail tail;
  marry marry;

  hfs.exceptions(std::ios::eofbit | std::ios::failbit | std::ios::badbit);

  try {
	unsigned src = 0;
	unsigned mark = 0;
	unsigned target = 0;
	po::options_description desc("Allowed options");

	desc.add_options()
		("src",     po::value<unsigned>(), "src image ID")
		("mark",    po::value<unsigned>(), "field offset")
		("target",  po::value<unsigned>(), "target image ID")

		("head",    "trim the head of an image")
		("tail",    "trim the tail of an image")
		("move",    "move an image")
		("swap",    "swap two images")
		("append",  "append one image to another")
		("marry",   "marry one image to another")

		("host",    po::value<std::string>(), "host dpm interface")
		("daemon", "run as a daemon process")

		("init",    "initialize a test MossDirectory ")
		("dump",    "dump MossDirectory ")
		("remove",  "remove MossDirectory ")

		("log",  "log to console, or as controlled by config file")
		("help",    "produce help message")
	;

	po::variables_map vm;
	po::store(po::parse_command_line(argc, argv, desc), vm);
	po::notify(vm);

	if (vm.count("log")) {
		try {
			l4::PropertyConfigurator::configure(".l4rc");
			std::cout << "logging configured" << std::endl;
			log.infoStream() << "logging started" << l4::eol;
		} catch(log4cpp::ConfigureFailure e) {
			l4::BasicConfigurator::configure();
			log.infoStream() << e.what() << l4::eol;
		}
	}

	if (vm.count("src")) {
		src = vm["src"].as<unsigned>();
	}
	if (vm.count("mark")) {
		mark = vm["mark"].as<unsigned>();
	}
	if (vm.count("target")) {
		target = vm["target"].as<unsigned>();
	}
	if (vm.count("head")) {
		head(1, src, mark, target);
	}
	if (vm.count("tail")) {
		tail(1, src, mark, target);
	}
	if (vm.count("swap")) {
		swap(1, src, target);
	}
	if (vm.count("move")) {
		move(1, src, target);
	}
	if (vm.count("append")) {
		append(1, src, target);
	}
	if (vm.count("marry")) {
		marry(1, src, target);
	}

	if (vm.count("init")) {
		blkaddr ba;
		for(size_t id=1; id<10; id++) {
			for(size_t field=1; field<10; field++) {
				std::stringstream video;
				std::stringstream key;
				video << "1:" << std::setfill('0') << std::setw(4) << id 
					<< ":" << std::setfill('0') << std::setw(4) << field 
					<< ":v" << std::ends;
				key << "1:" << std::setfill('0') << std::setw(4) << id 
					<< ":" << std::setfill('0') << std::setw(4) << field 
					<< ":k" << std::ends;
				ba.assign(video.str(), 1024, 1000);
		//		ba.assign(key.str(), 1024, 1000);
			}
		}
		log.infoStream() << "MossDirectory initialized" << l4::eol;
	}
	if (vm.count("dump")) {
		blkaddr ba;
		ba.dump(std::cout);
		log.infoStream() << "MossDirectory dumped" << l4::eol;
	}
	if (vm.count("remove")) {
		boost::interprocess::shared_memory_object::remove ("MossDirectory");
		log.infoStream() << "MossDirectory removed" << l4::eol;
	}
	if (vm.count("help")) {
		std::cerr << desc << std::endl;
	}

        if (vm.count("host")) {
		const std::string &device = vm["host"].as<std::string>();
                hfs.open(device.c_str());
        }
	if (vm.count("daemon")) {
		errno = 0;
		daemon(0,0);
		if(errno) {
			log.critStream() << "daemon() failed: " << strerror(errno) << l4::eol;
			throw std::runtime_error("daemon");
		}
	}

	if(hfs.is_open()) {
		std::string token;
		char buf[BUFSIZ];

		while("forever") {
			try {
				token.clear();
				hfs >> token;
			
				if(token == "EDIT=head") {
					hfs.getline(buf, BUFSIZ);
					sched_yield(); // I think we need this but Ingo would disagree
					head(buf);
					continue;
				}
				if(token == "EDIT=tail") { 
					hfs.getline(buf, BUFSIZ); 
					sched_yield(); // I think we need this but Ingo would disagree
					tail(buf); 
					continue;
				}
				if(token == "EDIT=append") {
					hfs.getline(buf, BUFSIZ);
					sched_yield(); // I think we need this but Ingo would disagree
					append(buf);
					continue;
				}
				if(token == "EDIT=marry") { 
					hfs.getline(buf, BUFSIZ); 
					sched_yield(); // I think we need this but Ingo would disagree
					marry(buf); 
					continue; 
				}
				if(token == "EDIT=swap") { 
					hfs.getline(buf, BUFSIZ); 
					sched_yield(); // I think we need this but Ingo would disagree
					swap(buf); 
					continue; 
				}
				if(token == "EDIT=move") { 
					hfs.getline(buf, BUFSIZ); 
					sched_yield(); // I think we need this but Ingo would disagree
					move(buf); 
					continue;
				}
				hfs.ignore(BUFSIZ, '\n');
			} catch ( std::ifstream::failure &e ) {
				const std::string &device = vm["host"].as<std::string>();
				log.warnStream() << "std::ifstream exception: " << e.what() << l4::eol;
  				if(hfs.is_open()) hfs.close();
				hfs.clear();
		                hfs.open(device.c_str());
  				if(!hfs.is_open()) break;
			}
		}
		std::cout << "normal exit." << std::endl;
	}
  } catch(std::exception &e) {
                std::cout << e.what() <<  " exiting..." << std::endl;
		retval = ~0;
  } catch(...) {
                std::cout << "unexpected exception, exiting..." << std::endl;
		retval = ~0;
  } 
  if(hfs.is_open()) hfs.close();
  l4::Category::shutdown();
  return retval;
}

