#include <iostream>
#include <iomanip>
#include <fstream>
#include <cerrno>
#include <vector>

#include <arpa/inet.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <net/route.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <boost/lexical_cast.hpp>
#include <boost/tokenizer.hpp>
#include <boost/program_options.hpp>
#include <boost/regex.hpp>

#include <log4cpp/Category.hh>
#include <log4cpp/BasicConfigurator.hh>
#include <log4cpp/PropertyConfigurator.hh>

namespace po = boost::program_options;
namespace l4 = log4cpp;

class calibrate
{
        l4::Category &log;
        const size_t def;
	std::string data;
	static const size_t numtaps = 64;
public:
        calibrate(const size_t size, const size_t def) : log(l4::Category::getInstance("calibrate")), def(def)
	{
		data = " ";
		for(size_t i=0; i<size; ++i) data += "1 ";
	}

        void operator() (const char *dev, size_t reps)
	{
		log.debugStream() << "set " << dev << " to " << reps << l4::eol;
		if(reps > numtaps) {
			log.infoStream() << "using default: " << def << l4::eol;
			reps = def;
		}
    		std::ofstream calfs(dev); ///< calibration device stream

		if(calfs.is_open()) {
			log.infoStream() << "opened: " << dev << l4::eol;

			log.infoStream() << "iterating to: " << reps << l4::eol;
			for(int i=0; i<reps; ++i) {
				calfs << data << std::endl;
			}
			calfs.close();
		} 
	}
};

class interrogate
{
        l4::Category &log;
        const size_t size;
	const size_t reps;
	static const size_t numtaps = 64;
public:
        interrogate(const size_t size) : log(l4::Category::getInstance("interrogate")), size(size), reps(9) {}

        size_t operator() (const char *dev)
	{
    		std::ifstream ifs(dev); ///< interrogation device stream
		std::vector<std::string> channels(size);

		std::vector<std::string>::iterator itr = channels.begin();
		while(itr!=channels.end()) {
			(*itr).assign(numtaps, '0');
			++itr;
		}

		if(ifs.is_open()) {
			log.debugStream() << "opened: " << dev << l4::eol;

			for(size_t i=0; i<(numtaps*reps); ++i) {
				ifs.ignore(BUFSIZ, ':');

				size_t tap;
				ifs >> tap;

				int value;
				std::vector<std::string>::iterator itr = channels.begin();
				while(itr!=channels.end()) {
					value = 0;
					ifs >> value;
					if(value==1) (*itr).at(tap)++;
					++itr;
				}

				ifs.ignore(BUFSIZ, '\n');
			}
			ifs.close();
		} 

		int mean = 0;
		const boost::regex e("([^01]*)([01]+)");
		for(int i=0; i<channels.size(); ++i) {
			boost::smatch what;
			boost::regex_search(channels[i], what, e);

			int value = (what.position(2) + what.length(2)/2) - what.length(2);
			if(value < 0) value = 0;
			if(value > numtaps) value = numtaps;
			if(i<10) {
				log.noticeStream() << "CH:0" << i << " " << channels[i] 
		//			<< " = " << what.position(2) << ":" << what.length(2) << l4::eol;
					<< " = " << value << l4::eol;
			} else {
				log.noticeStream() << "CH:" << i << " " << channels[i] 
		//			<< " = " << what.position(2) << ":" << what.length(2) << l4::eol;
					<< " = " << value << l4::eol;
			}
			mean += value;
		}
		return mean / channels.size();
	}
};

class options
{
        l4::Category &log;
        const char *dev;
	unsigned old;
public:
        options(const char *dev) : log(l4::Category::getInstance("options")), dev(dev) {}

        void operator() (const std::string &cmd)
	{
		log.debugStream() << "got: " << cmd << l4::eol;
		unsigned raw;

		if(cmd.find("DIMM=2") == std::string::npos) {
			if(cmd.find("1080i") != std::string::npos) raw=0x001;
			if(cmd.find("720p")  != std::string::npos) raw=0x101;
			if(cmd.find("625i")  != std::string::npos) raw=0x201;
			if(cmd.find("525i")  != std::string::npos) raw=0x301;
		} else {
			if(cmd.find("1080i") != std::string::npos) raw=0x401;
			if(cmd.find("720p")  != std::string::npos) raw=0x501;
			if(cmd.find("625i")  != std::string::npos) raw=0x601;
			if(cmd.find("525i")  != std::string::npos) raw=0x701;
		}
		if(raw == old) return;

		log.noticeStream() << "changing " 
					<< "0x" << std::hex << std::setfill('0') << std::setw(4) << old 
					<< " -> " 
					<< "0x" << std::hex << std::setfill('0') << std::setw(4) << raw 
					<< l4::eol;
    		std::ofstream cfg(dev); ///< cfg device stream
		if(cfg.is_open()) {
			cfg << "CTL raw=0x" << std::hex << std::setfill('0') << std::setw(4) << raw << std::endl;
			cfg.close();
		} 

		old = raw;
	}

        void operator() (const unsigned raw)
	{
		log.noticeStream() << "changing " 
					<< "0x" << std::hex << std::setfill('0') << std::setw(4) << old 
					<< " -> " 
					<< "0x" << std::hex << std::setfill('0') << std::setw(4) << raw 
					<< l4::eol;
    		std::ofstream cfg(dev); ///< cfg device stream
		if(cfg.is_open()) {
			cfg << "CTL raw=0x" << std::hex << std::setfill('0') << std::setw(4) << raw << std::endl;
			cfg.close();
		} 

		old = raw;
	}

};

class assign
{
        l4::Category &log;
        const struct timespec &pause;
	std::string old;
        void implementation(const std::string &hw, const std::string &ip, const std::string &nm, const std::string &gw);
public:
        assign( const struct timespec &pause) : log(l4::Category::getInstance("assign")), pause(pause) {}

        void operator()(const std::string cmd)
        {
		if(cmd == old) return;

		// ETH=0 HWADDR=00:B0:09:01:72:67 IPADDR=10.16.20.81 MASK=255.255.248.0 GATEWAY=10.16.16.1
		log.noticeStream() << "got: " << cmd << l4::eol;
		old = cmd;

		size_t hw_begin = cmd.find("HWADDR=");
		size_t ip_begin = cmd.find("IPADDR=");
		size_t nm_begin = cmd.find("MASK=");
		size_t gw_begin = cmd.find("GATEWAY=");

                if(hw_begin != std::string::npos) hw_begin += strlen("HWADDR=");
                if(ip_begin != std::string::npos) ip_begin += strlen("IPADDR=");
                if(nm_begin != std::string::npos) nm_begin += strlen("MASK=");
                if(gw_begin != std::string::npos) gw_begin += strlen("GATEWAY=");

		size_t hw_end = cmd.find_first_not_of(":0123456789abcdefABCDEF", hw_begin);
		size_t ip_end = cmd.find_first_not_of(".0123456789", ip_begin);
		size_t nm_end = cmd.find_first_not_of(".0123456789", nm_begin);
		size_t gw_end = cmd.find_first_not_of(".0123456789", gw_begin);

                if(hw_end == std::string::npos) hw_end = cmd.size();
                if(ip_end == std::string::npos) ip_end = cmd.size();
                if(hw_end == std::string::npos) nm_end = cmd.size();
                if(hw_end == std::string::npos) gw_end = cmd.size();


        	implementation(
				cmd.substr(hw_begin, hw_end-hw_begin),
				cmd.substr(ip_begin, ip_end-ip_begin),
				cmd.substr(nm_begin, nm_end-nm_begin),
				cmd.substr(gw_begin, gw_end-gw_begin)
			);
        }

        void operator() (const std::string &hw, const std::string &ip, const std::string &nm, const std::string &gw)
	{
        	implementation(hw,ip,nm,gw);
	}
};

void assign::implementation(const std::string &hw, const std::string &ip, const std::string &nm, const std::string &gw)
{
	int fd = socket(PF_INET, SOCK_DGRAM, 0);
	if(errno) throw std::runtime_error("socket");

	log.debugStream() << "hw: " << hw << l4::eol;
	log.debugStream() << "ip: " << ip << l4::eol;
	log.debugStream() << "nm: " << nm << l4::eol;
	log.debugStream() << "gw: " << gw << l4::eol;

	// you are going down!
	{
		struct ifreq ifr;
		memset(&ifr, 0, sizeof(ifr));
		strcpy(ifr.ifr_name, "eth0");

		ifr.ifr_flags = 0;

		errno = 0;
		ioctl(fd, SIOCSIFFLAGS, &ifr);
		if(errno) throw std::runtime_error("SIOCSIFFLAGS");
	}

	nanosleep(&pause, NULL); // time for the hardware to react

	// assign hardware address
	{
		struct ifreq ifr;
		memset(&ifr, 0, sizeof(ifr));
		strcpy(ifr.ifr_name, "eth0");

		int i;
		std::string buf(hw);
		for(size_t pos=buf.find(':'); pos != std::string::npos; pos=buf.find(':', pos)) {
			buf[pos] = ' ';
		}
		std::stringstream s(buf);
		s >> std::hex >> i; ifr.ifr_hwaddr.sa_data[0] = i;
		s >> std::hex >> i; ifr.ifr_hwaddr.sa_data[1] = i;
		s >> std::hex >> i; ifr.ifr_hwaddr.sa_data[2] = i;
		s >> std::hex >> i; ifr.ifr_hwaddr.sa_data[3] = i;
		s >> std::hex >> i; ifr.ifr_hwaddr.sa_data[4] = i;
		s >> std::hex >> i; ifr.ifr_hwaddr.sa_data[5] = i;
		if(!s) throw std::runtime_error("parsing");

		ifr.ifr_hwaddr.sa_family = ARPHRD_ETHER;

		errno = 0;
		ioctl(fd, SIOCSIFHWADDR, &ifr);
		if(errno) throw std::runtime_error("SIOCSIFHWADDR");
	}

	// assign ip address
	{
		struct ifreq ifr;
		struct sockaddr_in addr;
		memset(&ifr, 0, sizeof(ifr));
		memset(&addr, 0, sizeof(addr));

		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = inet_addr(ip.c_str());
		strcpy(ifr.ifr_name, "eth0");
		memcpy(&ifr.ifr_addr, &addr, sizeof(addr));

		errno = 0;
		ioctl(fd, SIOCSIFADDR, &ifr);
		if(errno) throw std::runtime_error("SIOCSIFADDR");
	}

	// assign netmask
	{
		struct ifreq ifr;
		struct sockaddr_in netmask;
		memset(&ifr, 0, sizeof(ifr));
		memset(&netmask, 0, sizeof(netmask));

		netmask.sin_family = AF_INET;
		netmask.sin_addr.s_addr = inet_addr(nm.c_str());
		strcpy(ifr.ifr_name, "eth0");
		memcpy(&ifr.ifr_netmask, &netmask, sizeof(netmask));

		errno = 0;
		ioctl(fd, SIOCSIFNETMASK, &ifr);
		if(errno) throw std::runtime_error("SIOCSIFNETMASK");
		}

	// start your engines
	{
		struct ifreq ifr;
		memset(&ifr, 0, sizeof(ifr));
		strcpy(ifr.ifr_name, "eth0");

		ifr.ifr_flags |= IFF_UP;

		errno = 0;
		ioctl(fd, SIOCSIFFLAGS, &ifr);
		if(errno) throw std::runtime_error("SIOCSIFFLAGS");
	}

	nanosleep(&pause, NULL); // time for the hardware to react

	// add default route
	{
		struct sockaddr_in dst;
		struct sockaddr_in gateway;
		struct sockaddr_in netmask;
		struct rtentry route;
		memset(&dst, 0, sizeof(dst));
		memset(&netmask, 0, sizeof(netmask));
		memset(&gateway, 0, sizeof(gateway));
		memset(&route, 0, sizeof(route));

		dst.sin_family = AF_INET;
		dst.sin_addr.s_addr = inet_addr("0.0.0.0");
		memcpy(&route.rt_dst, &dst, sizeof(dst));

		netmask.sin_family = AF_INET;
		netmask.sin_addr.s_addr = inet_addr(nm.c_str());
		memcpy(&route.rt_genmask, &netmask, sizeof(netmask));

		gateway.sin_family = AF_INET;
		gateway.sin_addr.s_addr = inet_addr(gw.c_str());
		memcpy(&route.rt_gateway, &gateway, sizeof(gateway));

		route.rt_flags |= RTF_UP;
		route.rt_flags |= RTF_GATEWAY;
		route.rt_flags |= RTF_STATIC;

		errno = 0;
		ioctl(fd, SIOCADDRT, &route);
		if(errno) throw std::runtime_error("SIOCADDRT");
	}

	close(fd);
}


int main(int ac, char* av[])
{
    int val = 0;
    l4::Category &log = l4::Category::getRoot();
    std::ifstream hfs; ///< host device stream
    std::string hw; ///< hardware (mac) address
    std::string ip; ///< ip address
    std::string nm; ///< netmask
    std::string gw; ///< default gateway
    const struct timespec pause = { 0, 4 * 1000000 }; // 4ms ~= 1/4 field
    assign assign(pause);
    options options("/dev/moCFG");
    interrogate dqs(18);
    calibrate dimm(18, 23);
    interrogate qr(11);
    calibrate rx(11, 14);

    hfs.exceptions(std::ios::eofbit | std::ios::failbit | std::ios::badbit);

    try {
        po::options_description desc("Allowed options");
        desc.add_options()
                ("help",                             "produce help message")
                ("log",                              "log to console")
                ("hw",    po::value<std::string>(),  "hardware (mac) address")
                ("dcm",                              "Digital Clock Manager reset")
                ("mgt",                              "Multi-Gigabit Transceiver reset")
                ("ip",    po::value<std::string>(),  "ip address")
                ("nm",    po::value<std::string>(),  "netmask")
                ("gw",    po::value<std::string>(),  "default gateway")
                ("up",                               "bring up the ethernet port")
                ("dqs",                              "query DIMM controller")
                ("dimm",  po::value<unsigned>(),     "calibrate DIMM timing")
                ("auto-dimm",                        "auto-calibrate DIMM timing")
                ("qr1",                              "query RX1 (input) timing")
                ("qr2",                              "query RX2 (input) timing")
                ("rx1",   po::value<unsigned>(),     "calibrate auto-timer 1")
                ("auto-rx1",                         "auto-calibrate auto-timer 1")
                ("rx2",   po::value<unsigned>(),     "calibrate auto-timer 2")
                ("auto-rx2",                         "auto-calibrate auto-timer 2")
                ("host",  po::value<std::string>(),  "host dpm interface")
                ("daemon",                           "run as a daemon process")
        ;

        po::variables_map vm;
        po::store(po::parse_command_line(ac, av, desc), vm);
        po::notify(vm);

	if (vm.count("help")) {
		std::cout << desc << std::endl;
	}
	if (vm.count("log")) {
		try {
			l4::PropertyConfigurator::configure(".l4rc");
			std::cout << "logging configured" << std::endl;
			log.noticeStream() << "logging started" << l4::eol;
		} catch(log4cpp::ConfigureFailure e) {
			l4::BasicConfigurator::configure();
			log.noticeStream() << "console logging" << l4::eol;
		}
	}
	if (vm.count("hw")) {
		hw = vm["hw"].as<std::string>();
	}
	if (vm.count("ip")) {
		ip = vm["ip"].as<std::string>();
	}
	if (vm.count("nm")) {
		nm = vm["nm"].as<std::string>();
	}
	if (vm.count("gw")) {
		gw = vm["gw"].as<std::string>();
	}
	if (vm.count("up")) {
		log.critStream() << "got: up" << l4::eol;
		assign(hw,ip,nm,gw);
		exit(0);
	}
	if (vm.count("dcm")) {
		log.critStream() << "got: dcm" << l4::eol;
    		std::ofstream cfs("/dev/moDCM"); ///< cfg device stream
		if(cfs.is_open()) {
			log.noticeStream() << "DCM reset" << l4::eol;
			cfs << "reset" << std::endl;
			cfs.close();
		}
	}
	if (vm.count("mgt")) {
		log.critStream() << "got: mgt" << l4::eol;
    		std::ofstream cfs("/dev/moMGT"); ///< cfg device stream
		if(cfs.is_open()) {
			log.noticeStream() << "MGT reset" << l4::eol;
			cfs << "reset" << std::endl;
			cfs.close();
		}
	}
	if (vm.count("dqs")) {
		log.critStream() << "got: dqs" << l4::eol;
    		std::ofstream cfs; ///< cfg device stream
		cfs.open("/dev/moVCT");
		if(cfs.is_open()) {
			log.noticeStream() << "set video controller training" << l4::eol;
			cfs << "reset" << std::endl;
			cfs.close();
		}
		dqs("/dev/moDQS");
		cfs.open("/dev/moVCA");
		if(cfs.is_open()) {
			log.noticeStream() << "set video controller active" << l4::eol;
			cfs << "reset" << std::endl;
			cfs.close();
		}
	}
	if (vm.count("dimm")) {
		unsigned value = vm["dimm"].as<unsigned>();
		log.critStream() << "got: dimm" << l4::eol;
		log.noticeStream() << "set calibration values=" << value << l4::eol;
    		std::ofstream cfs; ///< cfg device stream
		cfs.open("/dev/moVCT");
		if(cfs.is_open()) {
			log.noticeStream() << "set video controller training" << l4::eol;
			cfs << "reset" << std::endl;
			cfs.close();
		}
		dimm("/dev/moDQS", value);
		cfs.open("/dev/moVCA");
		if(cfs.is_open()) {
			log.noticeStream() << "set video controller active" << l4::eol;
			cfs << "reset" << std::endl;
			cfs.close();
		}
	}
	if (vm.count("auto-dimm")) {
		log.critStream() << "got: auto-dimm" << l4::eol;
    		std::ofstream cfs; ///< cfg device stream
		cfs.open("/dev/moVCT");
		if(cfs.is_open()) {
			log.noticeStream() << "set video controller training" << l4::eol;
			cfs << "reset" << std::endl;
			cfs.close();
		}
		dimm("/dev/moDQS", dqs("/dev/moDQS"));
		cfs.open("/dev/moVCA");
		if(cfs.is_open()) {
			log.noticeStream() << "set video controller active" << l4::eol;
			cfs << "reset" << std::endl;
			cfs.close();
		}
	}
	if(vm.count("qr1")) {
		log.critStream() << "got: qr1" << l4::eol;
		log.noticeStream() << "interrogating qr 1" << l4::eol;
    		std::ofstream cfg; ///< cfg device stream
    		cfg.open("/dev/moCFG"); ///< cfg device stream
		if(cfg.is_open()) {
			log.noticeStream() << "set auto-timer training" << l4::eol;
			cfg << "CTL raw=0x2" << std::endl;
			cfg.close();
		} 
		qr("/dev/moRX1");
		if(cfg.is_open()) {
			log.noticeStream() << "clear auto-timer training" << l4::eol;
			cfg << "CTL raw=0x0" << std::endl;
			cfg.close();
		} 
	}
	if (vm.count("qr2")) {
		log.critStream() << "got: qr2" << l4::eol;
		log.noticeStream() << "interrogating qr 2" << l4::eol;
    		std::ofstream cfg; ///< cfg device stream
    		cfg.open("/dev/moCFG"); ///< cfg device stream
		if(cfg.is_open()) {
			log.noticeStream() << "set auto-timer training" << l4::eol;
			cfg << "CTL raw=0x2" << std::endl;
			cfg.close();
		} 
		qr("/dev/moRX2");
		if(cfg.is_open()) {
			log.noticeStream() << "clear auto-timer training" << l4::eol;
			cfg << "CTL raw=0x0" << std::endl;
			cfg.close();
		} 
	}
	if (vm.count("rx1")) {
		unsigned value = vm["rx1"].as<unsigned>();
		log.critStream() << "got: rx1" << l4::eol;
		log.noticeStream() << "set calibration rx1=" << value << l4::eol;
		rx("/dev/moRX1", value);
		sleep(1);
	}
	if (vm.count("rx2")) {
		unsigned value = vm["rx2"].as<unsigned>();
		log.critStream() << "got: rx2" << l4::eol;

		log.noticeStream() << "set calibration rx2=" << value << l4::eol;
		rx("/dev/moRX2", value);
		sleep(1);
	}
	if(vm.count("auto-rx1")) {
		log.critStream() << "got: auto-rx1" << l4::eol;
    		std::ofstream cfg; ///< cfg device stream
    		cfg.open("/dev/moCFG"); ///< cfg device stream
		if(cfg.is_open()) {
			log.noticeStream() << "set auto-timer training" << l4::eol;
			cfg << "CTL raw=0x2" << std::endl;
			cfg.close();
		} 
		rx("/dev/moRX1", qr("/dev/moRX1"));
		if(cfg.is_open()) {
			log.noticeStream() << "clear auto-timer training" << l4::eol;
			cfg << "CTL raw=0x0" << std::endl;
			cfg.close();
		} 
	}
	if (vm.count("auto-rx2")) {
		log.critStream() << "got: auto-rx2" << l4::eol;
    		std::ofstream cfg; ///< cfg device stream
    		cfg.open("/dev/moCFG"); ///< cfg device stream
		if(cfg.is_open()) {
			log.noticeStream() << "set auto-timer training" << l4::eol;
			cfg << "CTL raw=0x2" << std::endl;
			cfg.close();
		} 
		rx("/dev/moRX2", qr("/dev/moRX2"));
		if(cfg.is_open()) {
			log.noticeStream() << "clear auto-timer training" << l4::eol;
			cfg << "CTL raw=0x0" << std::endl;
			cfg.close();
		} 
	}
	if (vm.count("host")) {
		const std::string &device = vm["host"].as<std::string>();
		hfs.open(device.c_str());
	}
	if (vm.count("daemon")) {
		errno = 0;
		daemon(0,0);
		if(errno) {
			log.critStream() << "daemon() failed: " << strerror(errno) << l4::eol;
			throw std::runtime_error("daemon");
		}
	}

        if(hfs.is_open()) {
                std::string token;
                char buf[BUFSIZ];

                while("forever") {
			try {
	                        token.clear();
				hfs >> token;
				if(token == "ETH=0") {
					hfs.getline(buf, BUFSIZ);
					sched_yield(); // I think we need this but Ingo would disagree
					assign(buf);
					continue;
				}
				if(token == "OPTIONS") {
					hfs.getline(buf, BUFSIZ);
					sched_yield(); // I think we need this but Ingo would disagree
					options(buf);
					continue;
				}
				hfs.ignore(BUFSIZ, '\n');
                        } catch ( std::ifstream::failure &e ) {
				const std::string &device = vm["host"].as<std::string>();
				log.warnStream() << "std::ifstream exception: " << e.what() << l4::eol;
				if(hfs.is_open()) hfs.close();
				hfs.clear();
				hfs.open(device.c_str());
				if(!hfs.is_open()) break;
                        }
                }

        }
    }
    catch(std::exception& e) {
        log.warnStream() << "error: " << e.what() << l4::eol;
        val = -1;
    }
    catch(...) {
        log.warnStream() << "Exception of unknown type" << l4::eol;
        val = -1;
    }

    if(hfs.is_open()) hfs.close();
    l4::Category::shutdown();
    return val;
}

