/// \file
/// \copyright 2008 Thomson Grass Valley
/// \author adam clarke
/// \date 2008-06-06
/// \brief respond to host controller messages 
///
#include <stdexcept>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <map>

#include <boost/program_options.hpp>

#include <log4cpp/Category.hh>
#include <log4cpp/BasicConfigurator.hh>
#include <log4cpp/PropertyConfigurator.hh>

#include "blkaddr.h"
#include "mosstiffload.hpp"
#include "mosstiffsave.hpp"

namespace po = boost::program_options;
namespace l4 = log4cpp;

/// \brief application entry point, launches a capture or display loop
/// - parse command line args with boost::program_options,
/// \param argc the number of arguments
/// \param argv the command line std::strings
int main(int argc, char *argv[])
{
  int val = 0; // normal return value
  unsigned suite;
  unsigned id;
  std::string std;
  std::string cache("/dev/moCM1");
  std::string mountload("/mnt/load");
  std::string mountsave("/mnt/save");
  std::string unc;
  std::vector< std::pair<std::string,std::string> > labels;
  std::ifstream src;  ///< host device stream
  std::ofstream sink; ///< host device stream
  long rowsperstrip = 31;
  bool lzw = true;
  l4::Category &log = l4::Category::getRoot();

  try {
	src.exceptions(std::ios::eofbit | std::ios::failbit | std::ios::badbit);
	sink.exceptions(std::ios::eofbit | std::ios::failbit | std::ios::badbit);

	po::options_description desc("Allowed options");

	desc.add_options()
		("suite",  po::value<unsigned>(), "suite number")
		("id",     po::value<unsigned>(), "image number")
		("std",    po::value<std::string>(), "linerate string [1080i|720p|625i|525i]")
		("cache",  po::value<std::string>(), "cache device")
		("mount",  po::value<std::string>(), "image mount point")
		("unc",    po::value<std::string>(), "unc encoded pathname")
		("rowsperstrip",   po::value<long>(), "rows per strip, default 31")
		("no-compression",    "use no compression")
		("load",   "load an image")
		("save",   "save an image")

		("host",   po::value<std::string>(), "host dpm interface")
		("daemon", "run as a daemon process")

		("dump",   "dump MossDirectory ")
		("remove", "remove MossDirectory ")

		("log",    "log to console")
		("help",   "produce help message")
	;
	po::variables_map vm;
	po::store(po::parse_command_line(argc, argv, desc), vm);
	po::notify(vm);

	if (vm.count("log")) {
		try {
			l4::PropertyConfigurator::configure(".l4rc");
			std::cout << "logging configured" << std::endl;
			log.noticeStream() << "logging started" << l4::eol;
		} catch(log4cpp::ConfigureFailure e) {
			l4::BasicConfigurator::configure();
			log.noticeStream() << "console logging" << l4::eol;
		}
	}

	if (vm.count("cache")) {
		cache = vm["cache"].as<std::string>();
		log.debugStream() << "cache=" << cache << l4::eol;
	}
	if (vm.count("mount")) {
		mountload = vm["mount"].as<std::string>();
		mountsave = vm["mount"].as<std::string>();
		log.debugStream() << "mountload =" << mountload << l4::eol;
		log.debugStream() << "mountsave =" << mountsave << l4::eol;
	}
	if (vm.count("suite")) {
		suite = vm["suite"].as<unsigned>();
		log.debugStream() << "suite=" << unc << l4::eol;
		if(suite > 2) suite = 0;
	}
	if (vm.count("id")) {
		id = vm["id"].as<unsigned>();
		log.debugStream() << "id=" << id << l4::eol;
		if(id > 9999) id = 0;
	}
	if (vm.count("std")) {
		std = vm["std"].as<std::string>();
		log.debugStream() << "std=" << unc << l4::eol;
	}
	if (vm.count("unc")) {
		unc = vm["unc"].as<std::string>();
		log.debugStream() << "unc=" << unc << l4::eol;
	}
	if (vm.count("rowsperstrip")) {
		rowsperstrip = vm["rowsperstrip"].as<long>();
		log.debugStream() << "rowsperstrip=" << rowsperstrip << l4::eol;
	}
	if (vm.count("no-compression")) {
		lzw = false;
		log.debugStream() << "use lzw=" << lzw << l4::eol;
	}
	if (vm.count("load")) {
		load load(cache, mountload, std::cout);
		log.debugStream() << "loading..." << l4::eol;
		load(suite, id, unc);
	}
	if (vm.count("save")) {
		blkaddr ba;
		save save(cache, mountsave, rowsperstrip, lzw, std::cout);
		log.debugStream() << "saving..." << l4::eol;
		if(std.find("720p") == std::string::npos) {
			ba.interlace(labels,suite,id);
		} else {
			ba.progressive(labels,suite,id);
		}
		save(labels, "video", unc);
	}

	if (vm.count("dump")) {
		blkaddr ba;
		log.debugStream() << "dumping..." << l4::eol;
		ba.dump(std::cout);
		log.noticeStream() << "MossDirectory dumped" << l4::eol;
	}
	if (vm.count("remove")) {
		log.debugStream() << "removing..." << l4::eol;
		boost::interprocess::shared_memory_object::remove("MossDirectory");
		log.noticeStream() << "MossDirectory removed" << l4::eol;
	}
	if (vm.count("help")) {
		log.debugStream() << "helping..." << l4::eol;
		std::cerr << desc << std::endl;
	}

	if (vm.count("host")) {
		const std::string &device = vm["host"].as<std::string>();
		log.debugStream() << "host=" << device << l4::eol;
		src.open(device.c_str());
		sink.open(device.c_str());
	}
	if (vm.count("daemon")) {
		log.debugStream() << "daemoning..." << l4::eol;
		errno = 0;
		if(daemon(0,0)) {
			exit(0);
		}
		if(errno) {
			log.critStream() << "daemon() failed: " << strerror(errno) << l4::eol;
			throw std::runtime_error("daemon");
		}
	}

	if(src.is_open() && sink.is_open()) {
		log.debugStream() << "src and sink open" << l4::eol;
		load load(cache, mountload, sink);
		save save(cache, mountsave, rowsperstrip, lzw, sink);
		std::string token;

		while("forever") {
			try {
				token.clear();
				src >> token;
				if(token == "JOB=load") { 
					sched_yield(); // I think we need this but Ingo would disagree
					load(src); 
					continue; 
				}
				if(token == "JOB=save") { 
					sched_yield(); // I think we need this but Ingo would disagree
					save(src); 
					continue; 
				}
				src.ignore(BUFSIZ, '\n');
			} catch ( std::fstream::failure &e ) {
				const std::string &device = vm["host"].as<std::string>();
				log.warnStream() << "std::fstream exception: " << e.what() << l4::eol;

				if(src.is_open()) src.close();
				src.clear();
				src.open(device.c_str());
				if(!src.is_open()) break;

				if(sink.is_open()) sink.close();
				sink.clear();
				sink.open(device.c_str());
				if(!sink.is_open()) break;
			}
		}
	}

  	log.noticeStream() << "normal exit." << l4::eol;
  } catch(std::exception &e) {
	log.critStream() << "unexpected exception, " << e.what() << l4::eol;
	val = ~0;
  } catch(...) {
	log.critStream() << "unexpected exception, exiting..." << l4::eol;
	val = ~0;
  }

  if(src.is_open()) src.close();
  if(sink.is_open()) sink.close();
  l4::Category::shutdown();
  return val;
}

