/// \file
/// \copyright 2008 Thomson Grass Valley
/// \author adam clarke
/// \date 2008-06-06
/// \brief the missouri image cache engine.
///
/// also passes embedded audio
///
/// example command lines:
/// - \code ./mossice --host /dev/moDPM --video /dev/moCH1 --key /dev/moCH2 [ runs a channel ]
/// - \code ./mossice --destroy \endcode [ destroys the shared memory directory ]
/// - \code ./mossice --dump \endcode [ dumps the shared memory directory to stdout ]
/// - \code ./mossice \endcode [ loads required shared objects and exits ]
///
#include <iostream>
#include <fstream>
#include <string>
#include <vector>

#include <boost/lexical_cast.hpp>
#include <boost/program_options.hpp>

#include <log4cpp/Category.hh>
#include <log4cpp/BasicConfigurator.hh>
#include <log4cpp/PropertyConfigurator.hh>

#include "mossio.h"
#include "blkaddr.h"

namespace po = boost::program_options;
namespace l4 = log4cpp;

/// \brief application entry point, launches a capture or display loop
///
/// - parse command line args with boost::program_options, 
/// - call worker (capture, display, response...) methods as required
/// - handle any exceptions that are thrown by the worker methods
///
/// \param argc the number of arguments
/// \param argv the array of command line strings
int main(int argc, char* argv[])
{
	videocmd::format fmt = videocmd::f1080i;
	size_t vintok = 0; ///< video input port, defaults to first input
	size_t kintok = 1; ///< key input port, defaults to second input
	size_t voutok = 2; ///< video output port, defaults to first output
	size_t koutok = 3; ///< key output port, defaults to second output

	std::fstream hfs; ///< host interrupt device stream
	std::fstream kfs; ///< key channel device stream
	std::fstream vfs; ///< video channel device stream
#ifdef ENABLE_AUDIO
	std::fstream efs1; ///< audio extractor device stream
	std::fstream efs2; ///< audio extractor device stream
	std::fstream ifs; ///< audio inserter device stream
#endif // ENABLE_AUDIO

        // respond to all I/O std::exceptions (just error out and exit)
        hfs.exceptions(std::ios::eofbit | std::ios::failbit | std::ios::badbit);
        kfs.exceptions(std::ios::eofbit | std::ios::failbit | std::ios::badbit);
        vfs.exceptions(std::ios::eofbit | std::ios::failbit | std::ios::badbit);
#ifdef ENABLE_AUDIO
        efs1.exceptions(std::ios::eofbit | std::ios::failbit | std::ios::badbit);
        efs2.exceptions(std::ios::eofbit | std::ios::failbit | std::ios::badbit);
        ifs.exceptions(std::ios::eofbit | std::ios::failbit | std::ios::badbit);
#endif // ENABLE_AUDIO

	const tlbr_t open = { 0, 0, 540, 2016 };
	const tlbr_t shut = { 0, 0, 0, 0 };
	const yuv_t err_clr = { 128, 512, 640 };

	try {
		po::options_description desc("Allowed options");

		desc.add_options()
			("help", "produce help message")
                        ("log",  "log to console")

			("daemon", "run as a daemon process")
			("dump", "dump the contents of the cache directory")
			("remove", "remove the cache directory")
#ifdef ENABLE_AUDIO
			("extractor", po::value<std::string>(), "extractor port")
			("inserter", po::value<std::string>(), "inserter port")
#endif // ENABLE_AUDIO
			("format",   po::value<std::string>(), "video format")
			("host",    po::value<std::string>(), "host dpm interface")
			("key", po::value<std::string>(), "key channel")
			("video", po::value<std::string>(), "video channel")
		;
		po::variables_map vm;
		po::store(po::parse_command_line(argc, argv, desc), vm);
		po::notify(vm);

		l4::Category &log = l4::Category::getRoot();
		if (vm.count("log")) {
			try {
				l4::PropertyConfigurator::configure(".l4rc");
				std::cout << "logging configured" << std::endl;
				log.noticeStream() << "logging started" << l4::eol;
			} catch(log4cpp::ConfigureFailure e) {
				l4::BasicConfigurator::configure();
				log.noticeStream() << "console logging" << l4::eol;
			}
		}
		if (vm.count("help")) {
			std::cerr << desc << std::endl;
		}
		if (vm.count("dump")) {
			blkaddr ba;
			ba.dump(std::cout);
		}
#ifdef ENABLE_AUDIO
		if (vm.count("extractor")) {
			efs1.open(vm["extractor"].as<std::string>().c_str(), std::fstream::in);
			efs2.open(vm["extractor"].as<std::string>().c_str(), std::fstream::out);
		}
		if (vm.count("inserter")) {
			ifs.open(vm["inserter"].as<std::string>().c_str(), std::fstream::out);
		}
#endif // ENABLE_AUDIO
		if (vm.count("format")) {
			const std::string &val = vm["format"].as<std::string>();
			if(val == "1080i") fmt = videocmd::f1080i;
			if(val == "720p") fmt = videocmd::f720p;
			if(val == "625i") fmt = videocmd::f625i;
			if(val == "525i") fmt = videocmd::f525i;
		}
		if (vm.count("host")) {
			hfs.open(vm["host"].as<std::string>().c_str(), std::fstream::in);
		}
		if (vm.count("key")) {
			kfs.open(vm["key"].as<std::string>().c_str(), std::fstream::out);
		}
		if (vm.count("video")) {
			vfs.open(vm["video"].as<std::string>().c_str(), std::fstream::out);
		}
		if (vm.count("remove")) {
			boost::interprocess::shared_memory_object::remove("MossDirectory");
			log.noticeStream() << "MossDirectory removed" << l4::eol;
		}
		if (vm.count("daemon")) {
			if(daemon(0,0)) {
				exit(0);
			}
		}

		{
			log.infoStream() << "initializing..." << l4::eol;
			const struct timespec pause = { 0, 4 * 1000000 }; // 4ms ~= 1/4 field
			for(int i=0; i<4; ++i) {
				if(vfs.is_open()) {
					vfs << videocmd(0, 0, 0, shut, shut, err_clr, err_clr, fmt) << std::endl;
				}
				if(kfs.is_open()) {
					kfs << videocmd(0, 0, 0, shut, shut, err_clr, err_clr, fmt) << std::endl;
				}
				nanosleep(&pause, NULL); // time for the hardware to react
			}
			log.infoStream() << "initialized." << l4::eol;
		}

		if(hfs.is_open()) {
#ifdef ENABLE_AUDIO
			std::string input[6];
#endif // ENABLE_AUDIO
			blkaddr ba; 
			char buf[BUFSIZ];

			std::vector<std::string> cmds(8);
			std::vector<std::string> t1(8);
			std::vector<std::string> t2(8);
			std::istringstream vin;
			std::istringstream vout;
			std::istringstream kin;
			std::istringstream kout;
			std::string token;

			std::string vfmt;
			std::string kfmt;

			std::string waddr;
			std::string raddr;

			int vstreak=0;
			int kstreak=0;

			int vrec=0;
			int krec=0;

			size_t vwblk = 0;
			size_t vrblk = 0;
			size_t kwblk = 0;
			size_t krblk = 0;

			tlbr_t vpos = shut;
			tlbr_t vcrop = shut;
			yuv_t  vmask = err_clr;
			tlbr_t kpos = shut;
			tlbr_t kcrop = shut;
			yuv_t  kmask = err_clr;

			int seq = 0;


			while("forever") {
				vin.clear();
				vout.clear();
				kin.clear();
				kout.clear();

				t2 = t1;
				t1 = cmds;
				++seq;

				// the host device blocks until a command is issued 
				try {
					std::vector<std::string>::iterator itr;
					for(itr=cmds.begin(); itr!=cmds.end(); itr++) {
						getline(hfs, (*itr)); 
						if((*itr).find("\tEND") != std::string::npos) break;
					}
				} catch(std::exception &e) {
					//throw std::runtime_error(std::string("host: ") + e.what());
					log.warnStream() << "rdstate=" << hfs.rdstate() << l4::eol;
					hfs.clear();
					continue;
				}

				bool vinChanged = (cmds[vintok]!=t1[vintok]) || (t1[vintok]!=t2[vintok]) || (seq < 600);
				bool voutChanged = (cmds[voutok]!=t1[voutok]) || (t1[voutok]!=t2[voutok]) || (seq < 600);

				if(voutChanged) {
					vstreak = 0; 
				} else { 
					++vstreak; 
				}


				// create unique field identifier, send the video channel command
				if(vfs.is_open()) {
					if(cmds[vintok].empty()) {
						vrec=0;
						vwblk = 0;
					} else if(vinChanged) {
						try {
							tlbr_t pos;
							//log.debugStream() << "vin cmd:  " << cmds[vintok] << l4::eol; 
							vin.str(cmds[vintok]);

							token.clear();
							vin >> token;
							//log.debugStream() << "TOKEN:  " << token << l4::eol; 

							waddr.clear();
							vin.getline(buf, sizeof(buf), '=');
							if(strstr(buf, "LABEL")) {
								vin >> waddr;
								if(waddr.find("0000") != std::string::npos) {
									log.warnStream() << "waddr:  " << waddr << l4::eol; 
									throw std::runtime_error("invalid label");
								}
							}

							vin.getline(buf, sizeof(buf), '=');
							if(strstr(buf, "POS")) {
								vin.ignore(5);
								vin >> pos.t;
								vin.ignore(1);
								vin >> pos.l;
								vin.ignore(1);
								vin >> pos.b;
								vin.ignore(1);
								vin >> pos.r;
								vin.ignore(3);
								//log.debugStream() << "POS=" 
								//	<< pos.t << ":" << pos.l << ":" << pos.b << ":" << pos.r
								//	<< l4::eol;
							}

							if(!waddr.empty()) {
								const size_t fldsiz = (pos.b - pos.t) * (1+(pos.r-pos.l)/112);
								vwblk = ba.assign(waddr, fldsiz+1, fldsiz);
								vrec=1;
							}
						} catch(...) {
							vrec=0;
							vwblk = 0;
							log.debugStream() << "vin parsing failed" << l4::eol;
						}
					}

					if(cmds[voutok].empty()) {
							vrblk = 0;
							vpos = vcrop = shut;
					} else if(vstreak<2) {
						try {
							//log.debugStream() << "vout cmd:  " << cmds[voutok] << l4::eol; 
							vout.str(cmds[voutok]);

							vfmt.clear();
							vout >> vfmt;
							//log.debugStream() << "FMT:  " << vfmt << l4::eol; 

							vout.getline(buf, sizeof(buf), '=');
							if(strstr(buf, "LABEL")) {
								vrblk = 0;
								vpos = vcrop = shut;
								raddr.clear();

								vout >> raddr;
								if((!raddr.empty()) && (raddr.find("0000") == std::string::npos)) {
									vrblk = ba.addr(raddr);
									vpos = vcrop = open;
								}
							} else {
								log.debugStream() << "LABEL expected, vout got: " << buf << l4::eol; 
								throw std::runtime_error("invalid label");
							}

							vout.getline(buf, sizeof(buf), '=');
							if(vrblk && strstr(buf, "POS")) {
								vout.ignore(5);
								vout >> vpos.t;
								vout.ignore(1);
								vout >> vpos.l;
								vout.ignore(1);
								vout >> vpos.b;
								vout.ignore(1);
								vout >> vpos.r;
								vout.ignore(3);
								//log.debugStream() << "POS=" 
								//	<< pos.t << ":" << pos.l << ":" << pos.b << ":" << pos.r
								//	<< l4::eol;
							}

							vout.getline(buf, sizeof(buf), '=');
							if(vrblk && strstr(buf, "CROP")) {
								vout.ignore(5);
								vout >> vcrop.t;
								vout.ignore(1);
								vout >> vcrop.l;
								vout.ignore(1);
								vout >> vcrop.b;
								vout.ignore(1);
								vout >> vcrop.r;
								vout.ignore(3);
								//log.debugStream() << "CROP=" 
								//	<< crop.t << ":" << crop.l << ":" << crop.b << ":" << crop.r
								//	<< l4::eol;
							}

							vout.getline(buf, sizeof(buf), '=');
							if(strstr(buf, "MASK")) {
								vout.ignore(6);
								vout >> vmask.y;
								vout.ignore(1);
								vout >> vmask.u;
								vout.ignore(1);
								vout >> vmask.v;
								vout.ignore(6);
							//	log.debugStream() << "MASK=" 
							//		<< vmask.y << "," << vmask.u << "," << vmask.v << l4::eol;
							}
						} catch(...) {
							vrblk = 0;
							vpos = vcrop = shut;
							vmask = err_clr;
							cmds[voutok] += " failed";
							log.debugStream() << "vout parsing failed" << l4::eol;
						}
					}
				}

				if(vinChanged || (vstreak<2)) {
					try {
						if(vfmt == "FMT=1080i") fmt = videocmd::f1080i;
						if(vfmt == "FMT=720p") fmt = videocmd::f720p;
						if(vfmt == "FMT=625i") fmt = videocmd::f625i;
						if(vfmt == "FMT=525i") fmt = videocmd::f525i;
						vfs << videocmd(vrec, vwblk, vrblk, vpos, vcrop, vmask, err_clr, fmt) << std::endl;
					} catch(std::exception &e) {
						throw std::runtime_error(std::string("video: ") + e.what());
					}
				}

				bool kinChanged = (cmds[kintok]!=t1[kintok]) || (t1[kintok]!=t2[kintok]) ||(seq < 600);
				bool koutChanged = (cmds[koutok]!=t1[koutok]) || (t1[koutok]!=t2[koutok]) ||(seq < 600);

				if(koutChanged) {
					kstreak = 0; 
				} else { 
					++kstreak; 
				}

				// create unique field identifier, send the key channel command
				if(kfs.is_open()) {
					if(cmds[kintok].empty()) {
						krec=0;
						kwblk = 0;
					} else if(kinChanged) {
						try {
							tlbr_t pos;
							//log.warnStream() << "kin cmd:  " << cmds[kintok] << l4::eol; 
							kin.str(cmds[kintok]);

							token.clear();
							kin >> token;
							//log.warnStream() << "TOKEN:  " << token << l4::eol; 

							waddr.clear();
							kin.getline(buf, sizeof(buf), '=');
							if(strstr(buf, "LABEL")) {
								kin >> waddr;
								if(waddr.find("0000") != std::string::npos) {
									throw std::runtime_error("invalid label");
								}
							}

							kin.getline(buf, sizeof(buf), '=');
							if(strstr(buf, "POS")) {
								kin.ignore(5);
								kin >> pos.t;
								kin.ignore(1);
								kin >> pos.l;
								kin.ignore(1);
								kin >> pos.b;
								kin.ignore(1);
								kin >> pos.r;
								kin.ignore(3);
								//log.debugStream() << "POS=" 
								//	<< pos.t << ":" << pos.l << ":" << pos.b << ":" << pos.r
								//	<< l4::eol;
							}

							if(!waddr.empty()) {
								const size_t fldsiz = (pos.b - pos.t) * (1+(pos.r-pos.l)/112);
								kwblk = ba.assign(waddr, fldsiz+1, fldsiz);
								krec=1;
							}
						} catch(...) {
							krec=0;
							kwblk = 0;
							log.debugStream() << "kin parsing failed" << l4::eol;
						}
					}

					if(cmds[koutok].empty()) {
							krblk = 0;
							kpos = kcrop = shut;
					} else if(kstreak<2) {
						try {
							kout.str(cmds[koutok]);

							kfmt.clear();
							kout >> kfmt;
							//log.warnStream() << "FMT:  " << kfmt << l4::eol; 

							kout.getline(buf, sizeof(buf), '=');
							if(strstr(buf, "LABEL")) {
								krblk = 0;
								kpos = kcrop = shut;
								raddr.clear();

								kout >> raddr;
								if((!raddr.empty()) && (raddr.find("0000") == std::string::npos)) {
									krblk = ba.addr(raddr);
									kpos = kcrop = open;
								}
							} else {
								log.debugStream() << "LABEL expected, kout got: " << buf << l4::eol; 
								throw std::runtime_error("invalid label");
							}

							kout.getline(buf, sizeof(buf), '=');
							if(krblk && strstr(buf, "POS")) {
								kout.ignore(5);
								kout >> kpos.t;
								kout.ignore(1);
								kout >> kpos.l;
								kout.ignore(1);
								kout >> kpos.b;
								kout.ignore(1);
								kout >> kpos.r;
								kout.ignore(3);
								//log.debugStream() << "POS=" 
								//	<< pos.t << ":" << pos.l << ":" << pos.b << ":" << pos.r 
								//	<< l4::eol;
							}

							kout.getline(buf, sizeof(buf), '=');
							if(krblk && strstr(buf, "CROP")) {
								kout.ignore(5);
								kout >> kcrop.t;
								kout.ignore(1);
								kout >> kcrop.l;
								kout.ignore(1);
								kout >> kcrop.b;
								kout.ignore(1);
								kout >> kcrop.r;
								kout.ignore(3);
								//log.debugStream() << "CROP=" 
								//	<< crop.t << ":" << crop.l << ":" << crop.b << ":" << crop.r
								//	<< l4::eol;
							}

							kout.getline(buf, sizeof(buf), '=');
							if(strstr(buf, "MASK")) {
								kout.ignore(6);
								kout >> kmask.y;
								kout.ignore(1);
								kout >> kmask.u;
								kout.ignore(1);
								kout >> kmask.v;
								kout.ignore(6);
								//log.debugStream() << "MASK=" 
								//	<< kmask.y << "," << kmask.u << "," << kmask.v << l4::eol;
							}
						} catch(...) {
							krec=0;
							kwblk = krblk = 0;
							kpos = kcrop = shut;
							kmask = err_clr;
							cmds[koutok] += " failed";
							log.debugStream() << "kout parsing failed" << l4::eol;
						}
					}
				}

				if(kinChanged || (kstreak<2)) {
					try {
						if(kfmt == "FMT=1080i") fmt = videocmd::f1080i;
						if(kfmt == "FMT=720p") fmt = videocmd::f720p;
						if(kfmt == "FMT=625i") fmt = videocmd::f625i;
						if(kfmt == "FMT=525i") fmt = videocmd::f525i;
						kfs << videocmd(krec, kwblk, krblk, kpos, kcrop, kmask, err_clr, fmt) << std::endl;
					} catch(std::exception &e) {
						throw std::runtime_error(std::string("key: ") + e.what());
					}
				}
#ifdef ENABLE_AUDIO
				// get extractor status
				// find the size of the extracted block
				if(efs1.is_open()) {
					try {
						efs1.seekg(0);
						for(int i=1; i<6; i++) {
							input[i].erase();
							getline(efs1,input[i]);
						}

						size_t start = input[1].find("size=");
						size_t end = input[1].find("blks");
						if((start != std::string::npos) && (end != std::string::npos)
							&& ((end-start) == 7) // size is in 10-99blks range
						) {
							start += strlen("size=");
							std::string buf = input[1].substr(start,end-start);
							ba.consume(buf, boost::lexical_cast<size_t>(buf));
						} else {
							ba.consume(buf, 42);
						}
					} catch(std::exception &e) {
						throw std::runtime_error(std::string("extractor status: ") + e.what());
					}
				}

				// create unique field identifier, send the extractor commands
				if(efs2.is_open()) {
					try {
						if(!cmd.label(vintok).empty()) {
							efs2.seekg(0);
							std::string buf = cmd.label(vintok) + "a";
							size_t addr = ba.assign(buf, 99+1);
							efs2 << "EXTRACTOR:enable addr=" << addr << std::endl;
						}
					} catch(std::exception &e) {
						throw std::runtime_error(std::string("extractor commands: ") + e.what());
					}
				}

				// create unique field identifier, send the inserter commands
				// (note the special case [no audio yet] for the first three fields
				if(ifs.is_open()) {
					try {
						if(!cmd.label(voutok).empty()) {
							std::string mode;
							if(!cmd.audio(voutok).empty()) {
								mode = "enable";
							} else {
								mode = "disable";
							}
							std::string buf = cmd.label(voutok) + "a";
							size_t addr = ba.addr(buf);
							size_t used = ba.used(buf);
							ifs << "INSERTER:"  << mode 
								<< " addr=" << addr
								<< " size=" << used << "blks"
								<< std::endl;
						}
					} catch(std::exception &e) {
						throw std::runtime_error(std::string("inserter commands: ") + e.what());
					}
				}
#endif // ENABLE_AUDIO
			}
		}
	} catch(std::exception& e) {
		std::cerr << "error: " << e.what() << std::endl;
	} catch(...) {
		std::cerr << "Exception of unknown type!" << std::endl;
	}

	// clean up if we are done
#ifdef ENABLE_AUDIO
	if(efs1.is_open()) efs1.close();
	if(efs2.is_open()) efs2.close();
	if(ifs.is_open()) ifs.close();
#endif // ENABLE_AUDIO
	if(hfs.is_open()) hfs.close();
	if(vfs.is_open()) vfs.close();
	if(kfs.is_open()) kfs.close();

	l4::Category::shutdown();
	return 0;
}

