/// \file
/// \copyright 2008 Thomson Grass Valley
/// \author adam clarke
/// \date 2008-06-06
/// \brief a collection of inserters and extractors to support ImageStore operations

#include <iostream>

#include "mossio.h"

/// \brief 1080i constants
namespace hd1080i {
	const int FIELD_SEGS	= 18; ///< number of segments in a horizontal line
	const int FIELD_LINES	= 540; ///< number of lines in a field

	const int PIXELS_PER_SEG	= 112; ///< number of pixels in a segment
	const int PIXELS_PER_LINE	= PIXELS_PER_SEG * FIELD_SEGS;
};

/// \brief 720p constants
namespace hd720p {
	const int FIELD_SEGS	= 12; ///< number of segments in a horizontal line
	const int FIELD_LINES	= 720; ///< number of lines in a field

	const int PIXELS_PER_SEG	= 112; ///< number of pixels in a segment
	const int PIXELS_PER_LINE	= PIXELS_PER_SEG * FIELD_SEGS;
};

/// \brief 625i constants
namespace sd625i {
	const int FIELD_SEGS	= 7; ///< number of segments in a horizontal line
	const int FIELD_LINES	= 576; ///< number of lines in a field

	const int PIXELS_PER_SEG	= 112; ///< number of pixels in a segment
	const int PIXELS_PER_LINE	= PIXELS_PER_SEG * FIELD_SEGS;
};

/// \brief 525i constants
namespace sd525i {
	const int FIELD_SEGS	= 7; ///< number of segments in a horizontal line
	const int FIELD_LINES	= 480; ///< number of lines in a field

	const int PIXELS_PER_SEG	= 112; ///< number of pixels in a segment
	const int PIXELS_PER_LINE	= PIXELS_PER_SEG * FIELD_SEGS;
};

/// \brief video channel command serialization functor
std::ostream &videocmd::operator()(std::ostream &sink) const {
	// find segs per frame: segs per line * lines per frame
	int segs;

	switch(fmt) {
	case f1080i:
		segs = ((hd1080i::PIXELS_PER_SEG/2) + pos.r - pos.l) / hd1080i::PIXELS_PER_SEG;
		break;
	case f720p:
		segs = ((hd720p::PIXELS_PER_SEG/2) + pos.r - pos.l) / hd720p::PIXELS_PER_SEG;
		break;
	case f625i:
		segs = ((sd625i::PIXELS_PER_SEG/2) + pos.r - pos.l) / sd625i::PIXELS_PER_SEG;
		break;
	case f525i:
		segs = ((sd525i::PIXELS_PER_SEG/2) + pos.r - pos.l) / sd525i::PIXELS_PER_SEG;
		break;
	default:
		segs = 0;
		break;
	}
	if(pos.t<0) {
		segs *= (pos.b);
	} else {
		segs *= (pos.b - pos.t);
	}

	// ChannelRegister
	sink << chreg << " ";
	// WriteStartX
	switch(fmt) {
	case f1080i:
		sink << (pos.l/2) << " ";
		break;
	case f720p:
		sink << (pos.l/2) << " ";
		break;
	case f625i:
		sink << (pos.l/2) << " ";
		break;
	case f525i:
		sink << (pos.l/2) << " ";
		break;
	default:
		sink << (pos.l/2) << " ";
		break;
	}
	// WriteStartY
	if(pos.t<0) {
		sink << 0 << " ";
	} else {
		sink << (pos.t) << " ";
	}
	// WriteCountX
	switch(fmt) {
	case f1080i:
		sink << 1+(pos.r-pos.l)/hd1080i::PIXELS_PER_SEG << " ";
		break;
	case f720p:
		sink << 1+(pos.r-pos.l)/hd720p::PIXELS_PER_SEG << " ";
		break;
	case f625i:
		sink << 1+(pos.r-pos.l)/sd625i::PIXELS_PER_SEG << " ";
		break;
	case f525i:
		sink << 1+(pos.r-pos.l)/sd525i::PIXELS_PER_SEG << " ";
		break;
	default:
		sink << 0 << " ";
		break;
	}
	// WriteCountY
	if(pos.t<0) {
		sink << (pos.b) << " ";
	} else {
		sink << (pos.b-pos.t) << " ";
	}
	// RecordAddress
	if(pos.t<0) {
		switch(fmt) {
		case f1080i:
			sink << wblk + hd1080i::FIELD_SEGS*(-pos.t) << " ";
			break;
		case f720p:
			sink << wblk + hd720p::FIELD_SEGS*(-pos.t) << " ";
			break;
		case f625i:
			sink << wblk + sd625i::FIELD_SEGS*(-pos.t) << " ";
			break;
		case f525i:
			sink << wblk + sd525i::FIELD_SEGS*(-pos.t) << " ";
			break;
		default:
			sink << wblk << " ";
			break;
		}
	} else {
		sink << wblk << " ";
	}
	// ReadStartX
	switch(fmt) {
	case f1080i:
		if(pos.l<-hd1080i::PIXELS_PER_SEG) {
			int i = pos.l;
			while(i<-hd1080i::PIXELS_PER_SEG) {
				i += hd1080i::PIXELS_PER_SEG;
			}
			sink << (i/2) << " ";
		} else {
			sink << (pos.l/2) << " ";
		}
		break;
	case f720p:
		if(pos.l<-hd720p::PIXELS_PER_SEG) {
			int i = pos.l;
			while(i<-hd720p::PIXELS_PER_SEG) {
				i += hd1080i::PIXELS_PER_SEG;
			}
			sink << (i/2) << " ";
		} else {
			sink << (pos.l/2) << " ";
		}
		break;
	case f625i:
		if(pos.l<-sd625i::PIXELS_PER_SEG) {
			int i = pos.l;
			while(i<-sd625i::PIXELS_PER_SEG) {
				i += hd1080i::PIXELS_PER_SEG;
			}
			sink << (i/2) << " ";
		} else {
			sink << (pos.l/2) << " ";
		}
		break;
	case f525i:
		if(pos.l<-sd525i::PIXELS_PER_SEG) {
			int i = pos.l;
			while(i<-sd525i::PIXELS_PER_SEG) {
				i += hd1080i::PIXELS_PER_SEG;
			}
			sink << (i/2) << " ";
		} else {
			sink << (pos.l/2) << " ";
		}
		break;
	default:
		sink << (pos.l/2) << " ";
		break;
	}
	// ReadStartY
	if(pos.t<0) {
		sink << 0 << " ";
	} else {
		sink << pos.t << " ";
	}
	// ReadCountX
	switch(fmt) {
	case f1080i:
		if(pos.l<-hd1080i::PIXELS_PER_SEG) {
			int i = pos.l;
			while(i<-hd1080i::PIXELS_PER_SEG) {
				i += hd1080i::PIXELS_PER_SEG;
			}
			sink << 1+(pos.r-i)/hd1080i::PIXELS_PER_SEG << " ";
		} else {
			sink << 1+(pos.r-pos.l)/hd1080i::PIXELS_PER_SEG << " ";
		}
		break;
	case f720p:
		if(pos.l<-hd720p::PIXELS_PER_SEG) {
			int i = pos.l;
			while(i<-hd720p::PIXELS_PER_SEG) {
				i += hd720p::PIXELS_PER_SEG;
			}
			sink << 1+(pos.r-i)/hd720p::PIXELS_PER_SEG << " ";
		} else {
			sink << 1+(pos.r-pos.l)/hd720p::PIXELS_PER_SEG << " ";
		}
		break;
	case f625i:
		if(pos.l<-sd625i::PIXELS_PER_SEG) {
			int i = pos.l;
			while(i<-sd625i::PIXELS_PER_SEG) {
				i += sd625i::PIXELS_PER_SEG;
			}
			sink << 1+(pos.r-i)/sd625i::PIXELS_PER_SEG << " ";
		} else {
			sink << 1+(pos.r-pos.l)/sd625i::PIXELS_PER_SEG << " ";
		}
		break;
	case f525i:
		if(pos.l<-sd525i::PIXELS_PER_SEG) {
			int i = pos.l;
			while(i<-sd525i::PIXELS_PER_SEG) {
				i += sd525i::PIXELS_PER_SEG;
			}
			sink << 1+(pos.r-i)/sd525i::PIXELS_PER_SEG << " ";
		} else {
			sink << 1+(pos.r-pos.l)/sd525i::PIXELS_PER_SEG << " ";
		}
		break;
	default:
		sink << 0 << " ";
		break;
	}
	// ReadCountY
	if(pos.t<0) {
		sink << (pos.b) << " ";
	} else {
		sink << (pos.b-pos.t) << " ";
	}
	// PlaybackAddress
	if(pos.t<0) {
		switch(fmt) {
		case f1080i:
			{
			size_t offset = 0;
			int i = pos.l;
			while(i<-hd1080i::PIXELS_PER_SEG) {
				i += hd1080i::PIXELS_PER_SEG;
				++offset;
			}
			sink << rblk + hd1080i::FIELD_SEGS*(-pos.t) + offset << " ";
			}
			break;
		case f720p:
			{
			size_t offset = 0;
			int i = pos.l;
			while(i<-hd720p::PIXELS_PER_SEG) {
				i += hd720p::PIXELS_PER_SEG;
				++offset;
			}
			sink << rblk + hd720p::FIELD_SEGS*(-pos.t) + offset << " ";
			}
			break;
		case f625i:
			{
			size_t offset = 0;
			int i = pos.l;
			while(i<-sd625i::PIXELS_PER_SEG) {
				i += sd625i::PIXELS_PER_SEG;
				++offset;
			}
			sink << rblk + sd625i::FIELD_SEGS*(-pos.t) + offset << " ";
			}
			break;
		case f525i:
			{
			size_t offset = 0;
			int i = pos.l;
			while(i<-sd525i::PIXELS_PER_SEG) {
				i += sd525i::PIXELS_PER_SEG;
				++offset;
			}
			sink << rblk + sd525i::FIELD_SEGS*(-pos.t) + offset << " ";
			}
			break;
		default:
			sink << rblk << " ";
			break;
		}
	} else {
		switch(fmt) {
		case f1080i:
			{
			size_t offset = 0;
			int i = pos.l;
			while(i<-hd1080i::PIXELS_PER_SEG) {
				i += hd1080i::PIXELS_PER_SEG;
				++offset;
			}
			sink << rblk + offset << " ";
			}
			break;
		case f720p:
			{
			size_t offset = 0;
			int i = pos.l;
			while(i<-hd720p::PIXELS_PER_SEG) {
				i += hd720p::PIXELS_PER_SEG;
				++offset;
			}
			sink << rblk + offset << " ";
			}
			break;
		case f625i:
			{
			size_t offset = 0;
			int i = pos.l;
			while(i<-sd625i::PIXELS_PER_SEG) {
				i += sd625i::PIXELS_PER_SEG;
				++offset;
			}
			sink << rblk + offset << " ";
			}
			break;
		case f525i:
			{
			size_t offset = 0;
			int i = pos.l;
			while(i<-sd525i::PIXELS_PER_SEG) {
				i += sd525i::PIXELS_PER_SEG;
				++offset;
			}
			sink << rblk + offset << " ";
			}
			break;
		default:
			sink << rblk << " ";
			break;
		}
	}
	// PlaybackOffset
	switch(fmt) {
	case f1080i:
		{
			size_t offset = 0;
			int i = pos.l;
			while(i<-hd1080i::PIXELS_PER_SEG) {
				i += hd1080i::PIXELS_PER_SEG;
				++offset;
			}
			sink << offset << " ";
		}
		break;
	case f720p:
		{
			size_t offset = 0;
			int i = pos.l;
			while(i<-hd720p::PIXELS_PER_SEG) {
				i += hd720p::PIXELS_PER_SEG;
				++offset;
			}
			sink << offset << " ";
		}
		break;
	case f625i:
		{
			size_t offset = 0;
			int i = pos.l;
			while(i<-sd625i::PIXELS_PER_SEG) {
				i += sd625i::PIXELS_PER_SEG;
				++offset;
			}
			sink << offset << " ";
		}
		break;
	case f525i:
		{
			size_t offset = 0;
			int i = pos.l;
			while(i<-sd525i::PIXELS_PER_SEG) {
				i += sd525i::PIXELS_PER_SEG;
				++offset;
			}
			sink << offset << " ";
		}
		break;
	default:
		sink << 0 << " ";
		break;
	}
	// MaskLeft
	sink << (mask.l/2) << " ";
	// MaskTop
	sink << mask.t << " ";
	// MaskRight
	sink << (mask.r/2) << " ";
	// MaskBottom
	sink << mask.b << " ";
	// MaskY
	sink << clr.y << " ";
	// MaskCb
	sink << clr.u << " ";
	// MaskCr
	sink << clr.v << " ";
	// BorderY
	sink << bdr.y << " ";
	// BorderCb
	sink << bdr.u << " ";
	// BorderCr
	sink << bdr.v << " ";

        return sink;
}

/// \brief insert a video channel command
std::ostream &operator<<(std::ostream &out, const videocmd &msg) { return msg(out); }

