/* Linux Image Engine Configuration Module */

#include <asm/current.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/byteorder.h>

#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/wait.h>
#include <linux/kernel.h>
#include <linux/device.h>

#include <xilinx/xparameters.h>
#if 0
#define RS_RX1_IDLY_CTL (XPAR_RAMSTORE2PLB_BASEADDR + 0x0C)
#define RS_RX2_IDLY_CTL (XPAR_RAMSTORE2PLB_BASEADDR + 0x10)
#define RS_RX1_RB (XPAR_RAMSTORE2PLB_BASEADDR + 0x40C)
#define RS_RX2_RB (XPAR_RAMSTORE2PLB_BASEADDR + 0x410)
#else
#define RS_RX1_IDLY_CTL (XPAR_XPS_REGISTERINT_BRAM_BASEADDR + 0x0C)
#define RS_RX2_IDLY_CTL (XPAR_XPS_REGISTERINT_BRAM_BASEADDR + 0x10)
#define RS_RX1_RB (XPAR_XPS_REGISTERINT_BRAM_BASEADDR + 0x40C)
#define RS_RX2_RB (XPAR_XPS_REGISTERINT_BRAM_BASEADDR + 0x410)
#endif
#define RX_IDLY_RST (1<<31)
#define RX_IDLY_INC (1<<30)
#define LINELEN (7 + 11*2 + 1)

#define MODULE_NAME "mossrx"
MODULE_AUTHOR("GVG");
MODULE_DESCRIPTION("ImageStore RX Calibration Driver");
MODULE_VERSION("0:0.1-experimental");

static struct resource *r = NULL;
static int major = 0; /* the major number assigned to our driver */
static int mrx_init(void);
static void mrx_cleanup(void);

module_init(mrx_init);
module_exit(mrx_cleanup);

int mrx_open(struct inode *inode,struct file *filep);
int mrx_release(struct inode *inode,struct file *filep);
ssize_t mrx_read(struct file *filep,char *buff,size_t count,loff_t *offp );
ssize_t mrx_write(struct file *filep,const char *buff,size_t count,loff_t *offp );
	
struct file_operations mrx_fops={
	open: mrx_open,
	read: mrx_read,
	write: mrx_write,
	release:mrx_release,
};

int mrx_open(struct inode *inode,struct file *filep)
{
	unsigned int minor = iminor(inode);
	switch(minor) {
	case 1:
	case 2:
		printk(KERN_DEBUG MODULE_NAME ": open minor=%d\n", minor);
		return 0;
	}
	return -ENODEV;
}

int mrx_release(struct inode *inode,struct file *filep)
{
	unsigned int minor = iminor(inode);
	switch(minor) {
	case 1:
	case 2:
		printk(KERN_DEBUG MODULE_NAME ": release minor=%d\n", minor);
		return 0;
	}
	return -ENODEV;
}

ssize_t mrx_read(struct file *filep,char *buff,size_t count,loff_t *offp )
{
	int err;
	char data[PAGE_SIZE];
	unsigned int minor = iminor(filep->f_dentry->d_inode);
	unsigned rb = 0;
	void *ctl = NULL;
	void *addr = NULL;
	size_t tap = ((*offp) & 0xFFFFFFFF);

        printk(KERN_DEBUG MODULE_NAME ": read, minor=%d offp=%d\n", minor, tap);
	tap /= LINELEN;
	tap %= 64;

	switch(minor) {
	case 1:
		ctl = ioremap(RS_RX1_IDLY_CTL, sizeof(unsigned long));
		addr = ioremap(RS_RX1_RB, sizeof(unsigned long));
		break;
	case 2:
		ctl = ioremap(RS_RX2_IDLY_CTL, sizeof(unsigned long));
		addr = ioremap(RS_RX2_RB, sizeof(unsigned long));
		break;
	default:
		return -ENODEV;
	}

	if(ctl) {
		if(!tap) {
			iowrite32(__arch__swab32(RX_IDLY_RST), ctl);
			barrier();

			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(HZ/1000);
			iowrite32(__arch__swab32(0), ctl);
			barrier();

			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(HZ/1000);
		}

		iowrite32(__arch__swab32(RX_IDLY_INC | 0x3FFFF), ctl);
		barrier();

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ/1000);
		iowrite32(__arch__swab32(RX_IDLY_INC), ctl);
		barrier();

		iounmap(ctl);
	}

	if(addr) {
		rb = __arch__swab32(ioread32(addr));
		iounmap(addr);
	}
       	snprintf(data, sizeof(data), 
			"RX1:%02d "
			"%c %c %c %c %c %c " 
			"%c %c %c %c %c " 
			"\n", 
			tap,
			(rb & (1<< 0) ? '1' : '0'),
			(rb & (1<< 1) ? '1' : '0'),
			(rb & (1<< 2) ? '1' : '0'),
			(rb & (1<< 3) ? '1' : '0'),
			(rb & (1<< 4) ? '1' : '0'),
			(rb & (1<< 5) ? '1' : '0'),
			(rb & (1<< 6) ? '1' : '0'),
			(rb & (1<< 7) ? '1' : '0'),
			(rb & (1<< 8) ? '1' : '0'),
			(rb & (1<< 9) ? '1' : '0'),
			(rb & (1<<10) ? '1' : '0')
		);

	/* copy kernel space buffer to user space*/
	err = copy_to_user(buff,data,strlen(data));
	if(err) {
		printk(KERN_ALERT MODULE_NAME " -> userspace copy failed!\n" );
		return -EFAULT;
	}

	*offp += strlen(data);
	return strlen(data);

}

ssize_t mrx_write(struct file *filep,const char *buff,size_t count,loff_t *offp )
{
	size_t i;
	int err;
	char data[PAGE_SIZE];
	char buf[11];
	int mask;
        unsigned int minor = iminor(filep->f_dentry->d_inode);
	void *ctl = NULL;
	size_t low = ((*offp) & 0xFFFFFFFF);

        printk(KERN_DEBUG MODULE_NAME ": write, minor=%d offp=%d\n", minor, low);

	switch(minor) {
	case 1:
		ctl = ioremap(RS_RX1_IDLY_CTL, sizeof(unsigned long));
		break;
	case 2:
		ctl = ioremap(RS_RX2_IDLY_CTL, sizeof(unsigned long));
		break;
	default:
		return -ENODEV;
	}

	if(!ctl) {
                printk(KERN_ALERT MODULE_NAME " -> RS_RX_IDLY_CTL remap failed!\n" );
                return -EFAULT;
        }

        memset(data, 0, sizeof(data));

        /* copy user space buffer to kernel space*/
        err = copy_from_user(data,buff,sizeof(data));
	if(err) {
                printk(KERN_ALERT MODULE_NAME " -> Kernel copy failed!\n" );
                return -EFAULT;
        }

	/* TODO: extract the mask from the user space data */
	mask = 0;
	i = sscanf(data, "%c %c %c %c %c %c %c %c %c %c %c",
			&buf[ 0], &buf[ 1], &buf[ 2], &buf[ 3], &buf[ 4], &buf[ 5], 
			&buf[ 6], &buf[ 7], &buf[ 8], &buf[ 9], &buf[10]  
		);
	if(i!= sizeof(buf)) {
		printk(KERN_ERR MODULE_NAME "-write: expected %d tokens, got %d\n", sizeof(buf), i);
		return -EIO;
	}
	for( i=0; i<sizeof(buf); i++ ) {
		if(buf[i]=='1') mask += (1<<i);
	}
 	mask &= 0x3FFFF;

	if(!low) {
		iowrite32(__arch__swab32(RX_IDLY_RST), ctl);
		barrier();

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ/1000);
		iowrite32(__arch__swab32(0), ctl);
		barrier();

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ/1000);
	}

	iowrite32(__arch__swab32(RX_IDLY_INC | mask), ctl);
	barrier();

	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(HZ/1000);
	iowrite32(__arch__swab32(RX_IDLY_INC), ctl);
	barrier();

	iounmap(ctl);

	*offp += strlen(data);
	return strlen(data);
}

static int mrx_init(void)
{
	printk(KERN_DEBUG MODULE_NAME ": initializing\n");
	major = register_chrdev(0,MODULE_NAME,&mrx_fops);
	if(!major){
		printk(KERN_ALERT MODULE_NAME ": failed to register\n");
		return -EIO;
	}
	printk(KERN_INFO MODULE_NAME ": major=%d\n", major);
        r = request_mem_region( RS_RX1_RB, sizeof(long), MODULE_NAME);
	if(!r) {
        	printk(KERN_ALERT MODULE_NAME ": cannot access FPGA registers\n");
		return -EFAULT;
	}
        r = request_mem_region( RS_RX1_IDLY_CTL, sizeof(long), MODULE_NAME);
	if(!r) {
        	printk(KERN_ALERT MODULE_NAME ": cannot access FPGA registers\n");
		return -EFAULT;
	}
	return 0;
}
static void mrx_cleanup(void)
{
        if(r) {
		release_mem_region(RS_RX1_RB, sizeof(long));
		release_mem_region(RS_RX1_IDLY_CTL, sizeof(long));
	}

	printk(KERN_DEBUG MODULE_NAME ": terminating\n");
	unregister_chrdev(major,MODULE_NAME);
	return ;
}

