#include <cerrno>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <cstdlib>

#include <sys/mount.h>
#include <sys/wait.h>

#include <boost/bind.hpp>
#include <boost/thread/thread.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>

#include <log4cpp/Category.hh>
#include <log4cpp/BasicConfigurator.hh>
#include <log4cpp/PropertyConfigurator.hh>

#include <tiffio.h>

#include "blkaddr.h"

namespace l4 = log4cpp;

class load
{
  typedef std::pair<std::string,std::string> strpair;
  static const loff_t BLKSIZ = 112 * sizeof(long);
  static char essence[sizeof(long)*2016*1080];
  static char mntcmds[5][PATH_MAX];

  size_t essence_width;
  size_t essence_length;
  l4::Category &log;
  TIFF *image;
  const std::string &dev;
  const std::string &mountpoint;
  std::ostream &sink;
  bool videokey;
  std::string desc;
  long pos;
  long len;
  long status;
  short suite;
  long item;
  std::string std;

  std::string pathname(const std::string &unc);
  void mount(const std::string &unc);
  int  read_video(loff_t offset, const size_t linelen);
  int  read_video(loff_t odd, loff_t even, const size_t linelen);
  void read_key(loff_t offset, const size_t linelen);
  void read_key(loff_t odd, loff_t even, const size_t linelen);
  void reader(const std::string &odd, const std::string &even);
  void reader(const std::string &label);
  void interlace (const short suite, const short id, const std::string pathname);
  void progressive (const short suite, const short id, const std::string pathname);

public:
	load(const std::string &dev, const std::string &mountpoint, std::ostream &sink) 
		: essence_width(0), essence_length(0), log(l4::Category::getInstance("load")), 
			image(NULL), dev(dev), mountpoint(mountpoint), sink(sink), 
			pos(0), len(~0), videokey(false), desc("---"), status(200), suite(0), item(0), std("unknown") {
				std::string ver = TIFFGetVersion();
				log.noticeStream() << ver.substr(0, ver.find('\n')) << l4::eol;
				TIFFSetErrorHandler(NULL);
				TIFFSetWarningHandler(NULL);
	};
	void operator()(std::istream &src);
	void operator() (const short suite, const short id, const std::string &unc);
};

char load::essence[sizeof(long)*2016*1080];

char load::mntcmds[5][PATH_MAX] = { "mount", "//menu/images", "/mnt/load", "-o", "ro,guest" };

std::string load::pathname(const std::string &unc)
{
	log.noticeStream() << " unc=" << unc << l4::eol;
	if(unc[0]!='\\' || unc[1]!='\\') {
		throw std::runtime_error("invalid unc");
	}

	std::string pathname(unc.substr(2));
	std::string::iterator itr;
	for(itr=pathname.begin(); itr!=pathname.end(); itr++) {
		if(*itr=='\\') *itr='/';
	}

	size_t hostpoint = pathname.find('/');
	size_t sharepoint = pathname.find('/', ++hostpoint);
	pathname.replace(0, sharepoint, mountpoint);

	log.noticeStream() << " unc=" << unc << " pathname=" << pathname << l4::eol;
	return pathname;
}

void load::mount(const std::string &unc)
{
	std::string path(unc);

	std::string::iterator itr;
	for(itr=path.begin(); itr!=path.end(); itr++) {
		if(*itr=='\\') *itr='/';
	}
	size_t hostpoint = path.find('/', 2);
	size_t sharepoint = path.find('/', ++hostpoint);
	path.erase(sharepoint);

	if(::strncmp(mntcmds[5],path.c_str(), path.size())==0) {
		log.noticeStream() << path << " already mounted"  << l4::eol;
		return;
	}

	log.debugStream() << " unmounting " << mountpoint << l4::eol;
	::umount(mountpoint.c_str());

	char buf[BUFSIZ];
	strncpy(buf, path.c_str(), sizeof(buf));
	log.noticeStream()	<< mntcmds[0] << " "
				<< buf << " "
				<< mntcmds[2] << " "
				<< mntcmds[3] << " "
				<< mntcmds[4] << " "
				<< l4::eol;

	errno = 0;
	pid_t pid = fork();
	switch(pid) {
	case -1:
		log.warnStream() << "fork: " << strerror(errno) << l4::eol;
		throw std::runtime_error("load fork failed");
		break;
	case 0:
		char * cmds[6];
		cmds[0] = mntcmds[0];
		cmds[1] = buf;
		cmds[2] = mntcmds[2];
		cmds[3] = mntcmds[3];
		cmds[4] = mntcmds[4];
		cmds[5] = NULL;

		l4::Category::shutdown();

		errno = 0;
		::execv("/bin/mount", cmds);
		::exit(errno ? -1 : 0);
		break;
	default:
		int status;
		errno = 0;
		::waitpid(pid, &status, 0);
		if(errno) {
			log.warnStream() << "waitpid: " << strerror(errno) << l4::eol;
			throw std::runtime_error("load wait failed");	
		}

		if(status) {
			log.warnStream() << "mount returns: " << status << l4::eol;
			throw std::runtime_error("load mount failed");	
		}

		log.noticeStream() << path << " mounted" << l4::eol;
		strncpy(mntcmds[1], path.c_str(), sizeof(mntcmds[1]));
		break;
	}
}

void load::operator()(std::istream &src)
{
	char buf[BUFSIZ];
	long itemID;
	short imageID;
	short suiteID;
	std::string unc;
	if(!src.good()) return;

	src.getline(buf, BUFSIZ, '=');
	if(!strstr(buf, "ITEM")) src.setstate(std::ios::failbit);
	src >> itemID;

	src.getline(buf, BUFSIZ, '=');
	if(!strstr(buf, "UNC")) src.setstate(std::ios::failbit);
	src.getline(buf, BUFSIZ, '\t');
	unc = buf;

	src.getline(buf, BUFSIZ, '=');
	if(!strstr(buf, "IMAGE")) src.setstate(std::ios::failbit);
	src >> imageID;

	src.getline(buf, BUFSIZ, '=');
	if(!strstr(buf, "STD")) src.setstate(std::ios::failbit);
	src >> std;

	src.getline(buf, BUFSIZ, '=');
	if(!strstr(buf, "SUITE")) src.setstate(std::ios::failbit);
	src >> suiteID;

	// complete the extraction 
	src.getline(buf, BUFSIZ);

	log.debugStream() << " JOB=load"
			<< " ITEM=" << itemID
			<< " UNC=" << unc
			<< " SUITE=" << suiteID
			<< " IMAGE=" << imageID
			<< " STD=" << std
			<< l4::eol;

	if((itemID != item) || (suiteID != suite)) {
		if(status != 202) {
			status = 202;
			log.noticeStream() << " new job: suite=" << suiteID << " item=" << itemID;
			suite = suiteID;
			item = itemID;
			try {
				mount(unc);
				if(std.find("720p") == std::string::npos ) {
					boost::thread thread(&load::interlace, this, suite, imageID, pathname(unc));
				} else {
					boost::thread thread(&load::progressive, this, suite, imageID, pathname(unc));
				}
			} catch(...) {
				log.errorStream() << "Cannot access " << unc << l4::eol;
				pos = 0;
				videokey = false;
				len = ~0;
				status = 401; // unauthorized
			}
		}
	}

	sink << "\tSUITE=" << suite;
	sink << "\tITEM=" << item;
	sink << "\tPOS=" << pos;
	sink << "\tLEN=" << len;
	sink << "\tSTD=" << std;
	sink << "\tPARTS=" << (videokey ? "videokey" : "video");
	sink << "\tDESC=" << desc;
	sink << "\tSTATUS=" << status;
	sink << std::endl;

	if(status != 200) {
		log.noticeStream() 
			<< " SUITE=" << suite
			<< " ITEM=" << item
			<< " POS=" << pos
			<< " LEN=" << len
			<< " STD=" << std
			<< " PARTS=" << (videokey ? "videokey" : "video")
			<< " DESC=" << desc
			<< " STATUS=" << status
			<< l4::eol;
	}
};

void load::operator() (const short suite, const short id, const std::string &unc)
{
	boost::thread thread(&load::interlace, this, suite, id, pathname(unc));
        while(!thread.timed_join(boost::posix_time::seconds(3))) {
		log.noticeStream() 
			<< " POS=" << pos
			<< " LEN=" << len
			<< " STD=" << std
			<< " PARTS=" << (videokey ? "videokey" : "video")
			<< " DESC=" << desc
			<< " STATUS=" << status
			<< l4::eol;
	}
	log.noticeStream() 
		<< " POS=" << pos
		<< " LEN=" << len
		<< " STD=" << std
		<< " PARTS=" << (videokey ? "videokey" : "video")
		<< " DESC=" << desc
		<< " STATUS=" << status
		<< l4::eol;
};

void load::interlace (const short suite, const short id, const std::string unc)
{
	pos = 0;
	videokey = false;
	// Open the TIFF file
	image = TIFFOpen(unc.c_str(), "r");
	if(image) {
		// Add time stamping variables
		boost::posix_time::ptime			start(boost::posix_time::microsec_clock::local_time()),
											fini;

		log.noticeStream() << " TIFF opened: " << unc << l4::eol;
		int pages = TIFFNumberOfDirectories(image);
		log.noticeStream() << " found" << pages << " pages" << l4::eol;

		TIFFGetField(image, TIFFTAG_IMAGEWIDTH, &essence_width);
		TIFFGetField(image, TIFFTAG_IMAGELENGTH, &essence_length);
		if(essence_length%2) ++essence_length; 
		len = pages * essence_length;
		status = 0;
		int i=1;
		do {
			std::stringstream odd;
			odd << suite 
				<< ":" << std::setfill('0') << std::setw(4) << id 
				<< ":" << std::setfill('0') << std::setw(4) << i++
				<< std::ends;
			std::stringstream even;
			even << suite 
				<< ":" << std::setfill('0') << std::setw(4) << id 
				<< ":" << std::setfill('0') << std::setw(4) << i++
				<< std::ends;
			reader(odd.str(),even.str());
		} while (TIFFReadDirectory(image)==1);

		// Close the file
		TIFFClose(image);
		status = 200;
		log.noticeStream() << " TIFF closed: " << unc << l4::eol;

		fini = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds

		// dump our findings on time used by each operation
		log.noticeStream() 
			<< "load::interlace(start time)          = "	<< start		<< l4::eol
			<< "load::interlace(finish time)         = "	<< fini			<< l4::eol
			<< "load::interlace(total time)          = "	<< (fini-start)	<< l4::eol
			<< l4::eol;
	} else {
		log.warnStream() << "Could not open " << unc << " for reading" << l4::eol;
		status = 404;
		if(image) TIFFClose(image);
	}
	image = NULL;
};

void load::progressive (const short suite, const short id, const std::string unc)
{
	pos = 0;
	videokey = false;

	// Open the TIFF file
	image = TIFFOpen(unc.c_str(), "r");
	if(image) {
		// Add time stamping variables
		boost::posix_time::ptime			start(boost::posix_time::microsec_clock::local_time()),
											fini;

		log.noticeStream() << " TIFF opened: " << unc << l4::eol;
		int pages = TIFFNumberOfDirectories(image);
		log.noticeStream() << " found" << pages << " pages" << l4::eol;

		TIFFGetField(image, TIFFTAG_IMAGEWIDTH, &essence_width);
		TIFFGetField(image, TIFFTAG_IMAGELENGTH, &essence_length);
		if(essence_length%2) ++essence_length;
		len = pages * essence_length;
		status = 0;
		int i=1;
		do {
			std::stringstream label;
			label << suite 
				<< ":" << std::setfill('0') << std::setw(4) << id 
				<< ":" << std::setfill('0') << std::setw(4) << i++
				<< std::ends;
			reader(label.str());
		} while (TIFFReadDirectory(image)==1);

		// Close the file
		TIFFClose(image);
		status = 200;
		log.noticeStream() << " TIFF closed: " << unc << l4::eol;

		fini = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds

		// dump our findings on time used by each operation
		log.noticeStream() 
			<< "load::progressive(start time)        = "	<< start		<< l4::eol
			<< "load::progressive(finish time)       = "	<< fini			<< l4::eol
			<< "load::progressive(total time)        = "	<< (fini-start)	<< l4::eol
			<< l4::eol;
	} else {
		log.warnStream() << "Could not open " << unc << " for reading" << l4::eol;
		status = 404;
		if(image) TIFFClose(image);
	}
	image = NULL;
};

int load::read_video(loff_t odd, loff_t even, const size_t linesiz)
{
	log.noticeStream() << " odd=" << odd << " even=" << even << l4::eol;
	if(!image) {
		log.warnStream() << "no image" << l4::eol;
		throw std::runtime_error("no image");
	}

	if(status) {
		if(!TIFFReadDirectory(image)) {
			log.warnStream() << "directory not found" << l4::eol;
			throw std::runtime_error("directory not found");
		}
	} else {
		status = 202;
	}

	char *ptr = NULL;
	TIFFGetField(image, TIFFTAG_IMAGEDESCRIPTION, &ptr); 
	desc = (ptr && strlen(ptr)) ? ptr : "---";

	std::ofstream cm(dev.c_str()); // cache memory stream
	if(cm.is_open()) {
		// Add time stamping variables
		boost::posix_time::ptime            start(boost::posix_time::microsec_clock::local_time()),
											lcur(start),
											cur,
											fini;

		// Add duration variables to profile with
		boost::posix_time::time_duration    t_read, t_write;

		const size_t NumberOfStrips = TIFFNumberOfStrips(image);
		const size_t StripSize = TIFFStripSize(image);

		long RowsPerStrip = 1;
		TIFFGetField(image, TIFFTAG_ROWSPERSTRIP, &RowsPerStrip); 

		log.noticeStream() << "read_video: "
			<< " # strips = "   << NumberOfStrips
			<< " strip size = "   << StripSize
			<< " rows per strip = "   << RowsPerStrip
			<< l4::eol;

		// Read essence from the file
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
		for(size_t y=0; y<NumberOfStrips; y++) {
			char *ptr = essence+(y*StripSize);
			if( TIFFReadEncodedStrip(image, y, ptr, StripSize) < 0) throw std::runtime_error("TIFFReadEncodedStrip");
			pos+=(essence_length/NumberOfStrips);
		}
		cur = boost::posix_time::microsec_clock::local_time();  // Fetch the current time in microseconds
		t_read = cur - lcur;                                   // accumulate duration taken for read

		// Write essence to the cache
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
		for(size_t strip=0; strip<NumberOfStrips; strip++) {
			for(size_t row=0; row<RowsPerStrip; row++) {
				size_t y = strip*RowsPerStrip + row;
				if(y>=essence_length) break;
				loff_t offset = y%2 ? even : odd;
				cm.seekp(offset+(y/2)*linesiz);
				cm.write(essence+((strip*StripSize)+(row*(StripSize/RowsPerStrip))), linesiz);
				cm.clear();
			}
		}
		cur = boost::posix_time::microsec_clock::local_time();  // Fetch the current time in microseconds
		t_write = cur - lcur;                                  // accumulate duration taken for write

		cm.close();

		fini = boost::posix_time::microsec_clock::local_time();     // Fetch the current time in microseconds

		// dump our findings on time used by each operation
		log.noticeStream() << "read_video: " 
			<< " time reading = "   << t_read
			<< " time writing = "   << t_write
			<< " total time = "   << (fini-start)
			<< l4::eol;
	} else {
		log.warnStream() << "cache memory not accessible" << l4::eol;
		throw std::runtime_error("no memory stream");
	}

	uint16 count;
	uint32 *diroff;
	if(!TIFFGetField(image, TIFFTAG_SUBIFD, &count, &diroff))
		throw std::runtime_error("no subdirectories");
	if(count > 1) {
		videokey = true;
	}

	for(size_t i=0; i<count; i++) {
		log.noticeStream() << "subdirectory " << i+1 << " offset=" << diroff[i] << l4::eol;
	}

	return count;
}

int load::read_video(loff_t offset, const size_t linesiz)
{
	log.noticeStream() << " offset=" << offset << l4::eol;
	if(!image) {
		log.warnStream() << "no image" << l4::eol;
		throw std::runtime_error("no image");
	}

	if(status) {
		if(!TIFFReadDirectory(image)) {
			log.warnStream() << "directory not found" << l4::eol;
			throw std::runtime_error("directory not found");
		}
	} else {
		status = 202;
	}

	char *ptr = NULL;
	TIFFGetField(image, TIFFTAG_IMAGEDESCRIPTION, &ptr); 
	desc = (ptr && strlen(ptr)) ? ptr : "---";

	std::ofstream cm(dev.c_str()); // cache memory stream
	if(cm.is_open()) {
		// Add time stamping variables
		boost::posix_time::ptime            start(boost::posix_time::microsec_clock::local_time()),
											lcur(start),
											cur,
											fini;

		// Add duration variables to profile with
		boost::posix_time::time_duration    t_read, t_write;

		const size_t NumberOfStrips = TIFFNumberOfStrips(image);
		const size_t StripSize = TIFFStripSize(image);

		long RowsPerStrip = 1;
		TIFFGetField(image, TIFFTAG_ROWSPERSTRIP, &RowsPerStrip); 

		log.noticeStream() << "read_video: "
			<< " # strips = "   << NumberOfStrips
			<< " strip size = "   << StripSize
			<< " rows per strip = "   << RowsPerStrip
			<< l4::eol;

		// Read essence from the file
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
		for(size_t y=0; y<NumberOfStrips; y++) {
			char *ptr = essence+(y*StripSize);
			if( TIFFReadEncodedStrip(image, y, ptr, StripSize) < 0) throw std::runtime_error("TIFFReadEncodedStrip");
			pos+=(essence_length/NumberOfStrips);
		}
		cur = boost::posix_time::microsec_clock::local_time();  // Fetch the current time in microseconds
		t_read = cur - lcur;                                   // accumulate duration taken for read

		// Write essence to the cache
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
		for(size_t strip=0; strip<NumberOfStrips; strip++) {
			for(size_t row=0; row<RowsPerStrip; row++) {
				size_t y = strip*RowsPerStrip + row;
				if(y>=essence_length) break;
				cm.seekp(offset+y*linesiz);
				cm.write(essence+((strip*StripSize)+(row*(StripSize/RowsPerStrip))), linesiz);
				cm.clear();
			}
		}
		cur = boost::posix_time::microsec_clock::local_time();  // Fetch the current time in microseconds
		t_write = cur - lcur;                                  // accumulate duration taken for write

		cm.close();

		fini = boost::posix_time::microsec_clock::local_time();     // Fetch the current time in microseconds

		// dump our findings on time used by each operation
		log.noticeStream() << "read_video: "
			<< " time reading = " << t_read
			<< " time writing = " << t_write
			<< " total time = "  << (fini-start)
			<< l4::eol;
	} else {
		log.warnStream() << "cache memory not accessible" << l4::eol;
		throw std::runtime_error("no memory stream");
	}

	uint16 count;
	uint32 *diroff;
	if(!TIFFGetField(image, TIFFTAG_SUBIFD, &count, &diroff))
		throw std::runtime_error("no subdirectories");
	if(count > 1) {
		videokey = true;
	}

	for(size_t i=0; i<count; i++) {
		log.noticeStream() << "subdirectory " << i+1 << " offset=" << diroff[i] << l4::eol;
	}

	return count;
}

void load::read_key(loff_t offset, const size_t linesiz)
{
	log.noticeStream() << " offset=" << offset << l4::eol;
	if(!image) {
		log.warnStream() << "no image" << l4::eol;
		throw std::runtime_error("no image");
	}

	uint16 count;
	uint32 *diroff;
	TIFFGetField(image, TIFFTAG_SUBIFD, &count, &diroff);
	if(!count) throw std::runtime_error("no subdirectories");

	if(!TIFFSetSubDirectory(image, diroff[count-1])) {
		videokey = false;
		log.warnStream() << "subdirectory " << " [" << diroff[count-1] << "] not found" << l4::eol;
		throw std::runtime_error("subdirectory not found");
	}

	std::ofstream cm(dev.c_str()); // cache memory stream
	if(cm.is_open()) {
		// Add time stamping variables
		boost::posix_time::ptime            start(boost::posix_time::microsec_clock::local_time()),
											lcur(start),
											cur,
											fini;

		// Add duration variables to profile with
		boost::posix_time::time_duration    t_read, t_write;

		const size_t NumberOfStrips = TIFFNumberOfStrips(image);
		const size_t StripSize = TIFFStripSize(image);

		long RowsPerStrip = 1;
		TIFFGetField(image, TIFFTAG_ROWSPERSTRIP, &RowsPerStrip); 

		log.noticeStream() << "read_video: "
			<< " # strips = "   << NumberOfStrips
			<< " strip size = "   << StripSize
			<< " rows per strip = "   << RowsPerStrip
			<< l4::eol;

		// Read essence from the file
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
		for(size_t y=0; y<NumberOfStrips; y++) {
			char *ptr = essence+(y*StripSize);
			if( TIFFReadEncodedStrip(image, y, ptr, StripSize) < 0) throw std::runtime_error("TIFFReadEncodedStrip");
			pos+=(essence_length/NumberOfStrips);
		}
		cur = boost::posix_time::microsec_clock::local_time();  // Fetch the current time in microseconds
		t_read = cur - lcur;                                   // accumulate duration taken for read

		// Write essence to the cache
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
		for(size_t strip=0; strip<NumberOfStrips; strip++) {
			for(size_t row=0; row<RowsPerStrip; row++) {
				size_t y = strip*RowsPerStrip + row;
				if(y>=essence_length) break;
				cm.seekp(offset+y*linesiz);
				cm.write(essence+((strip*StripSize)+(row*(StripSize/RowsPerStrip))), linesiz);
				cm.clear();
			}
		}
		cur = boost::posix_time::microsec_clock::local_time();  // Fetch the current time in microseconds
		t_write = cur - lcur;                                  // accumulate duration taken for write

		cm.close();

		fini = boost::posix_time::microsec_clock::local_time();     // Fetch the current time in microseconds

		// dump our findings on time used by each operation
		log.noticeStream() << "read_key: " 
			<< " time reading = "   << t_read
			<< " time writing = "   << t_write
			<< " total time = "   << (fini-start)
			<< l4::eol;
	} else {
		log.warnStream() << "cache memory not accessible" << l4::eol;
		throw std::runtime_error("no memory stream");
	}

}

void load::read_key(loff_t odd, loff_t even, const size_t linesiz)
{
	log.noticeStream() << " odd=" << odd << " even=" << even << l4::eol;
	if(!image) {
		log.warnStream() << "no image" << l4::eol;
		throw std::runtime_error("no image");
	}

	uint16 count;
	uint32 *diroff;
	TIFFGetField(image, TIFFTAG_SUBIFD, &count, &diroff);
	if(!count) throw std::runtime_error("no subdirectories");

	if(!TIFFSetSubDirectory(image, diroff[count-1])) {
		videokey = false;
		log.warnStream() << "subdirectory " << " [" << diroff[count-1] << "] not found" << l4::eol;
		throw std::runtime_error("subdirectory not found");
	}

	std::ofstream cm(dev.c_str()); // cache memory stream
	if(cm.is_open()) {
		// Add time stamping variables
		boost::posix_time::ptime            start(boost::posix_time::microsec_clock::local_time()),
											lcur(start),
											cur,
											fini;

		// Add duration variables to profile with
		boost::posix_time::time_duration    t_read, t_write;

		const size_t NumberOfStrips = TIFFNumberOfStrips(image);
		const size_t StripSize = TIFFStripSize(image);

		long RowsPerStrip = 1;
		TIFFGetField(image, TIFFTAG_ROWSPERSTRIP, &RowsPerStrip); 

		log.noticeStream() << "read_video: "
			<< " # strips = "   << NumberOfStrips
			<< " strip size = "   << StripSize
			<< " rows per strip = "   << RowsPerStrip
			<< l4::eol;

		// Read essence from the file
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
		for(size_t y=0; y<NumberOfStrips; y++) {
			char *ptr = essence+(y*StripSize);
			if( TIFFReadEncodedStrip(image, y, ptr, StripSize) < 0) throw std::runtime_error("TIFFReadEncodedStrip");
			pos+=(essence_length/NumberOfStrips);
		}
		cur = boost::posix_time::microsec_clock::local_time();  // Fetch the current time in microseconds
		t_read = cur - lcur;                                   // accumulate duration taken for read

		// Write essence to the cache
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
		for(size_t strip=0; strip<NumberOfStrips; strip++) {
			for(size_t row=0; row<RowsPerStrip; row++) {
				size_t y = strip*RowsPerStrip + row;
				if(y>=essence_length) break;
				loff_t offset = y%2 ? even : odd;
				cm.seekp(offset+(y/2)*linesiz);
				cm.write(essence+((strip*StripSize)+(row*(StripSize/RowsPerStrip))), linesiz);
				cm.clear();
			}
		}
		cur = boost::posix_time::microsec_clock::local_time();  // Fetch the current time in microseconds
		t_write = cur - lcur;                                  // accumulate duration taken for write

		cm.close();

		fini = boost::posix_time::microsec_clock::local_time();     // Fetch the current time in microseconds

		// dump our findings on time used by each operation
		log.noticeStream() << "read_key: " 
			<< " time reading = "   << t_read
			<< " time writing = "   << t_write
			<< " total time = "   << (fini-start)
			<< l4::eol;
	} else {
		log.warnStream() << "cache memory not accessible" << l4::eol;
		throw std::runtime_error("no memory stream");
	}

}

void load::reader(const std::string &odd, const std::string &even)
{
	blkaddr ba;
	int num_subdirectories = 0;
	log.noticeStream() << "odd=" << odd << " even=" << even << l4::eol;

	try {
		std::string v(":v");
		size_t numblks = essence_length * (1+essence_width/112);
                loff_t oddset  = ba.assign(odd.c_str()+v, 1+(numblks/2), (numblks/2));
                loff_t evenset = ba.assign(even.c_str()+v, 1+(numblks/2), (numblks/2));
		log.noticeStream() << "read image odd=" << oddset << " even=" << evenset << l4::eol;
		switch(numblks) {
		case 2*9720:
			num_subdirectories = read_video(oddset*BLKSIZ, evenset*BLKSIZ, 18*BLKSIZ);
			break;
		case 8640:
			num_subdirectories = read_video(oddset*BLKSIZ, evenset*BLKSIZ, 12*BLKSIZ);
			break;
		case 2*2016:
			num_subdirectories = read_video(oddset*BLKSIZ, evenset*BLKSIZ, 7*BLKSIZ);
			break;
		case 2*1701:
		case 2*1708:
			num_subdirectories = read_video(oddset*BLKSIZ, evenset*BLKSIZ, 7*BLKSIZ);
			break;
		default:
			log.errorStream() << "unsupported format, numblks=" << numblks << l4::eol;
			throw std::runtime_error("unsupported format");
			break;
		}
	} catch(...) {
		log.warnStream() << "essence not loaded" << l4::eol;
	}

	if(num_subdirectories > 1 ) try {
		len += essence_length; // a key means the load becomes longer
		std::string k(":k");
		size_t numblks = (1+essence_width/112)*essence_length;
                loff_t oddset  = ba.assign(odd.c_str()+k, 1+(numblks/2), (numblks/2));
                loff_t evenset = ba.assign(even.c_str()+k, 1+(numblks/2), (numblks/2));
		log.noticeStream() << "read key odd=" << oddset << " even=" << evenset << l4::eol;
		switch(numblks) {
		case 2*9720:
			read_key(oddset*BLKSIZ, evenset*BLKSIZ, 18*BLKSIZ);
			break;
		case 8640:
			read_key(oddset*BLKSIZ, evenset*BLKSIZ, 12*BLKSIZ);
			break;
		case 2*2016:
			read_key(oddset*BLKSIZ, evenset*BLKSIZ, 7*BLKSIZ);
			break;
		case 2*1701:
		case 2*1708:
			read_key(oddset*BLKSIZ, evenset*BLKSIZ, 7*BLKSIZ);
			break;
		default:
			log.errorStream() << "unsupported format, numblks=" << numblks << l4::eol;
			throw std::runtime_error("unsupported format");
			break;
		}
	} catch(...) {
		log.noticeStream() << "key not loaded" << l4::eol;
	}
}

void load::reader(const std::string &label)
{
	blkaddr ba;
	int num_subdirectories = 0;
	log.noticeStream() << "label=" << label << l4::eol;

	try {
		std::string v(":v");
		size_t numblks = essence_length * (1+essence_width/112);
                loff_t offset  = ba.assign(label.c_str()+v, 1+numblks, numblks);
		log.noticeStream() << "read image offset=" << offset << l4::eol;
		switch(numblks) {
		case 2*9720:
			num_subdirectories = read_video(offset*BLKSIZ, 18*BLKSIZ);
			break;
		case 8640:
			num_subdirectories = read_video(offset*BLKSIZ, 12*BLKSIZ);
			break;
		case 2*2016:
			num_subdirectories = read_video(offset*BLKSIZ, 7*BLKSIZ);
			break;
		case 2*1701:
		case 2*1708:
			num_subdirectories = read_video(offset*BLKSIZ, 7*BLKSIZ);
			break;
		default:
			log.errorStream() << "unsupported format, numblks=" << numblks << l4::eol;
			throw std::runtime_error("unsupported format");
			break;
		}
	} catch(...) {
		log.warnStream() << "essence not loaded" << l4::eol;
	}

	if(num_subdirectories > 1 ) try {
		len += essence_length; // a key means the load becomes longer
		std::string k(":k");
		size_t numblks = (1+essence_width/112)*essence_length;
                loff_t offset  = ba.assign(label.c_str()+k, 1+numblks, numblks);
		log.noticeStream() << "read key offset=" << offset << l4::eol;
		switch(numblks) {
		case 2*9720:
			read_key(offset*BLKSIZ, 18*BLKSIZ);
			break;
		case 8640:
			read_key(offset*BLKSIZ, 12*BLKSIZ);
			break;
		case 2*2016:
			read_key(offset*BLKSIZ, 12*BLKSIZ);
			break;
		case 2*1701:
		case 2*1708:
			read_key(offset*BLKSIZ, 12*BLKSIZ);
			break;
		default:
			log.errorStream() << "unsupported format, numblks=" << numblks << l4::eol;
			throw std::runtime_error("unsupported format");
			break;
		}
	} catch(...) {
		log.noticeStream() << "key not loaded" << l4::eol;
	}
}

