#include <cerrno>
#include <fstream>
#include <iostream>
#include <stdexcept>

#include <sys/mount.h>
#include <sys/wait.h>

#include <boost/bind.hpp>
#include <boost/thread/thread.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>

#include <log4cpp/Category.hh>
#include <log4cpp/BasicConfigurator.hh>
#include <log4cpp/PropertyConfigurator.hh>

#include <tiffio.h>

#include "blkaddr.h"

namespace l4 = log4cpp;

class save
{
  typedef std::pair<std::string,std::string> strpair;

  static const size_t thumb_width  = 84;
  static const size_t thumb_length = 48;
  static const loff_t BLKSIZ = 112 * sizeof(long);
  static char essence[sizeof(uint32)*2016*1080];
  static const char *make;
  static const char *model;
  static const char *ver;
  static char mntcmds[5][PATH_MAX];

  size_t essence_width;
  size_t essence_length;
  l4::Category &log;
  TIFF *image;
  const std::string &dev;
  const std::string &mountpoint;
  long rowsperstrip;
  short compression;
  std::ostream &sink;
  bool videokey;
  std::string desc;
  long pos;
  long len;
  long status;
  std::string suite;
  long item;
  const std::string inprocess;
  std::string std;

  unsigned char ycbcr2r(char *ptr);
  unsigned char ycbcr2g(char *ptr);
  unsigned char ycbcr2b(char *ptr);
  std::string pathname(const std::string &unc);
  void mount(const std::string &unc);
  void write_thumb(const size_t hBias, const size_t hGain, const size_t vBias, const size_t vGain, loff_t offset, const size_t len);
  void write_video(loff_t offset, const std::string &parts, const size_t linesiz);
  void write_video(loff_t odd, loff_t even, const std::string &parts, const size_t linesiz);
  void write_key(loff_t offset, const size_t linesiz);
  void write_key(loff_t odd, loff_t even, const size_t linesiz);
  void writer(const strpair &label, const std::string &parts);
  void handler (const std::vector<strpair> labels, const std::string parts, const std::string pathname);

public:
	save(const std::string &dev, const std::string &mountpoint, const long rowsperstrip, bool lzw, std::ostream &sink) 
		: essence_width(0), essence_length(0), log(l4::Category::getInstance("save")), 
			image(NULL), dev(dev), mountpoint(mountpoint), rowsperstrip(rowsperstrip), compression(lzw?COMPRESSION_LZW:COMPRESSION_NONE), sink(sink), 
			videokey(false), pos(0), len(~0), status(0), suite("1"), item(0), inprocess(".tmp"), std("unknown") {
				std::string ver = TIFFGetVersion();
				log.noticeStream() << ver.substr(0, ver.find('\n')) << l4::eol;
				TIFFSetErrorHandler(NULL);
				TIFFSetWarningHandler(NULL);
	};
	void operator()(std::istream &src);
	void operator() (const std::vector< std::pair<std::string,std::string> > &labels, const std::string &parts, const std::string &unc);
};

char save::essence[sizeof(long)*2016*1080];
const char *save::make = "Thomson GVG";
const char *save::model = "Kayenne";
const char *save::ver = "Missouri v0.0.0";

char save::mntcmds[5][PATH_MAX] = { "mount", "//menu/images", "/mnt/save", "-o", "rw,guest" };

std::string save::pathname(const std::string &unc)
{
	log.noticeStream() << " unc=" << unc << l4::eol;
	if(unc[0]!='\\' || unc[1]!='\\') {
		throw std::runtime_error("invalid unc");
	}

	std::string pathname(unc.substr(2));
	std::string::iterator itr;
	for(itr=pathname.begin(); itr!=pathname.end(); itr++) {
		if(*itr=='\\') *itr='/';
	}

	size_t hostpoint = pathname.find('/');
	size_t sharepoint = pathname.find('/', ++hostpoint);
	pathname.replace(0, sharepoint, mountpoint);

	log.noticeStream() << " unc=" << unc << " pathname=" << pathname << l4::eol;
	return pathname;
}

void save::mount(const std::string &unc)
{
	std::string path(unc);
	std::string::iterator itr;
	for(itr=path.begin(); itr!=path.end(); itr++) {
		if(*itr=='\\') *itr='/';
	}
	size_t hostpoint = path.find('/', 2);
	size_t sharepoint = path.find('/', ++hostpoint);
	path.erase(sharepoint);

	if(::strncmp(mntcmds[5],path.c_str(), path.size())==0) {
		log.noticeStream() << path << " already mounted"  << l4::eol;
		return;
	}

	log.debugStream() << " unmounting " << mountpoint << l4::eol;
	::umount(mountpoint.c_str());

	char buf[BUFSIZ];
	strncpy(buf,path.c_str(),sizeof(buf));
        log.noticeStream()      << mntcmds[0] << " "
                                << buf << " "
                                << mntcmds[2] << " "
                                << mntcmds[3] << " "
                                << mntcmds[4] << " "
                                << l4::eol;

	errno = 0;
	pid_t pid = fork();
	switch(pid) {
	case -1:
		log.warnStream() << "fork: " << strerror(errno) << l4::eol;
		throw std::runtime_error("save fork failed");
		break;
	case 0:
		char * cmds[6];
		cmds[0] = mntcmds[0];
		cmds[1] = buf;
		cmds[2] = mntcmds[2];
		cmds[3] = mntcmds[3];
		cmds[4] = mntcmds[4];
		cmds[5] = NULL;

                l4::Category::shutdown();

		errno = 0;
		::execv("/bin/mount", cmds);
		exit(errno ? -1 : 0);
		break;
	default:
		int status;
		errno = 0;
		::waitpid(pid, &status, 0);
		if(errno) {
			log.warnStream() << "waitpid: " << strerror(errno) << l4::eol;
			throw std::runtime_error("save wait failed");
		}

		if(status) {
                	log.warnStream() << "mount returns: " << status << l4::eol;
			throw std::runtime_error("save mount failed");
		}

		log.noticeStream() << path << " mounted" << l4::eol;
		strncpy(mntcmds[1],path.c_str(),sizeof(mntcmds[1]));
		break;
	}
}

void save::operator()(std::istream &src)
{
	char buf[BUFSIZ];
	long itemID;
	std::string imageID;
	std::string suiteID;
	std::string parts;
	std::string unc;
	if(!src.good()) return;

	src.getline(buf, BUFSIZ, '=');
	if(!strstr(buf, "ITEM")) src.setstate(std::ios::failbit);
	src >> itemID;

	src.getline(buf, BUFSIZ, '=');
	if(!strstr(buf, "UNC")) src.setstate(std::ios::failbit);
	src.getline(buf, BUFSIZ, '\t');
	unc = buf;

	src.getline(buf, BUFSIZ, '=');
	if(!strstr(buf, "IMAGE")) src.setstate(std::ios::failbit);
	src >> imageID;

	src.getline(buf, BUFSIZ, '=');
	if(!strstr(buf, "STD")) src.setstate(std::ios::failbit);
	src >> std;

	src.getline(buf, BUFSIZ, '=');
	if(!strstr(buf, "PARTS")) src.setstate(std::ios::failbit);
	src >> parts;

	src.getline(buf, BUFSIZ, '=');
	if(!strstr(buf, "DESC")) src.setstate(std::ios::failbit);
	src.getline(buf, BUFSIZ, '\t');
	desc = buf;

	src.getline(buf, BUFSIZ, '=');
	if(!strstr(buf, "SUITE")) src.setstate(std::ios::failbit);
	src >> suiteID;

	// complete the extraction 
	src.getline(buf, BUFSIZ);

	log.debugStream() << " JOB=save"
			<< " ITEM=" << itemID
			<< " UNC=" << imageID
			<< " SUITE=" << suiteID
			<< " IMAGE=" << imageID
			<< l4::eol;
	
	if((itemID != item) || (suiteID != suite)) {
		if(status != 202) {
			status = 202;
	                blkaddr ba;
			std::vector<strpair> labels;
			log.noticeStream() << " new job: suite=" << suiteID << " item=" << itemID;
			suite = suiteID;
			item = itemID;
			try {
				mount(unc);
				log.debugStream() << " list: suite=" << suiteID << " image=" << imageID;
		                if(std.find("720p") == std::string::npos) {
					ba.interlace(labels, boost::lexical_cast<unsigned>(suiteID), boost::lexical_cast<unsigned>(imageID));
				} else {
					ba.progressive(labels, boost::lexical_cast<unsigned>(suiteID), boost::lexical_cast<unsigned>(imageID));
				}
				boost::thread thread(&save::handler, this, labels, parts, pathname(unc));
			} catch(std::runtime_error e) {
				log.errorStream() << "runtime error: " << e.what() << l4::eol;
				pos = 0;
				len = ~0;
				status = 505; // internal error
			} catch(...) {
				log.errorStream() << "Cannot access " << unc << l4::eol;
				pos = 0;
				len = ~0;
				status = 401; // unauthorized
			}
		}
	}

	if(!suite.empty()) sink << "\tSUITE=" << suite;
	if(item)  {
		sink << "\tITEM=" << item;
		sink << "\tPOS=" << pos;
		sink << "\tLEN=" << (essence_length ? (len*essence_length) : ~0);
		sink << "\tSTD=" << std;
		sink << "\tPARTS=" << (videokey ? "videokey" : "video");
		sink << "\tDESC=" << desc;
		sink << "\tSTATUS=" << status;
		sink << std::endl;
	}
	if(status != 200)  {
		log.noticeStream() 
			<< " SUITE=" << suite
			<< " ITEM=" << item
			<< " POS=" << pos
			<< " LEN=" << (essence_length ? (len*essence_length) : ~0)
			<< " STD=" << std
			<< " PARTS=" << (videokey ? "videokey" : "video")
			<< " DESC=" << desc
			<< " STATUS=" << status
			<< l4::eol;
	}
};

void save::operator() (const std::vector< std::pair<std::string,std::string> > &labels, const std::string &parts, const std::string &unc)
{
	boost::thread thread(&save::handler, this, labels, parts, pathname(unc));
	while(!thread.timed_join(boost::posix_time::seconds(3))) {
		log.noticeStream() 
			<< " POS=" << pos
			<< " LEN=" << (essence_length ? (len*essence_length) : ~0)
			<< " STD=" << std
			<< " PARTS=" << (videokey ? "videokey" : "video")
			<< " DESC=" << desc
			<< " STATUS=" << status
			<< l4::eol;
	}
	log.noticeStream() 
		<< " POS=" << pos
		<< " LEN=" << (essence_length ? (len*essence_length) : ~0)
		<< " STD=" << std
		<< " PARTS=" << (videokey ? "videokey" : "video")
		<< " DESC=" << desc
		<< " STATUS=" << status
		<< l4::eol;
};

void save::handler (const std::vector<strpair> labels, const std::string parts, const std::string pathname)
{
	if(labels.size()==0) {
		log.warnStream() << "no labels" << l4::eol;
		pos = 0;
		status = 500;
		return;
	}

	std::vector<strpair>::const_iterator itr;
	for(itr=labels.begin(); itr!=labels.end(); itr++) {
		log.noticeStream() << (*itr).first << ":" << (*itr).second << " -> " << pathname << l4::eol;
	}

	pos = 0;

	// Open the TIFF file
	image = TIFFOpen((pathname+inprocess).c_str(), "w");
	if(image) {

		// Add time stamping variables
		boost::posix_time::ptime			start(boost::posix_time::microsec_clock::local_time()),
											lcur(start),
											cur,
											fini;

		// Add duration variables to profile with
		boost::posix_time::time_duration	t_creat, t_rename;

		log.noticeStream() << " TIFF opened: " << pathname << inprocess << l4::eol;
		len = labels.size();
		if(parts=="videokey") len *= 2;
		status = 202;
		for_each(labels.begin(), labels.end(), boost::bind(&save::writer, this, _1, parts));

		// Close the file
		TIFFClose(image);
		log.noticeStream() << " TIFF closed: " << pathname << inprocess << l4::eol;

		cur = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds
		t_creat += cur - start;					// duration taken to create the TIFF (read/write)
		lcur = cur;						// Save the time for next calculation

		errno = 0;

		::rename((pathname+inprocess).c_str(), pathname.c_str());
		if(errno) {
			log.warnStream() << "rename error " << errno << " "  << strerror(errno) << l4::eol;
			throw std::runtime_error("rename");
		}

		fini = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds
		t_rename += fini - lcur;				// duration taken to rename the TIFF

		status = 200;
		log.noticeStream() << " TIFF renamed: " << pathname << l4::eol;

		// dump our findings on time used by each operation
		log.noticeStream() 
			<< "save::handler(start time)             = "	<< start		<< l4::eol
			<< "save::handler(TIFF creation time)     = "	<< t_creat		<< l4::eol
			<< "save::handler(TIFF rename time)       = "	<< t_rename		<< l4::eol
			<< "save::handler(finish time)            = "	<< fini			<< l4::eol
			<< "save::handler(total time)             = "	<< (fini-start)	<< l4::eol
			<< l4::eol;

	} else {
		log.warnStream() << "Could not open " << pathname << " for writing" << l4::eol;
		status = 404;
		if(image) TIFFClose(image);
	}
	image = NULL;
};

/// \brief extracts an 8bit red value from a 10bit ycbcr data stream
/// \param pos points to a 10bit luminance value, which is preceeded by a Cr value
/// \returns the derived an 8bit (0-255) red value
/// \note this is an approximate (CPU conserving) integer math based implementation,
/// suitable for use in generating thumbnails, but not suitable for use in essence conversion
unsigned char save::ycbcr2r(char *pos)
{
	long y = *(unsigned short *)pos;
	y /= 4;
	if(y < 0) y=0;
	if(y > 255) y=255;

	long cr = *(unsigned short *)(pos-sizeof(short));
	cr /= 4;
	if(cr < 0) cr=0;
	if(cr > 255) cr=255;

	long r = ( 298 * y + 408 * cr ) / 256 - 222;
	if(r < 16) return 16;
	if(r > 255) return 255;
	return r;
}

/// \brief extracts an 8bit green value from a 10bit ycbcr data stream
/// \param pos points to a 10bit luminance value, which is preceeded by a Cr value, and followed by a Cb value
/// \returns the derived an 8bit (0-255) green value
/// \note this is an approximate (CPU conserving) integer math based implementation,
/// suitable for use in generating thumbnails, but not suitable for use in essence conversion
unsigned char save::ycbcr2g(char *pos)
{
	long y = *(unsigned short *)pos;
	y /= 4;
	if(y < 0) y = 0;
	if(y > 255) y = 255;

	long cb = *(unsigned short *)(pos+sizeof(short));
	cb /= 4;
	if(cb < 0) cb=0;
	if(cb > 255) cb=255;

	long cr = *(unsigned short *)(pos-sizeof(short));
	cr /= 4;
	if(cr < 0) cr=0;
	if(cr > 255) cr=255;

	long g = ( 298 * y - 100 * cb - 208 * cr ) / 256 + 135;
	if(g < 16) return 16;
	if(g > 255) return 255;
	return g;
}

/// \brief extracts an 8bit blue value from a 10bit ycbcr data stream
/// \param pos points to a 10bit luminance value, which is followed by a Cb value
/// \returns the derived an 8bit (0-255) blue value
/// \note this is an approximate (CPU conserving) integer math based implementation,
/// suitable for use in generating thumbnails, but not suitable for use in essence conversion
unsigned char save::ycbcr2b(char *pos)
{
	short y = *(unsigned short *)pos;
	y /= 4;
	if(y < 0) y=0;
	if(y > 255) y=255;

	short cb = *(unsigned short *)(pos+sizeof(short));
	cb /= 4;
	if(cb < 0) cb=0;
	if(cb > 255) cb=255;

	long b = ( 298 * y + 516 * cb ) / 256 - 276;
	if(b < 16) return 16;
	if(b > 255) return 255;
	return b;
}

// Progressive video write mode
void save::write_video(loff_t offset, const std::string &parts, const size_t linesiz)
{
	log.noticeStream() << " offset=" << offset << l4::eol;
	if(!image) {
		log.warnStream() << "no image" << l4::eol;
		throw std::runtime_error("no image");
	}

	TIFFCreateDirectory(image);
	TIFFSetField(image, TIFFTAG_MAKE, make);
	TIFFSetField(image, TIFFTAG_MODEL, model);
	TIFFSetField(image, TIFFTAG_SOFTWARE, ver);

	// We need to set some values for basic tags before we can add any data
	TIFFSetField(image, TIFFTAG_IMAGEWIDTH, essence_width);
	TIFFSetField(image, TIFFTAG_IMAGELENGTH, essence_length);
	TIFFSetField(image, TIFFTAG_BITSPERSAMPLE, 16);
	TIFFSetField(image, TIFFTAG_SAMPLESPERPIXEL, 2);
	TIFFSetField(image, TIFFTAG_EXTRASAMPLES, EXTRASAMPLE_UNSPECIFIED);
	TIFFSetField(image, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
	TIFFSetField(image, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
	TIFFSetField(image, TIFFTAG_ROWSPERSTRIP, rowsperstrip);
	TIFFSetField(image, TIFFTAG_COMPRESSION, compression);

	const char *ptr = desc.c_str();
        TIFFSetField(image, TIFFTAG_IMAGEDESCRIPTION, ptr);

	std::ifstream cm(dev.c_str()); // cache memory stream
	if(cm.is_open()) {
		// Add time stamping variables
		boost::posix_time::ptime			start(boost::posix_time::microsec_clock::local_time()),
											lcur(start),
											cur,
											fini;


		const size_t NumberOfStrips = TIFFNumberOfStrips(image);
		const size_t StripSize = TIFFStripSize(image);
	        log.noticeStream() << "write_video: " 
			<< " # strips = "   << NumberOfStrips
			<< " strip size = "   << StripSize
			<< " rows per strip = "   << rowsperstrip
			<< l4::eol;

		// Read essence from cache
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
                for(size_t strip=0; strip<NumberOfStrips; strip++) {
                        for(size_t row=0; row<rowsperstrip; row++) {
                                size_t y = strip*rowsperstrip + row;
				cm.seekg(offset+y*linesiz);
                                cm.read(essence+((strip*StripSize)+(row*(StripSize/rowsperstrip))), linesiz);
                        }
                }
		cur = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds
		boost::posix_time::time_duration t_read = cur - lcur;	// accumulate duration taken for read

		// Write essence to the file
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
		for(size_t y=0; y<NumberOfStrips; y++) {
			char *ptr = essence+(y*StripSize);
			if( TIFFWriteEncodedStrip(image, y, ptr, StripSize) < 0) throw std::runtime_error("TIFFWriteEncodedStrip");
			pos+=rowsperstrip;
		}
		cur = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds
		boost::posix_time::time_duration t_write = cur - lcur; 	// accumulate duration taken for write

		cm.close();

		fini = boost::posix_time::microsec_clock::local_time();		// Fetch the current time in microseconds

		// dump our findings on time used by each operation
	        log.noticeStream() << "write_video: " 
			<< " time reading = " << t_read
			<< " time writing = " << t_write 
			<< " total time = " << (fini-start)
			<< l4::eol;
	}

	uint16 count = 1;
	uint32 diroff[2] = { 0, 0 };
	if(parts=="videokey") count = 2;
	TIFFSetField(image, TIFFTAG_SUBIFD, count, diroff);

	if( TIFFWriteDirectory(image) != 1 ) throw std::runtime_error("TIFFWriteDirectory");
}

// Interlaced video write mode
void save::write_video(loff_t odd, loff_t even, const std::string &parts, const size_t linesiz)
{
	log.noticeStream() << " odd=" << odd << " even=" << even << l4::eol;
	if(!image) {
		log.warnStream() << "no image" << l4::eol;
		throw std::runtime_error("no image");
	}

	TIFFCreateDirectory(image);
	TIFFSetField(image, TIFFTAG_MAKE, make);
	TIFFSetField(image, TIFFTAG_MODEL, model);
	TIFFSetField(image, TIFFTAG_SOFTWARE, ver);

	// We need to set some values for basic tags before we can add any data
	TIFFSetField(image, TIFFTAG_IMAGEWIDTH, essence_width);
	TIFFSetField(image, TIFFTAG_IMAGELENGTH, essence_length);
	TIFFSetField(image, TIFFTAG_BITSPERSAMPLE, 16);
	TIFFSetField(image, TIFFTAG_SAMPLESPERPIXEL, 2);
	TIFFSetField(image, TIFFTAG_EXTRASAMPLES, EXTRASAMPLE_UNSPECIFIED);
	TIFFSetField(image, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
	TIFFSetField(image, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
	TIFFSetField(image, TIFFTAG_ROWSPERSTRIP, rowsperstrip);
	TIFFSetField(image, TIFFTAG_COMPRESSION, compression);

	const char *ptr = desc.c_str();
        TIFFSetField(image, TIFFTAG_IMAGEDESCRIPTION, ptr);

	std::ifstream cm(dev.c_str()); // cache memory stream
	if(cm.is_open()) {
		// Add time stamping variables
		boost::posix_time::ptime			start(boost::posix_time::microsec_clock::local_time()),
											lcur(start),
											cur,
											fini;


		const size_t NumberOfStrips = TIFFNumberOfStrips(image);
		const size_t StripSize = TIFFStripSize(image);
	        log.noticeStream() << "write_video: " 
			<< " # strips = "   << NumberOfStrips
			<< " strip size = "   << StripSize
			<< " rows per strip = "   << rowsperstrip
			<< l4::eol;

		// Read essence from cache
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
                for(size_t strip=0; strip<NumberOfStrips; strip++) {
                        for(size_t row=0; row<rowsperstrip; row++) {
                                size_t y = strip*rowsperstrip + row;
				loff_t offset = y%2 ? even : odd;
				cm.seekg(offset+(y/2)*linesiz);
                                cm.read(essence+((strip*StripSize)+(row*(StripSize/rowsperstrip))), linesiz);
                        }
                }
		cur = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds
		boost::posix_time::time_duration t_read = cur - lcur; 	// accumulate duration taken for read

		// Write essence to file
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
		for(size_t y=0; y<NumberOfStrips; y++) {
			char *ptr = essence+(y*StripSize);
			if( TIFFWriteEncodedStrip(image, y, ptr, StripSize) < 0) throw std::runtime_error("TIFFWriteEncodedStrip");
			pos+=rowsperstrip;

		}
		cur = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds
		boost::posix_time::time_duration t_write = cur - lcur;	// accumulate duration taken for write

		cm.close();

		fini = boost::posix_time::microsec_clock::local_time();		// Fetch the current time in microseconds

		// dump our findings on time used by each operation
	        log.noticeStream() << "write_video: " 
			<< " time reading = " << t_read
			<< " time writing = " << t_write 
			<< " total time = " << (fini-start)
			<< l4::eol;
	}

	uint16 count = 1;
	uint32 diroff[2] = { 0, 0 };
	if(parts=="videokey") count = 2;
	TIFFSetField(image, TIFFTAG_SUBIFD, count, diroff);

	if( TIFFWriteDirectory(image) != 1 ) throw std::runtime_error("TIFFWriteDirectory");
}

void save::write_key(loff_t offset, const size_t linesiz)
{
	log.noticeStream() << " offset=" << offset << l4::eol;
	if(!image) {
		log.warnStream() << "no image" << l4::eol;
		throw std::runtime_error("no image");
	}

	TIFFCreateDirectory(image);
	TIFFSetField(image, TIFFTAG_MAKE, make);
	TIFFSetField(image, TIFFTAG_MODEL, model);
	TIFFSetField(image, TIFFTAG_SOFTWARE, ver);

	// We need to set some values for basic tags before we can add any data
	TIFFSetField(image, TIFFTAG_IMAGEWIDTH, essence_width);
	TIFFSetField(image, TIFFTAG_IMAGELENGTH, essence_length);
	TIFFSetField(image, TIFFTAG_BITSPERSAMPLE, 16);
	TIFFSetField(image, TIFFTAG_SAMPLESPERPIXEL, 2);
	TIFFSetField(image, TIFFTAG_EXTRASAMPLES, EXTRASAMPLE_UNSPECIFIED);
	TIFFSetField(image, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
	TIFFSetField(image, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
	TIFFSetField(image, TIFFTAG_ROWSPERSTRIP, rowsperstrip);
	TIFFSetField(image, TIFFTAG_COMPRESSION, compression);

	std::ifstream cm(dev.c_str()); // cache memory stream
	if(cm.is_open()) {
		// Add time stamping variables
		boost::posix_time::ptime			start(boost::posix_time::microsec_clock::local_time()),
											lcur(start),
											cur,
											fini;

		const size_t NumberOfStrips = TIFFNumberOfStrips(image);
		const size_t StripSize = TIFFStripSize(image);
	        log.noticeStream() << "write_key: " 
			<< " # strips = "   << NumberOfStrips
			<< " strip size = "   << StripSize
			<< " rows per strip = "   << rowsperstrip
			<< l4::eol;

		// Read essence from cache
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
                for(size_t strip=0; strip<NumberOfStrips; strip++) {
                        for(size_t row=0; row<rowsperstrip; row++) {
                                size_t y = strip*rowsperstrip + row;
				cm.seekg(offset+y*linesiz);
                                cm.read(essence+((strip*StripSize)+(row*(StripSize/rowsperstrip))), linesiz);
                        }
                }
		cur = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds
		boost::posix_time::time_duration t_read = cur - lcur;	// accumulate duration taken for read

		// Write essence to file
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
		for(size_t y=0; y<NumberOfStrips; y++) {
			char *ptr = essence+(y*StripSize);
			if( TIFFWriteEncodedStrip(image, y, ptr, StripSize) < 0) throw std::runtime_error("TIFFWriteEncodedStrip");
			pos+=rowsperstrip;
		}
		cur = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds
		boost::posix_time::time_duration t_write = cur - lcur; 	// accumulate duration taken for write

		cm.close();

		fini = boost::posix_time::microsec_clock::local_time();		// Fetch the current time in microseconds

		// dump our findings on time used by each operation
	        log.noticeStream() << "write_key: " 
			<< " time reading = " << t_read
			<< " time writing = " << t_write 
			<< " total time = " << (fini-start)
			<< l4::eol;
	}

	TIFFSetField(image, TIFFTAG_SUBFILETYPE, FILETYPE_MASK);

	if( TIFFWriteDirectory(image) != 1 ) throw std::runtime_error("TIFFWriteDirectory");
}

void save::write_key(loff_t odd, loff_t even, const size_t linesiz)
{
	log.noticeStream() << " odd=" << odd << " even=" << even << l4::eol;
	if(!image) {
		log.warnStream() << "no image" << l4::eol;
		throw std::runtime_error("no image");
	}

	TIFFCreateDirectory(image);
	TIFFSetField(image, TIFFTAG_MAKE, make);
	TIFFSetField(image, TIFFTAG_MODEL, model);
	TIFFSetField(image, TIFFTAG_SOFTWARE, ver);

	// We need to set some values for basic tags before we can add any data
	TIFFSetField(image, TIFFTAG_IMAGEWIDTH, essence_width);
	TIFFSetField(image, TIFFTAG_IMAGELENGTH, essence_length);
	TIFFSetField(image, TIFFTAG_BITSPERSAMPLE, 16);
	TIFFSetField(image, TIFFTAG_SAMPLESPERPIXEL, 2);
	TIFFSetField(image, TIFFTAG_EXTRASAMPLES, EXTRASAMPLE_UNSPECIFIED);
	TIFFSetField(image, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
	TIFFSetField(image, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
	TIFFSetField(image, TIFFTAG_ROWSPERSTRIP, rowsperstrip);
	TIFFSetField(image, TIFFTAG_COMPRESSION, compression);

	std::ifstream cm(dev.c_str()); // cache memory stream
	if(cm.is_open()) {
		// Add time stamping variables
		boost::posix_time::ptime			start(boost::posix_time::microsec_clock::local_time()),
											lcur(start),
											cur,
											fini;

		const size_t NumberOfStrips = TIFFNumberOfStrips(image);
		const size_t StripSize = TIFFStripSize(image);
	        log.noticeStream() << "write_key: " 
			<< " # strips = "   << NumberOfStrips
			<< " strip size = "   << StripSize
			<< " rows per strip = "   << rowsperstrip
			<< l4::eol;

		// Read essence from cache
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
                for(size_t strip=0; strip<NumberOfStrips; strip++) {
                        for(size_t row=0; row<rowsperstrip; row++) {
                                size_t y = strip*rowsperstrip + row;
				loff_t offset = y%2 ? even : odd;
				cm.seekg(offset+(y/2)*linesiz);
                                cm.read(essence+((strip*StripSize)+(row*(StripSize/rowsperstrip))), linesiz);
                        }
                }
		cur = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds
		boost::posix_time::time_duration t_read = cur - lcur; 	// accumulate duration taken for read

		// Write essence to file
		lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds
		for(size_t y=0; y<NumberOfStrips; y++) {
			char *ptr = essence+(y*StripSize);
			if( TIFFWriteEncodedStrip(image, y, ptr, StripSize) < 0) throw std::runtime_error("TIFFWriteEncodedStrip");
			pos+=rowsperstrip;
		}
		cur = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds
		boost::posix_time::time_duration t_write = cur - lcur;	// accumulate duration taken for write

		cm.close();

		fini = boost::posix_time::microsec_clock::local_time();		// Fetch the current time in microseconds

		// dump our findings on time used by each operation
	        log.noticeStream() << "write_key: " 
			<< " time reading = " << t_read
			<< " time writing = " << t_write 
			<< " total time = " << (fini-start)
			<< l4::eol;
	}

	TIFFSetField(image, TIFFTAG_SUBFILETYPE, FILETYPE_MASK);

	if( TIFFWriteDirectory(image) != 1 ) throw std::runtime_error("TIFFWriteDirectory");
}

void save::write_thumb(const size_t hBias, const size_t hGain, const size_t vBias, const size_t vGain, loff_t offset, const size_t len)
{
	char pixels[84*3];
	char buf[len];

	log.noticeStream() << " offset=" << offset << l4::eol;
	if(!image) {
		log.warnStream() << "no image" << l4::eol;
		throw std::runtime_error("no image");
	}

	TIFFCreateDirectory(image);
	TIFFSetField(image, TIFFTAG_MAKE, make);
	TIFFSetField(image, TIFFTAG_MODEL, model);
	TIFFSetField(image, TIFFTAG_SOFTWARE, ver);

	// We need to set some values for basic tags before we can add any data
	TIFFSetField(image, TIFFTAG_IMAGEWIDTH, thumb_width);
	TIFFSetField(image, TIFFTAG_IMAGELENGTH, thumb_length);
	TIFFSetField(image, TIFFTAG_BITSPERSAMPLE, 8);
	TIFFSetField(image, TIFFTAG_SAMPLESPERPIXEL, 3);
	TIFFSetField(image, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
	TIFFSetField(image, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
	TIFFSetField(image, TIFFTAG_ROWSPERSTRIP, 1);

	std::ifstream cm(dev.c_str()); // cache memory stream
	if(cm.is_open()) {

		// Add time stamping variables
		boost::posix_time::ptime			start(boost::posix_time::microsec_clock::local_time()),
											lcur(start),
											cur,
											fini;

		// Add duration variables to profile with
		boost::posix_time::time_duration	t_read, t_adj, t_write;

		for(size_t y=0; y<48; y++) {

			lcur = boost::posix_time::microsec_clock::local_time(); // Fetch the current time in microseconds

			cm.seekg(offset+vGain*(y+vBias)*len);

			cur = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds
			lcur = cur;						// Save the time for next calculation

			cm.read(buf, len);

			cur = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds
			t_read = cur - lcur; 					// accumulate duration taken for read
			lcur = cur;						// Save the time for next calculation

			for(size_t x=0; x<84; x++) {
				pixels[3*x+0] = ycbcr2r(buf+(hBias+x*hGain));
				pixels[3*x+1] = ycbcr2g(buf+(hBias+x*hGain));
				pixels[3*x+2] = ycbcr2b(buf+(hBias+x*hGain));
			}

			cur = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds
			t_adj = cur - lcur; 					// accumulate duration taken for adjusting thumbnail
			lcur = cur;						// Save the time for next calculation

			if( TIFFWriteScanline(image, pixels, y) != 1) throw std::runtime_error("TIFFWriteScanline");

			cur = boost::posix_time::microsec_clock::local_time();	// Fetch the current time in microseconds
			t_write = cur - lcur; 					// accumulate duration taken for write
			lcur = cur;						// Save the time for next calculation

		}
		cm.close();

		fini = boost::posix_time::microsec_clock::local_time();		// Fetch the current time in microseconds

		// dump our findings on time used by each operation
	        log.noticeStream() << "write_thumb: " 
			<< " time reading = " << t_read
			<< " time writing = " << t_write 
			<< " total time = " << (fini-start)
			<< l4::eol;
	}

	TIFFSetField(image, TIFFTAG_SUBFILETYPE, FILETYPE_REDUCEDIMAGE);

	if( TIFFWriteDirectory(image) != 1 ) throw std::runtime_error("TIFFWriteDirectory");
}

void save::writer(const strpair &label, const std::string &parts)
{
	const std::string &odd  = label.first;
	const std::string &even = label.second;
	blkaddr ba;
        log.noticeStream() << "odd=" << odd << " even=" << even << l4::eol;

	try {
                loff_t oddset  = ba.addr(odd+":v");
                loff_t evenset = (even.empty() ? 0 : ba.addr(even+":v"));
		switch(ba.used(odd+":v")) {
		case 9720:
			essence_width  = 1920;
			essence_length = 1080;
			std = "full1080i";
	                log.noticeStream() << "write 1080i image odd=" << oddset << " even=" << evenset << l4::eol;
			if(even.empty()) {
				write_video(oddset*BLKSIZ, parts, 18*BLKSIZ);
			} else {
				write_video(oddset*BLKSIZ, evenset*BLKSIZ, parts, 18*BLKSIZ);
			}
			log.noticeStream() << "write thumb offset=" << oddset << l4::eol;
			write_thumb(40, 88, 5, 10, oddset*BLKSIZ, 18*BLKSIZ);
			break;
		case 8640:
			essence_width  = 1280;
			essence_length = 720;
			std = "full720p";
	                log.noticeStream() << "write 720p image offset=" << oddset << l4::eol;
			if(even.empty()) {
				write_video(oddset*BLKSIZ, parts, 12*BLKSIZ);
			} else {
				write_video(oddset*BLKSIZ, evenset*BLKSIZ, parts, 12*BLKSIZ);
			}
			log.noticeStream() << "write thumb offset=" << oddset << l4::eol;
			write_thumb(16, 56, 6, 12, oddset*BLKSIZ, 12*BLKSIZ);
			break;
		case 2016:
			essence_width  = 720;
			essence_length = 576;
			std = "full625i";
	                log.noticeStream() << "write 625i image odd=" << oddset << " even=" << evenset << l4::eol;
			if(even.empty()) {
				write_video(oddset*BLKSIZ, parts, 7*BLKSIZ);
			} else {
				write_video(oddset*BLKSIZ, evenset*BLKSIZ, parts, 7*BLKSIZ);
			}
			log.noticeStream() << "write thumb offset=" << oddset << l4::eol;
			write_thumb(8, 32, 5, 5, oddset*BLKSIZ, 7*BLKSIZ);
			break;
		case 1701:
		case 1708:
			essence_width  = 720;
			essence_length = 486;
			std = "full525i";
	                log.noticeStream() << "write 525i image odd=" << oddset << " even=" << evenset << l4::eol;
			if(even.empty()) {
				write_video(oddset*BLKSIZ, parts, 7*BLKSIZ);
			} else {
				write_video(oddset*BLKSIZ, evenset*BLKSIZ, parts, 7*BLKSIZ);
			}
			log.noticeStream() << "write thumb offset=" << oddset << l4::eol;
			write_thumb(8, 32, 4, 4, oddset*BLKSIZ, 7*BLKSIZ);
			break;
		default:
			essence_width  = 0;
			essence_length = 0;
	                log.errorStream() << "unknown image format [" << ba.used(odd+":v") << "]" << l4::eol;
			throw std::runtime_error("unknown image format");
		}
	} catch(...) {
		log.warnStream() << "essence not saved" << l4::eol;
	}

	if(parts=="videokey") {
		try {
			loff_t oddset  = ba.addr(odd+":k");
                	loff_t evenset = (even.empty() ? 0 : ba.addr(even+":k"));
			switch(ba.used(odd+":k")) {
			case 9720:
        		        log.noticeStream() << "write key odd=" << oddset << " even=" << evenset << l4::eol;
				write_key(oddset*BLKSIZ, evenset*BLKSIZ, 18*BLKSIZ);
				break;
			case 8640:
        		        log.noticeStream() << "write key offset=" << oddset << l4::eol;
				write_key(oddset*BLKSIZ, 12*BLKSIZ);
				break;
			case 2016:
        		        log.noticeStream() << "write key odd=" << oddset << " even=" << evenset << l4::eol;
				write_key(oddset*BLKSIZ, evenset*BLKSIZ, 7*BLKSIZ);
				break;
			case 1701:
			case 1708:
        		        log.noticeStream() << "write key odd=" << oddset << " even=" << evenset << l4::eol;
				write_key(oddset*BLKSIZ, evenset*BLKSIZ, 7*BLKSIZ);
				break;
			default:
				essence_width  = 0;
				essence_length = 0;
	        	        log.errorStream() << "unknown image format [" << ba.used(odd+":k") << "]" << l4::eol;
				throw std::runtime_error("unknown image format");
			}
		} catch(...) {
			log.noticeStream() << "key not saved" << l4::eol;
        	        loff_t offset = ba.addr(odd+":v");
			write_thumb(40, 88, 5, 10, offset*BLKSIZ, 18*BLKSIZ);
		}
	}
}

