/// \file
/// \copyright 2008 Thomson Grass Valley
/// \author adam clarke
/// \date 2008-06-06
/// \brief a collection of inserters and extractors to support ImageStore operations

#include <iostream>
#include <sstream>
#include <iterator>
#include <iomanip>
#include <map>

#include <log4cpp/Category.hh>

#include "blkaddr.h"

namespace l4 = log4cpp;

/// \brief a shared memory space for constructing shared objects associated with a name.
///
/// establishes a connection to an existing shared memory segment 
boost::interprocess::managed_shared_memory blkaddr::segment (boost::interprocess::open_or_create,"MossDirectory" , 8*65536);         
blkaddr::MossMap * blkaddr::directory = segment.find<MossMap>("MossMap").first;
size_t * blkaddr::nextAvail = segment.find<size_t>("MossNextAvail").first;

void blkaddr::dump(std::ostream &sink) const {
	if(!directory) throw std::runtime_error("blkaddr::dump - moss directory not found");

	blkaddr::MossMap::const_iterator itr;
	for(itr=directory->begin();itr!=directory->end();itr++) {
		sink	<< (itr->first) 
			<< " [offset=" << (itr->second.offset)
			<< " allocated=" << (itr->second.allocated)
			<< " consumed=" << (itr->second.consumed) << "]"
			<< std::endl;
	}
}

blkaddr::blkaddr(void) : log(l4::Category::getInstance("blkaddr")), mutex(boost::interprocess::open_or_create, "moss") {
	//Find the map
	if(!directory) {
		//Initialize the shared memory STL-compatible allocator
		ShmemAllocator alloc_inst (segment.get_segment_manager());

		directory = segment.construct<MossMap>("MossMap")      //object name
						(std::less<KeyType>(), //first  ctor parameter
						alloc_inst);     //second ctor parameter
	}
	//Find the next available offset, initialize to 1000
	if(!nextAvail) {
		nextAvail = segment.construct<size_t>("MossNextAvail") (1000);
	}
}

/// \brief finds a cache memory offset for a given id
/// creates a new cache memory offset if one does not already exist
size_t blkaddr::assign(const std::string &label, const size_t segs, const size_t used) {
	if(!directory) throw std::runtime_error("blkaddr::assign - moss directory not found");
	if(!nextAvail) throw std::runtime_error("blkaddr::assign - moss nextavail not found");
	log.debugStream() << "request " << label << l4::eol;

	static KeyType key(segment.get_segment_manager());
	key = label.c_str();
	MossMap::iterator itr = directory->find(key);
	if(itr == directory->end()) {
		boost::interprocess::scoped_lock<boost::interprocess::named_mutex> lock(mutex);
		MappedType avail = { (*nextAvail), segs, used };
		(*nextAvail) += segs;
		directory->insert( std::make_pair(key, avail));
		log.infoStream() << "assigned: " << segs << " to " << key.c_str() << l4::eol;
	} else {
		itr->second.allocated = segs;
		log.debugStream() << "found: " << itr->first.c_str() << l4::eol;
	}
	return (*directory)[key].offset;
}

/// \brief record consumption of a cache memory block
void blkaddr::consume(const std::string &id, const size_t segs) {
	if(!directory) throw std::runtime_error("blkaddr::consume - moss directory not found");

	static KeyType key(segment.get_segment_manager());
	key = id.c_str();
	MossMap::iterator itr = directory->find(key);
	if(itr == directory->end()) throw std::runtime_error("blkaddr::consume - " + id + " - not found");
	itr->second.consumed = segs;
}

/// \brief find address of a cache memory block
size_t blkaddr::addr(const std::string &id) const {
	if(!directory) throw std::runtime_error("blkaddr::addr - moss directory not found");

	static KeyType key(segment.get_segment_manager());
	key = id.c_str();
	MossMap::const_iterator itr = directory->find(key);
	if(itr == directory->end()) throw std::runtime_error("blkaddr::addr - " + id + " - not found");
	return itr->second.offset;
}

/// \brief find consumption of a cache memory block
size_t blkaddr::used(const std::string &id) const {
	if(!directory) throw std::runtime_error("blkaddr::used - moss directory not found");

	static KeyType key(segment.get_segment_manager());
	key = id.c_str();
	MossMap::const_iterator itr = directory->find(key);
	if(itr == directory->end()) throw std::runtime_error("blkaddr::used - " + id + " - not found");
	return itr->second.consumed;
}

/// \brief move a cached image
bool blkaddr::move(const short suite, const long srcID, const long targetID) {
	static KeyType key(segment.get_segment_manager());
	std::ostringstream buf;

	if(!directory) throw std::runtime_error("blkaddr::move - moss directory not found");

	// enumerate the sources
	buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID) << std::ends;
	key = buf.str().c_str();
	MossMap::iterator sBegin = directory->lower_bound(key);
	if(sBegin == directory->end()) {
		log.debugStream() << "blkaddr::move src - " + buf.str() + " - not found" << l4::eol;
		return false;
	}
	if(sBegin->first.find(buf.str().c_str())==std::string::npos) {
		log.debugStream() << "blkaddr::move src - " + buf.str() + " - wrong label found" << l4::eol;
		return false;
	}
	buf.str("");

	MossMap::iterator sEnd;
	if(srcID<9999) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID+1) << std::ends;
	} else {
		buf << (suite+1) << ":0000" << std::ends;
	}
	key = buf.str().c_str();
	sEnd = directory->upper_bound(key);
	buf.str("");

	std::vector<std::string> fields;
	enumerate(fields,sBegin,sEnd);
	if(fields.empty()) {
		log.debugStream() << "empty move" << l4::eol;
		return false;
	}

	log.noticeStream() << "move src=" << suite << ":" << srcID << " target=" << suite << ":" << targetID << l4::eol;

	// create the derived targets
	for(std::vector<std::string>::iterator itr=fields.begin();itr!=fields.end();itr++) {
		MappedType avail = { 0, 0, 0 };
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID) 
			<< ":" << itr->substr(7) << std::ends;
		key = buf.str().c_str();
		boost::interprocess::scoped_lock<boost::interprocess::named_mutex> lock(mutex);
		directory->insert(std::make_pair(key, avail));
		log.debugStream() << "added: " << buf.str().c_str() << l4::eol;
		buf.str("");
	}

	// copy the source avail into the derived targets
	buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID) << std::ends;
	key = buf.str().c_str();
	MossMap::iterator cBegin = directory->lower_bound(key);
	if(cBegin == directory->end()) throw std::runtime_error("blkaddr::move - " + buf.str() + " - not found");
	if(cBegin->first.find(buf.str().c_str())==std::string::npos) {
		throw std::runtime_error("blkaddr::move - " + buf.str() + " - not found");
	}
	buf.str("");

	MossMap::iterator cEnd;
	if(targetID<9999) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID+1) << std::ends;
	} else {
		buf << (suite+1) << ":0000" << std::ends;
	}
	key = buf.str().c_str();
	cEnd = directory->upper_bound(key);
	buf.str("");

	for(MossMap::iterator itr=cBegin; itr!=cEnd; itr++) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID) 
			<< ":" << itr->first.substr(7) << std::ends;
		key = buf.str().c_str();
		itr->second = (*directory)[key];
		log.debugStream() << "copy src=" << key << " target=" << itr->first << l4::eol;
		buf.str("");
	}

	// final step, erase the src
	buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID) << std::ends;
	key = buf.str().c_str();
	MossMap::iterator eBegin = directory->lower_bound(key);
	if(eBegin == directory->end()) throw std::runtime_error("blkaddr::move - " + buf.str() + " - not found");
	if(eBegin->first.find(buf.str().c_str())==std::string::npos) {
		throw std::runtime_error("blkaddr::move - " + buf.str() + " - not found");
	}
	buf.str("");

	MossMap::iterator eEnd;
	if(srcID<9999) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID+1) << std::ends;
	} else {
		buf << (suite+1) << ":0000" << std::ends;
	}
	key = buf.str().c_str();
	eEnd = directory->upper_bound(key);
	buf.str("");

	directory->erase(eBegin, eEnd);

	log.noticeStream() << "move complete" << l4::eol;
	return true;
}

/// \brief swap the adresses of two cached images
/// \note we leave the keys alone, so nothing is ever invalid
bool blkaddr::swap(const short suite, const long srcID, const long targetID) {
	static KeyType key(segment.get_segment_manager());
	std::ostringstream buf;

	if(!directory) throw std::runtime_error("blkaddr::swap - moss directory not found");

	buf << suite << ":" << std::setfill('0') << std::setw(4) << targetID << std::ends;
	key = buf.str().c_str();
	MossMap::iterator start = directory->lower_bound(key);
	if(start == directory->end()) { 
		log.debugStream() << "target - " << buf.str() << " - not found" << l4::eol;
		return false;
	}
	if(start->first.find(buf.str().c_str())==std::string::npos) {
		log.debugStream() << "target - " << buf.str() << " - not found" << l4::eol;
		return false;
	}
	buf.str("");

	MossMap::iterator end;
	if(targetID<9999) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID+1) << std::ends;
	} else {
		buf << (suite+1) << ":0000" << std::ends;
	}
	key = buf.str().c_str();
	end = directory->upper_bound(key);
	buf.str("");

	if(start<end) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << srcID << std::ends;
		key = buf.str().c_str();
		MossMap::iterator s = directory->lower_bound(key);
		if(s == directory->end()) {
			log.debugStream() << "src - " << buf.str() << " - not found" << l4::eol;
			return false;
		}
		if(s->first.find(buf.str().c_str())==std::string::npos) {
			log.debugStream() << "src - " << buf.str() << " - not found" << l4::eol;
			return false;
		}
		buf.str("");

		for(MossMap::iterator itr=start; itr<end; itr++, s++) {
			if(s == directory->end()) break;
			log.debugStream() << "swapping src=" << s->first << " target=" << itr->first << l4::eol;
			MappedType tmp = itr->second;
			itr->second = s->second;
			s->second = tmp;
		}
	} else {
		log.warnStream() << "swap failed, no target" << l4::eol;
		return false;
	}

	log.noticeStream() << "swap complete" << l4::eol;
	return true;
}

/// \brief marry one image to another
void blkaddr::marry(const short suite, const long srcID, const long targetID) {
	static KeyType key(segment.get_segment_manager());
	std::ostringstream buf;

	if(!directory) throw std::runtime_error("blkaddr::marry - moss directory not found");

	// enumerate the sources
	buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID) << std::ends;
	key = buf.str().c_str();
	MossMap::iterator sBegin = directory->lower_bound(key);
	if(sBegin == directory->end()) throw std::runtime_error("blkaddr::marry - " + buf.str() + " - not found");
	if(sBegin->first.find(buf.str().c_str())==std::string::npos) {
		throw std::runtime_error("blkaddr::marry - " + buf.str() + " - not found");
	}
	buf.str("");

	MossMap::iterator sEnd;
	if(srcID<9999) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID+1) << std::ends;
	} else {
		buf << (suite+1) << ":0000" << std::ends;
	}
	key = buf.str().c_str();
	sEnd = directory->upper_bound(key);
	buf.str("");

	std::vector<std::string> fields;
	enumerate(fields,sBegin,sEnd);
	if(fields.empty()) {
		log.debugStream() << "empty marry" << l4::eol;
		return;
	}

	// create the derived targets
	for(std::vector<std::string>::iterator itr=fields.begin();itr!=fields.end();itr++) {
		if(itr->find(":v") != std::string::npos) {
			MappedType avail = { 0, 0, 0 };
			buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID) 
				<< ":" << itr->substr(7,4) << ":k" << std::ends;
			key = buf.str().c_str();
			boost::interprocess::scoped_lock<boost::interprocess::named_mutex> lock(mutex);
			directory->insert(std::make_pair(key, avail));
			log.debugStream() << "added: " << buf.str().c_str() << l4::eol;
		}
		buf.str("");
	}

	// copy the source avail into the derived targets
	buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID) << std::ends;
	key = buf.str().c_str();
	MossMap::iterator cBegin = directory->lower_bound(key);
	if(cBegin == directory->end()) throw std::runtime_error("blkaddr::marry - " + buf.str() + " - not found");
	if(cBegin->first.find(buf.str().c_str())==std::string::npos) {
		throw std::runtime_error("blkaddr::marry - " + buf.str() + " - not found");
	}
	buf.str("");

	MossMap::iterator cEnd;
	if(targetID<9999) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID+1) << std::ends;
	} else {
		buf << (suite+1) << ":0000" << std::ends;
	}
	key = buf.str().c_str();
	cEnd = directory->upper_bound(key);
	buf.str("");

	buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID) << std::ends;
	key = buf.str().c_str();
	MossMap::iterator point = directory->lower_bound(key);
	if(point == directory->end()) throw std::runtime_error("blkaddr::marry - " + buf.str() + " - not found");
	if(point->first.find(buf.str().c_str())==std::string::npos) {
		throw std::runtime_error("blkaddr::marry - " + buf.str() + " - not found");
	}
	buf.str("");

	for(MossMap::iterator itr=cBegin; itr!=cEnd; itr++) {
		if(itr->first.find(":k") != std::string::npos) {
			while(point->first.find(":k") != std::string::npos) {
				point++;
				if(point == directory->end()) {
					throw std::runtime_error("blkaddr::marry - " + buf.str() + " - not found");
				}
			}
			itr->second = point->second;
			log.infoStream() << "marry src=" << point->first << " target=" << itr->first << l4::eol;
			point++;
		}
	}

	// final step, erase the src
	buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID) << std::ends;
	key = buf.str().c_str();
	MossMap::iterator eBegin = directory->lower_bound(key);
	if(eBegin == directory->end()) throw std::runtime_error("blkaddr::marry - " + buf.str() + " - not found");
	if(eBegin->first.find(buf.str().c_str())==std::string::npos) {
		throw std::runtime_error("blkaddr::marry - " + buf.str() + " - not found");
	}
	buf.str("");

	MossMap::iterator eEnd;
	if(srcID<9999) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID+1) << std::ends;
	} else {
		buf << (suite+1) << ":0000" << std::ends;
	}
	key = buf.str().c_str();
	eEnd = directory->upper_bound(key);
	buf.str("");

	directory->erase(eBegin, eEnd);

	log.noticeStream() << "marry complete" << l4::eol;
}

/// \brief append one image to another
void blkaddr::append(const short suite, const long srcID, const long targetID) {
	static KeyType key(segment.get_segment_manager());
	std::ostringstream buf;

	if(!directory) throw std::runtime_error("blkaddr::append - moss directory not found");

	// enumerate the sources
	buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID) << std::ends;
	key = buf.str().c_str();
	MossMap::iterator sBegin = directory->lower_bound(key);
	if(sBegin == directory->end()) throw std::runtime_error("blkaddr::append - " + buf.str() + " - not found");
	if(sBegin->first.find(buf.str().c_str())==std::string::npos) {
		throw std::runtime_error("blkaddr::append - " + buf.str() + " - not found");
	}
	buf.str("");

	MossMap::iterator sEnd;
	if(srcID<9999) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID+1) << std::ends;
	} else {
		buf << (suite+1) << ":0000" << std::ends;
	}
	key = buf.str().c_str();
	sEnd = directory->upper_bound(key);
	buf.str("");

	std::vector<std::string> fields;
	enumerate(fields,sBegin,sEnd);
	if(fields.empty()) {
		log.debugStream() << "empty append" << l4::eol;
		return;
	}

	// get the size of the target image
	buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID) << std::ends;
	key = buf.str().c_str();
	MossMap::iterator tBegin = directory->lower_bound(key);
	if(tBegin == directory->end()) throw std::runtime_error("blkaddr::append - " + buf.str() + " - not found");
	if(tBegin->first.find(buf.str().c_str())==std::string::npos) {
		throw std::runtime_error("blkaddr::append - " + buf.str() + " - not found");
	}
	buf.str("");

	MossMap::iterator tEnd;
	if(targetID<9999) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID+1) << std::ends;
	} else {
		buf << (suite+1) << ":0000" << std::ends;
	}
	key = buf.str().c_str();
	tEnd = directory->upper_bound(key);
	buf.str("");

	int tLen=0;
	for(MossMap::iterator itr=tBegin; itr!=tEnd; itr++) {
		// increment the field count for video fields only
		if(itr->first.find('v') != std::string::npos) tLen++; 
	}
	log.debugStream() << "targetlength=" << tLen << l4::eol;
	tLen++;

	// create the derived targets
	int fld = tLen;
	for(std::vector<std::string>::iterator itr=fields.begin();itr!=fields.end();itr++) {
		MappedType avail = { 0, 0, 0 };
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID) 
			<< ":" << std::setfill('0') << std::setw(4) << (fld) 
			<< ":" << itr->substr(12) << std::ends;
		key = buf.str().c_str();
		boost::interprocess::scoped_lock<boost::interprocess::named_mutex> lock(mutex);
		directory->insert(std::make_pair(key, avail));
		log.debugStream() << "added: " << buf.str().c_str() << l4::eol;
		buf.str("");
		if(itr->find('v') != std::string::npos) fld++; 
	}

	// copy the sources to the targets
	buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID) 
			<< ":" << std::setfill('0') << std::setw(4) << (tLen) << std::ends; 
	key = buf.str().c_str();
	MossMap::iterator cBegin = directory->lower_bound(key);
	if(cBegin == directory->end()) throw std::runtime_error("blkaddr::append - " + buf.str() + " - not found");
	if(cBegin->first.find(buf.str().c_str())==std::string::npos) {
		throw std::runtime_error("blkaddr::append - " + buf.str() + " - not found");
	}
	buf.str("");

	MossMap::iterator cEnd;
	if(targetID<9999) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID+1) << std::ends;
	} else {
		buf << (suite+1) << ":0000" << std::ends;
	}
	key = buf.str().c_str();
	cEnd = directory->upper_bound(key);
	buf.str("");

	buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID) << std::ends;
	key = buf.str().c_str();
	MossMap::iterator point = directory->lower_bound(key);
	if(point == directory->end()) throw std::runtime_error("blkaddr::append - " + buf.str() + " - not found");
	if(point->first.find(buf.str().c_str())==std::string::npos) {
		throw std::runtime_error("blkaddr::append - " + buf.str() + " - not found");
	}
	buf.str("");

	for(MossMap::iterator itr=cBegin; itr!=cEnd; itr++, point++) {
		itr->second = point->second;
		log.infoStream() << "copy src=" << point->first << " target=" << itr->first << l4::eol;
	}

	// final step, erase the src
	buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID) << std::ends;
	key = buf.str().c_str();
	MossMap::iterator eBegin = directory->lower_bound(key);
	if(eBegin == directory->end()) throw std::runtime_error("blkaddr::append - " + buf.str() + " - not found");
	if(eBegin->first.find(buf.str().c_str())==std::string::npos) {
		throw std::runtime_error("blkaddr::append - " + buf.str() + " - not found");
	}
	buf.str("");

	MossMap::iterator eEnd;
	if(srcID<9999) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID+1) << std::ends;
	} else {
		buf << (suite+1) << ":0000" << std::ends;
	}
	key = buf.str().c_str();
	eEnd = directory->upper_bound(key);
	buf.str("");

	directory->erase(eBegin, eEnd);
	log.noticeStream() << "append complete" << l4::eol;
}

/// \brief trim the head of a clip
/// \note we do this by copying the addresses down, then erase the tail
/// \todo save the head
void blkaddr::head(const short suite, const long srcID, const long mark, const long targetID) {
	static KeyType key(segment.get_segment_manager());
	std::ostringstream buf;

	if(!directory) throw std::runtime_error("blkaddr::head - moss directory not found");

	// if there is a targetID, copy the head to the target
	if(targetID) {
		// enumerate the sources
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID) << std::ends;
		key = buf.str().c_str();
		MossMap::iterator sBegin = directory->lower_bound(key);
		if(sBegin == directory->end()) throw std::runtime_error("blkaddr::head - " + buf.str() + " - not found");
		if(sBegin->first.find(buf.str().c_str())==std::string::npos) {
			throw std::runtime_error("blkaddr::head - " + buf.str() + " - not found");
		}
		buf.str("");

		buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID)
			<< ":" << std::setfill('0') << std::setw(4) << (mark) << std::ends;
		key = buf.str().c_str();
		MossMap::iterator sEnd = directory->upper_bound(key);
		buf.str("");

		std::vector<std::string> fields;
		enumerate(fields,sBegin,sEnd);
		if(fields.empty()) {
			log.debugStream() << "empty head" << l4::eol;
			return;
		}

		// create the derived targets
		for(std::vector<std::string>::iterator itr=fields.begin();itr!=fields.end();itr++) {
			MappedType avail = { 0, 0, 0 };
			buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID) 
				<< ":" << itr->substr(7) << std::ends;
			key = buf.str().c_str();
			boost::interprocess::scoped_lock<boost::interprocess::named_mutex> lock(mutex);
			directory->insert(std::make_pair(key, avail));
			log.debugStream() << "added: " << buf.str().c_str() << l4::eol;
			buf.str("");
		}

		// copy the sources to the targets
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID) << std::ends;
		key = buf.str().c_str();
		MossMap::iterator cBegin = directory->lower_bound(key);
		if(cBegin == directory->end()) throw std::runtime_error("blkaddr::head - " + buf.str() + " - not found");
		if(cBegin->first.find(buf.str().c_str())==std::string::npos) {
			throw std::runtime_error("blkaddr::head - " + buf.str() + " - not found");
		}
		buf.str("");

		MossMap::iterator cEnd;
		if(targetID<9999) {
			buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID+1) << std::ends;
		} else {
			buf << (suite+1) << ":0000" << std::ends;
		}
		key = buf.str().c_str();
		cEnd = directory->upper_bound(key);
		buf.str("");

		buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID) << std::ends;
		key = buf.str().c_str();
		MossMap::iterator point = directory->lower_bound(key);
		if(point == directory->end()) throw std::runtime_error("blkaddr::head - " + buf.str() + " - not found");
		if(point->first.find(buf.str().c_str())==std::string::npos) {
			throw std::runtime_error("blkaddr::head - " + buf.str() + " - not found");
		}
		buf.str("");

		for(MossMap::iterator itr=cBegin; itr!=cEnd; itr++, point++) {
			itr->second = point->second;
			log.infoStream() << "head src=" << point->first << " target=" << itr->first << l4::eol;
		}
	}

	// copy the addresses down
	buf << suite << ":" << std::setfill('0') << std::setw(4) << srcID << std::ends;
	key = buf.str().c_str();
	MossMap::iterator first = directory->lower_bound(key);
	if(first == directory->end()) throw std::runtime_error("blkaddr::head - " + buf.str() + " - not found");
	if(first->first.find(buf.str().c_str())==std::string::npos) {
		throw std::runtime_error("blkaddr::head - " + buf.str() + " - not found");
	}
	buf.str("");

	buf << suite 
		<< ":" << std::setfill('0') << std::setw(4) << srcID 
		<< ":" << std::setfill('0') << std::setw(4) << (mark+1)
		<< std::ends;
	key = buf.str().c_str();
	MossMap::iterator point = directory->upper_bound(key);
	if(point == directory->end()) throw std::runtime_error("blkaddr::head - " + buf.str() + " - not found");
	buf.str("");

	MossMap::iterator last;
	if(srcID<9999) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID+1) << std::ends;
	} else {
		buf << (suite+1) << ":0000" << std::ends;
	}
	key = buf.str().c_str();
	last = directory->upper_bound(key);
	buf.str("");

	MossMap::iterator to, from;
	for(to=first,from=point; from<last; to++, from++) {
		to->second = from->second;
	}

	// erase the tail
	directory->erase(to, last);
	log.noticeStream() << "head complete" << l4::eol;
}

/// \brief trim the tail of a clip
/// \todo save the tail
void blkaddr::tail(const short suite, const long srcID, const long mark, const long targetID) {
	static KeyType key(segment.get_segment_manager());
	std::ostringstream buf;

	if(!directory) throw std::runtime_error("blkaddr::tail - moss directory not found");

	// if there is a targetID, copy the head to the target
	if(targetID) {
		// enumerate the sources
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID)
			<< ":" << std::setfill('0') << std::setw(4) << (mark) << std::ends;
		key = buf.str().c_str();
		MossMap::iterator sBegin = directory->lower_bound(key);
		if(sBegin == directory->end()) throw std::runtime_error("blkaddr::tail - " + buf.str() + " - not found");
		if(sBegin->first.find(buf.str().c_str())==std::string::npos) {
			throw std::runtime_error("blkaddr::tail - " + buf.str() + " - not found");
		}
		buf.str("");

		MossMap::iterator sEnd;
		if(srcID<9999) {
			buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID+1) << std::ends;
		} else {
			buf << (suite+1) << ":0000" << std::ends;
		}
		key = buf.str().c_str();
		sEnd = directory->upper_bound(key);
		buf.str("");

		std::vector<std::string> fields;
		enumerate(fields,sBegin,sEnd);
		if(fields.empty()) {
			log.debugStream() << "empty tail" << l4::eol;
			return;
		}

		// create the derived targets
		int i=0;
		for(std::vector<std::string>::iterator itr=fields.begin();itr!=fields.end();itr++) {
			MappedType avail = { 0, 0, 0 };
			buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID) 
				<< ":" << std::setfill('0') << std::setw(4) << (i) 
				<< ":" << itr->substr(12) << std::ends;
			key = buf.str().c_str();
			boost::interprocess::scoped_lock<boost::interprocess::named_mutex> lock(mutex);
			directory->insert(std::make_pair(key, avail));
			log.debugStream() << "added: " << buf.str().c_str() << l4::eol;

			// increment the field count for video fields only
			if(buf.str().find('v') != std::string::npos) i++; 
			buf.str("");
		}

		// copy the sources to the targets
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID) << std::ends;
		key = buf.str().c_str();
		MossMap::iterator cBegin = directory->lower_bound(key);
		if(cBegin == directory->end()) throw std::runtime_error("blkaddr::tail - " + buf.str() + " - not found");
		if(cBegin->first.find(buf.str().c_str())==std::string::npos) {
			throw std::runtime_error("blkaddr::tail - " + buf.str() + " - not found");
		}
		buf.str("");

		MossMap::iterator cEnd;
		if(targetID<9999) {
			buf << suite << ":" << std::setfill('0') << std::setw(4) << (targetID+1) << std::ends;
		} else {
			buf << (suite+1) << ":0000" << std::ends;
		}
		key = buf.str().c_str();
		cEnd = directory->upper_bound(key);
		buf.str("");

		buf << suite 
			<< ":" << std::setfill('0') << std::setw(4) << srcID 
			<< ":" << std::setfill('0') << std::setw(4) << mark
			<< std::ends;
		key = buf.str().c_str();
		MossMap::iterator point = directory->lower_bound(key);
		if(point == directory->end()) throw std::runtime_error("blkaddr::tail - " + buf.str() + " - not found");
		if(point->first.find(buf.str().c_str())==std::string::npos) {
			throw std::runtime_error("blkaddr::tail - " + buf.str() + " - not found");
		}
		buf.str("");

		for(MossMap::iterator itr=cBegin; itr!=cEnd; itr++, point++) {
			itr->second = point->second;
			log.infoStream() << "tail src=" << point->first << " target=" << itr->first << l4::eol;
		}
	}

	buf << suite 
		<< ":" << std::setfill('0') << std::setw(4) << srcID 
		<< ":" << std::setfill('0') << std::setw(4) << mark
		<< std::ends;
	key = buf.str().c_str();
	MossMap::iterator point = directory->lower_bound(key);
	if(point == directory->end()) throw std::runtime_error("blkaddr::tail - " + buf.str() + " - not found");
	if(point->first.find(buf.str().c_str())==std::string::npos) {
		throw std::runtime_error("blkaddr::tail - " + buf.str() + " - not found");
	}
	buf.str("");

	MossMap::iterator last;
	if(srcID<9999) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (srcID+1) << std::ends;
	} else {
		buf << (suite+1) << ":0000" << std::ends;
	}
	key = buf.str().c_str();
	last = directory->upper_bound(key);
	buf.str("");

	// erase the tail
	directory->erase(point, last);
	log.noticeStream() << "tail complete" << l4::eol;
}

void blkaddr::enumerate(std::vector<std::string> &fields, const MossMap::const_iterator start, const MossMap::const_iterator end)
{
	for(MossMap::const_iterator itr=start; itr!=end; itr++) {
		std::string buf(itr->first.c_str());
		fields.push_back(buf);
		log.debugStream() << "enumerated " << itr->first << l4::eol;
	}
}

void blkaddr::interlace(std::vector< std::pair <std::string,std::string> > &listing, const short suite, const short id)
{
	static KeyType key(segment.get_segment_manager());
	std::ostringstream buf;

	log.debugStream() << " suite=" << suite << " image=" << id << l4::eol;
	if(!directory) throw std::runtime_error("blkaddr::list - moss directory not found");

	if(!suite) throw std::runtime_error("blkaddr::list - no suite");
	if(!id) throw std::runtime_error("blkaddr::list - no image id");
	std::string odd;
	std::string even;

	buf << suite << ":" << std::setfill('0') << std::setw(4) << id << std::ends;
	key = buf.str().c_str();
	MossMap::iterator start = directory->lower_bound(key);
	if(start == directory->end()) throw std::runtime_error("blkaddr::list - " + buf.str() + " - not found");
	if(start->first.find(buf.str().c_str())==std::string::npos) {
		throw std::runtime_error("blkaddr::list - " + buf.str() + " - not found");
	}
	log.debugStream() << "start=" << buf.str() << l4::eol;
	buf.str("");

	MossMap::iterator end;
	if(id<9999) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (id+1) << std::ends;
	} else {
		buf << (suite+1) << ":0000" << std::ends;
	}
	key = buf.str().c_str();
	end = directory->upper_bound(key);
	buf.str("");

	for(MossMap::const_iterator itr=start; itr!=end; itr++) {
		if(itr->first.find("1:v") != std::string::npos) {
			odd = itr->first.c_str();
		}
		if(itr->first.find("2:v") != std::string::npos) {
			even = itr->first.c_str();
		}
		if(itr->first.find("3:v") != std::string::npos) {
			odd = itr->first.c_str();
		}
		if(itr->first.find("4:v") != std::string::npos) {
			even = itr->first.c_str();
		}
		if(itr->first.find("5:v") != std::string::npos) {
			odd = itr->first.c_str();
		}
		if(itr->first.find("6:v") != std::string::npos) {
			even = itr->first.c_str();
		}
		if(itr->first.find("7:v") != std::string::npos) {
			odd = itr->first.c_str();
		}
		if(itr->first.find("8:v") != std::string::npos) {
			even = itr->first.c_str();
		}
		if(itr->first.find("9:v") != std::string::npos) {
			odd = itr->first.c_str();
		}
		if(itr->first.find("0:v") != std::string::npos) {
			even = itr->first.c_str();
		}
		if(!odd.empty() && !even.empty()) {
			odd.erase(odd.find(":v"));
			even.erase(even.find(":v"));
			listing.push_back(std::make_pair(odd, even));
			odd.clear();
			even.clear();
		}
	}
}

void blkaddr::progressive(std::vector< std::pair <std::string,std::string> > &listing, const short suite, const short id)
{
	static KeyType key(segment.get_segment_manager());
	std::ostringstream buf;

	log.debugStream() << " suite=" << suite << " image=" << id << l4::eol;
	if(!directory) throw std::runtime_error("blkaddr::list - moss directory not found");

	if(!suite) throw std::runtime_error("blkaddr::list - no suite");
	if(!id) throw std::runtime_error("blkaddr::list - no image id");
	std::string label;
	std::string empty;

	buf << suite << ":" << std::setfill('0') << std::setw(4) << id << std::ends;
	key = buf.str().c_str();
	MossMap::iterator start = directory->lower_bound(key);
	if(start == directory->end()) throw std::runtime_error("blkaddr::list - " + buf.str() + " - not found");
	if(start->first.find(buf.str().c_str())==std::string::npos) {
		throw std::runtime_error("blkaddr::list - " + buf.str() + " - not found");
	}
	buf.str("");

	MossMap::iterator end;
	if(id<9999) {
		buf << suite << ":" << std::setfill('0') << std::setw(4) << (id+1) << std::ends;
	} else {
		buf << (suite+1) << ":0000" << std::ends;
	}
	key = buf.str().c_str();
	end = directory->upper_bound(key);
	buf.str("");


	for(MossMap::const_iterator itr=start; itr!=end; itr++) {
		if(itr->first.find(":v") != std::string::npos) {
			label = itr->first.c_str();
			label.erase(label.find(":v"));
			listing.push_back(std::make_pair(label, empty));
			label.clear();
		}
	}
}

