/// \file
/// \copyright 2008 Thomson Grass Valley
/// \author adam clarke
/// \date 2008-06-06
/// \brief a collection of inserters and extractors to support ImageStore operations
#ifndef __blkaddr_h__
#define __blkaddr_h__

#include <boost/interprocess/sync/scoped_lock.hpp>
#include <boost/interprocess/sync/named_mutex.hpp>
#include <boost/interprocess/managed_shared_memory.hpp>
#include <boost/interprocess/allocators/allocator.hpp>
#include <boost/interprocess/containers/string.hpp>
#include <boost/interprocess/containers/flat_map.hpp>

#include <log4cpp/Category.hh>

/// \brief block address serialization functor
class blkaddr
{
        log4cpp::Category &log;
	boost::interprocess::named_mutex mutex;
public:
	/// \note that map<Key, MappedType>'s value_type is std::pair<const Key, MappedType>,
	/// so the allocator must allocate that pair.
	typedef boost::interprocess::allocator<char, boost::interprocess::managed_shared_memory::segment_manager> char_allocator;
	typedef boost::interprocess::basic_string<char, std::char_traits<char>, char_allocator> KeyType;
	typedef struct { size_t offset, allocated, consumed; }  MappedType;
	typedef std::pair<const KeyType, MappedType> ValueType;

	/// \brief Alias an STL compatible allocator of for the map.
	/// This allocator will allow to place containers
	/// in managed shared memory segments
	typedef boost::interprocess::allocator<ValueType, boost::interprocess::managed_shared_memory::segment_manager> ShmemAllocator;

	/// \brief Alias a map of ints that uses the previous STL-like allocator.
	/// \note the third parameter argument is the ordering function
	/// of the map, just like with std::map, used to compare the keys.
	typedef boost::interprocess::flat_map<KeyType, MappedType, std::less<KeyType>, ShmemAllocator> MossMap;

	static boost::interprocess::managed_shared_memory blkaddr::segment;  ///< shared memory segment name
	static MossMap *directory; ///< points to the shared memory map
	static size_t *nextAvail; ///< point to offset of next available memory block

	void enumerate(std::vector<std::string> &fields, const MossMap::const_iterator start, const MossMap::const_iterator end);
public:
	blkaddr(void);
	void dump(std::ostream &sink) const;
        size_t assign(const std::string &label, const size_t segs, const size_t used=0);
        void consume(const std::string &id, const size_t segs);
        size_t addr(const std::string &id) const;
        size_t used(const std::string &id) const;

        bool move(const short suite, const long srcID, const long targetID);
        bool swap(const short suite, const long srcID, const long targetID);
        void append(const short suite, const long srcID, const long targetID);
        void marry(const short suite, const long srcID, const long targetID);
        void head(const short suite, const long srcID, const long mark, const long targetID=0);
        void tail(const short suite, const long srcID, const long mark, const long targetID=0);

	void interlace( std::vector< std::pair <std::string,std::string> > &listings, const short suite, const short id);
	void progressive( std::vector< std::pair <std::string,std::string> > &listings, const short suite, const short id);
};

#endif // __blkaddr_h__

