#include <cerrno>
#include <fstream>
#include <iostream>

#include <sys/mount.h>
#include <sys/wait.h>

#include <boost/lexical_cast.hpp>
#include <boost/program_options.hpp>

#include <log4cpp/Category.hh>
#include <log4cpp/BasicConfigurator.hh>
#include <log4cpp/PropertyConfigurator.hh>

#include <tiffio.h>

#include "blkaddr.h"

namespace po = boost::program_options;
namespace l4 = log4cpp;

static char mntcmds[7][80] = {
			"mount",
			"-t", "smbfs",
			"-o", "guest",
			"\\\\10.16.20.99\\SharedFolder", "/mnt/images"
	};
static const size_t essence_width  = 1920;
static const size_t essence_length = 1080;
static const size_t thumb_width  = 84;
static const size_t thumb_length = 48;
static const loff_t BLKSIZ = 112 * sizeof(long);

/// \brief extracts an 8bit red value from a 10bit ycbcr data stream
/// \param pos points to a 10bit luminance value, which is preceeded by a Cr value
/// \returns the derived an 8bit (0-255) red value
/// \note this is an approximate (CPU conserving) integer math based implementation,
/// suitable for use in generating thumbnails, but not suitable for use in essence conversion
unsigned char ycbcr2r(char *pos)
{
	short y = *(unsigned short *)pos;
	y >>= 2;
	if((y) < 0) y=0;
	if((y) > 255) y=255;

	short cr = *(unsigned short *)(pos-sizeof(short));
	cr >>= 2; cr -= 128;
	if((cr) < -128) y=-128;
	if((cr) > 128) y=128;

	short r = y + 7*cr/5;
	if((r) < 0) return 0;
	if((r) > 255) return 255;
	return r;
}

/// \brief extracts an 8bit green value from a 10bit ycbcr data stream
/// \param pos points to a 10bit luminance value, which is preceeded by a Cr value, and followed by a Cb value
/// \returns the derived an 8bit (0-255) green value
/// \note this is an approximate (CPU conserving) integer math based implementation,
/// suitable for use in generating thumbnails, but not suitable for use in essence conversion
unsigned char ycbcr2g(char *pos)
{
	short y = *(unsigned short *)pos;
	y >>= 2;
	if((y) < 0) y = 0;
	if((y) > 255) y = 255;

	short cb = *(unsigned short *)(pos+sizeof(short));
	cb >>= 2; cb -= 128;
	if((cb) < -128) y=-128;
	if((cb) > 128) y=128;

	short cr = *(unsigned short *)(pos-sizeof(short));
	cr >>= 2; cr -= 128;
	if((cr) < -128) y=-128;
	if((cr) > 128) y=128;

	short g = y - cb/3 - 3*cr/4;
	if((g) < 0) return 0;
	if((g) > 255) return 255;
	return g;
}

/// \brief extracts an 8bit blue value from a 10bit ycbcr data stream
/// \param pos points to a 10bit luminance value, which is followed by a Cb value
/// \returns the derived an 8bit (0-255) blue value
/// \note this is an approximate (CPU conserving) integer math based implementation,
/// suitable for use in generating thumbnails, but not suitable for use in essence conversion
unsigned char ycbcr2b(char *pos)
{
	short y = *(unsigned short *)pos;
	y >>= 2;
	if((y) < 0) y=0;
	if((y) > 255) y=255;

	short cb = *(unsigned short *)(pos+sizeof(short));
	cb >>= 2; cb -= 128;
	if((cb) < -128) y=-128;
	if((cb) > 128) y=128;

	short b = y + 9*cb/5;
	if((b) < 0) return 0;
	if((b) > 255) return 255;
	return b;
}

void write_video(const char *device, loff_t even, loff_t odd, TIFF *image)
{

  TIFFCreateDirectory(image);

  // We need to set some values for basic tags before we can add any data
  TIFFSetField(image, TIFFTAG_IMAGEWIDTH, essence_width);
  TIFFSetField(image, TIFFTAG_IMAGELENGTH, essence_length);
  TIFFSetField(image, TIFFTAG_BITSPERSAMPLE, 16);
  TIFFSetField(image, TIFFTAG_SAMPLESPERPIXEL, 2);
  TIFFSetField(image, TIFFTAG_EXTRASAMPLES, EXTRASAMPLE_UNSPECIFIED);
  TIFFSetField(image, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
  TIFFSetField(image, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
  TIFFSetField(image, TIFFTAG_ROWSPERSTRIP, 1);

  std::ifstream cm(device); // cache memory stream
  if(cm.is_open()) {
    char buf[18 * BLKSIZ];

    // Write the information to the file
    for(size_t y=0; y<essence_length; y++) {
      loff_t offset = y%1 ? even : odd;
      cm.seekg(offset+(y/2)*sizeof(buf));
      cm.read(buf, sizeof(buf));
      TIFFWriteScanline(image, buf, y);
    }
    cm.close();
  }

  uint32 diroff[2] = { 0, 0 };
  TIFFSetField(image, TIFFTAG_SUBIFD, 2, diroff);

  TIFFWriteDirectory(image);
}

void read_video(const char *device, loff_t even, loff_t odd, TIFF *image)
{
    if(TIFFReadDirectory(image)) {

	std::ofstream cm(device); // cache memory stream
	if(cm.is_open()) {
		char buf[18 * BLKSIZ];

		// Write the information to the file
		for(size_t y=0; y<essence_length; y++) {
			loff_t offset = y%1 ? even : odd;
			cm.seekp(offset+(y/2)*sizeof(buf));
			TIFFReadScanline(image, buf, y);
			cm.write(buf, sizeof(buf));
		}
		cm.close();
	}
    } else {
	throw std::runtime_error("essence not found");
    }
}

void write_key(const char *device, loff_t even, loff_t odd, TIFF *image)
{
  TIFFCreateDirectory(image);

  // We need to set some values for basic tags before we can add any data
  TIFFSetField(image, TIFFTAG_IMAGEWIDTH, essence_width);
  TIFFSetField(image, TIFFTAG_IMAGELENGTH, essence_length);
  TIFFSetField(image, TIFFTAG_BITSPERSAMPLE, 16);
  TIFFSetField(image, TIFFTAG_SAMPLESPERPIXEL, 2);
  TIFFSetField(image, TIFFTAG_EXTRASAMPLES, EXTRASAMPLE_UNSPECIFIED);
  TIFFSetField(image, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
  TIFFSetField(image, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
  TIFFSetField(image, TIFFTAG_ROWSPERSTRIP, 1);

  std::ifstream cm(device); // cache memory stream
  if(cm.is_open()) {
    char buf[18 * BLKSIZ];

    // Write the information to the file
    for(size_t y=0; y<essence_length; y++) {
      loff_t offset = y%1 ? even : odd;
      cm.seekg(offset+(y/2)*sizeof(buf));
      cm.read(buf, sizeof(buf));
      TIFFWriteScanline(image, buf, y);
    }
    cm.close();
  }

  TIFFSetField(image, TIFFTAG_SUBFILETYPE, FILETYPE_MASK);

  TIFFWriteDirectory(image);
}

void read_key(const char *device, loff_t even, loff_t odd, TIFF *image)
{
  uint16 count;
  uint32 diroff[32];
  TIFFGetField(image, TIFFTAG_SUBIFD, &count, &diroff);

  if(TIFFSetSubDirectory(image, diroff[0])) {
	std::ofstream cm(device); // cache memory stream
	if(cm.is_open()) {
		char buf[18 * BLKSIZ];

		// Write the information to the file
		for(size_t y=0; y<essence_length; y++) {
			loff_t offset = y%1 ? even : odd;
			cm.seekp(offset+(y/2)*sizeof(buf));
			TIFFReadScanline(image, buf, y);
			cm.write(buf, sizeof(buf));
		}
		cm.close();
	}
    } else {
	throw std::runtime_error("key not found");
    }
}

void write_thumb(const char *device, loff_t offset, TIFF *image)
{

  TIFFCreateDirectory(image);

  TIFFSetField(image, TIFFTAG_IMAGEWIDTH, thumb_width);
  TIFFSetField(image, TIFFTAG_IMAGELENGTH, thumb_length);
  TIFFSetField(image, TIFFTAG_BITSPERSAMPLE, 8);
  TIFFSetField(image, TIFFTAG_SAMPLESPERPIXEL, 3);
  TIFFSetField(image, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
  TIFFSetField(image, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
  TIFFSetField(image, TIFFTAG_ROWSPERSTRIP, 1);

  std::ifstream cm(device); // cache memory stream
  if(cm.is_open()) {
    char pixels[84*3];
    char buf[18*BLKSIZ];

    for(size_t y=0; y<48; y++) {
	cm.seekg(offset+10*(y+5)*sizeof(buf));
	cm.read(buf, sizeof(buf));
	for(size_t x=0; x<84; x++) {
		pixels[3*x+2] = ycbcr2r(buf+(40+x*88));
		pixels[3*x+1] = ycbcr2g(buf+(40+x*88));
		pixels[3*x+0] = ycbcr2b(buf+(40+x*88));
	}
        TIFFWriteScanline(image, pixels, y);
    }
    cm.close();
  }

  TIFFSetField(image, TIFFTAG_SUBFILETYPE, FILETYPE_REDUCEDIMAGE);

  TIFFWriteDirectory(image);
}

class imagereader
{
	const char *dev;
	TIFF *image;
	l4::Category &log;
public:
	imagereader(const char *dev, TIFF *image) 
		: dev(dev), image(image), log(l4::Category::getInstance("imagereader")) {};

	void operator()(const std::string &key)
	{
		std::string odd, even;
		blkaddr ba;
		try {
			size_t pos = key.find('/');
			if(pos == std::string::npos) {
				odd = even = key;
			} else {
				odd  = key.substr(0,pos);
				even = key.substr(pos,key.length());
			}
		} catch(...) {
			log.warnStream() << "bad arguments" << l4::eol;
			return;
		}

		try {
	                loff_t oddset  = ba.assign(odd+":v", 18*1080+1) * BLKSIZ;
	                loff_t evenset = ba.assign(even+":v", 18*1080+1) * BLKSIZ;
			read_video(dev, oddset, evenset, image);
			log.noticeStream() << "read image" << l4::eol;
		} catch(...) {
			log.warnStream() << "essence not cached" << l4::eol;
		}
/*
		try {
	                loff_t oddset  = ba.assign(odd+":k", 18*1080+1) * BLKSIZ;
	                loff_t evenset = ba.assign(even+":k", 18*1080+1) * BLKSIZ;
			read_key(dev, oddset, evenset, image);
			log.noticeStream() << "read key" << l4::eol;
		} catch(...) {
			log.noticeStream() << "key not cached" << l4::eol;
		}
*/
	}
};

class imagewriter
{
	const char *dev;
	TIFF *image;
	l4::Category &log;
public:
	imagewriter(const char *dev, TIFF *image) 
		: dev(dev), image(image), log(l4::Category::getInstance("imagewriter")) {};

	void operator()(const std::string &key) 
	{
		std::string odd, even;
		blkaddr ba;
		try {
			size_t pos = key.find('/');
			if(pos == std::string::npos) {
				odd = even = key;
			} else {
				odd  = key.substr(0,pos);
				even = key.substr(pos,key.length());
			}
		} catch(...) {
			log.warnStream() << "bad arguments" << l4::eol;
			return;
		}

		try {
	                loff_t oddset  = ba.addr(odd+":v") * BLKSIZ;
	                loff_t evenset = ba.addr(even+":v") * BLKSIZ;
			write_video(dev, oddset, evenset, image);
			log.noticeStream() << "wrote image" << l4::eol;
			write_thumb(dev, oddset, image);
			log.noticeStream() << "wrote thumb" << l4::eol;
		} catch(...) {
			log.warnStream() << "essence not found" << l4::eol;
		}

		try {
	                loff_t oddset  = ba.addr(odd+":k") * BLKSIZ;
	                loff_t evenset = ba.addr(even+":k") * BLKSIZ;
			write_key(dev, oddset, evenset, image);
			log.noticeStream() << "wrote key" << l4::eol;
		} catch(...) {
			log.noticeStream() << "key not found" << l4::eol;
	                loff_t offset = ba.addr(odd+":v") * BLKSIZ;
			write_thumb(dev, offset, image);
		}
	}
};

class blkreader
{
	const char *dev;
	TIFF *image;
	l4::Category &log;
public:
	blkreader(const char *dev, TIFF *image) 
		: dev(dev), image(image), log(l4::Category::getInstance("blkreader")) {};

	void operator()(loff_t offset) { }
};

class blkwriter
{
	const char *dev;
	TIFF *image;
	l4::Category &log;
public:
	blkwriter(const char *dev, TIFF *image) 
		: dev(dev), image(image), log(l4::Category::getInstance("blkwriter")) {};

	void operator()(loff_t offset) 
	{
		write_video(dev, offset, offset, image);
		log.noticeStream() << "wrote image" << l4::eol;
		write_thumb(dev, offset, image);
		log.noticeStream() << "wrote thumb" << l4::eol;
		write_key(dev, offset, offset, image);
		log.noticeStream() << "wrote key" << l4::eol;
	}
};

void blkaddr(loff_t &addr)
{
	addr *= BLKSIZ;
}

int main(int argc, char *argv[])
{
   try {
	std::vector<std::string> keys;
	std::vector<loff_t> addrs;
	const char *dev = "/dev/moCM1";
	std::string unc("/mnt/images/YCbCr.tif");

	po::options_description desc("Allowed options");

	desc.add_options()
		("help",  "produce help message")
		("log",  "log to console, or as controlled by config file")
		("cache", po::value<std::string>(), "cache memory device")
		("blkaddr", po::value< std::vector<loff_t> >(), "starting memory block addresses, repeat for each field")
		("key", po::value< std::vector<std::string> >(), "field address keys, repeat for each field")
		("read", po::value<std::string>(), "source filename, default /mnt/images/YCbCr.tif")
		("write", po::value<std::string>(), "destination filename, /mnt/images/default YCbCr.tif")
		("share", po::value<std::string>(), "share name, defaults to local filesystem")
	;

	po::variables_map vm;
	po::store(po::parse_command_line(argc, argv, desc), vm);
	po::notify(vm);

	l4::Category &log = l4::Category::getRoot();
	if (vm.count("log")) {
		try {
			l4::PropertyConfigurator::configure(".l4rc");
			std::cout << "logging configured" << std::endl;
			log.noticeStream() << "logging started" << l4::eol;
		} catch(log4cpp::ConfigureFailure e) {
			l4::BasicConfigurator::configure();
			log.noticeStream() << "console logging" << l4::eol;
		}
	}
	if (vm.count("help")) {
		std::cerr << desc << std::endl;
   		l4::Category::shutdown();
		exit(42);
	}
	if (vm.count("blkaddr")) {
		addrs = vm["blkaddr"].as< std::vector<loff_t> >();
		for_each(addrs.begin(), addrs.end(), blkaddr);
	}
	if (vm.count("cache")) {
		dev = vm["cache"].as<std::string>().c_str();
	}
	if (vm.count("key")) {
		keys = vm["key"].as< std::vector<std::string> >();
	}
	if (vm.count("tiff")) {
		log.noticeStream() << TIFFGetVersion() << l4::eol;
	}

	if (vm.count("share")) {
		strncpy(mntcmds[5],vm["share"].as<std::string>().c_str(),80);

		errno = 0;
		pid_t pid = fork();
		switch(pid) {
		case -1:
			log.warnStream() << "fork: " << strerror(errno) << l4::eol;
			l4::Category::shutdown();
			exit(-42);
		case 0:
			char * cmds[7];
			cmds[0] = mntcmds[0];
			cmds[1] = mntcmds[1];
			cmds[2] = mntcmds[2];
			cmds[3] = mntcmds[3];
			cmds[4] = mntcmds[4];
			cmds[5] = mntcmds[5];
			cmds[6] = mntcmds[6];
			cmds[7] = NULL;

			errno = 0;
			execv("/bin/mount", cmds);
			if(errno) log.warnStream() << "mount: " << strerror(errno) << l4::eol;
			l4::Category::shutdown();
			exit(42);
		default:
			int status;
			errno = 0;
			waitpid(pid, &status, 0);
			if(errno) log.warnStream() << "waitpid: " << strerror(errno) << l4::eol;
			break;
		}
	}

	if (vm.count("write")) {
		const std::string &unc = vm["write"].as<std::string>();
		// Open the TIFF file
		TIFF *image = TIFFOpen(unc.c_str(), "w");
		if(image) {
			imagewriter iw(dev, image);
			for_each(keys.begin(), keys.end(), iw);

			blkwriter bw(dev, image);
			for_each(addrs.begin(), addrs.end(), bw);

			// Close the file
			TIFFClose(image);
		} else {
			log.warnStream() << "Could not open " << unc << " for writing" << l4::eol;
			if (vm.count("share")) sleep(2); // give time for mount to complete
		}
	}

	if (vm.count("read")) {
		unc = vm["read"].as<std::string>();
		// Open the TIFF file
		TIFF *image = TIFFOpen(unc.c_str(), "r");
		if(image) {
			imagereader ir(dev, image);
			for_each(keys.begin(), keys.end(), ir);

			blkreader br(dev, image);
			for_each(addrs.begin(), addrs.end(), br);

			// Close the file
			TIFFClose(image);
		} else {
			log.warnStream() << "Could not open " << unc << " for reading" << l4::eol;
			if (vm.count("share")) sleep(2); // give time for mount to complete
		}
	}

	if (vm.count("share")) {
		// unmount the share
		errno = 0;
		umount("/mnt/images");
		if(errno) log.warnStream() << "umount: " << strerror(errno) << l4::eol;
	}

   } catch(...) {
	std::cerr << "caught unexpected exception" << std::endl;
	l4::Category::shutdown();
	return -42;
   }

   l4::Category::shutdown();
   return 0;
}


